#
# Board dependent Makefile for WRG-N16A
#

MYNAME:=WRG-N16A
Q?=@

#############################################################################
# build root file system image

build_tools:
	$(Q)[ -d images ] || mkdir -p images
	$(Q)make -C ./tools/squashfs-tools
	$(Q)make -C ./tools/buildimg

squashfs: build_tools clean_CVS remove_fsimg
	@echo -e "\033[32m$(MYNAME): building squashfs!\033[0m"
	$(Q)./tools/squashfs-tools/mksquashfs rootfs $(ROOTFS_IMG) -le

squashfs_lzma: build_tools clean_CVS remove_fsimg
	@echo -e "\033[32m$(MYNAME): building squashfs (LZMA)!\033[0m"
	$(Q)./tools/squashfs-tools/mksquashfs-lzma rootfs $(ROOTFS_IMG) -le
	$(Q)chmod 664 $(ROOTFS_IMG)

.PHONY: build_tools squashfs squashfs_lzma

#############################################################################
# build kernel image

kernel_image:
	@echo -e "\033[32m$(MYNAME): building kernel image file ...\033[0m"
	$(Q)rm -f $(KERNEL_IMG)
	$(Q)mipsel-linux-objcopy -O binary -R .note -R .comment -S $(KERNELDIR)/vmlinux vmlinux.bin
	$(Q)gzip vmlinux.bin
	$(Q)mv vmlinux.bin.gz $(KERNEL_IMG)

kernel_image_lzma: build_tools
	@echo -e "\033[32m$(MYNAME): building kernel image (LZMA)...\033[0m"
	$(Q)rm -f vmlinux.bin $(KERNEL_IMG)
	$(Q)mipsel-linux-objcopy -O binary -R .note -R .comment -S $(KERNELDIR)/vmlinux vmlinux.bin
	$(Q)./tools/squashfs-tools/lzma/lzma e vmlinux.bin $(KERNEL_IMG)
	$(Q)rm -f vmlinux.bin

.PHONY: kernel_image kernel_image_lzma

##########################################################################
#
# Common target
#

kernel_clean:
	@echo -e "\033[32m$(MYNAME): cleaning kernel ...\033[0m"
	$(Q)make -C kernel mrproper

kernel: kernel_clean
	@echo -e "\033[32m$(MYNAME) Building kernel ...\033[0m"
	$(Q)cp progs.board/kernel.config kernel/.config
	$(Q)make -C kernel oldconfig
	$(Q)make -C kernel dep
	$(Q)make -C kernel

ifeq (buildno, $(wildcard buildno))

ifdef PREDEFINE_RELIMAGE
RELIMAGE:=$(PREDEFINE_RELIMAGE)
else
BUILDNO:=$(shell cat buildno)
RELIMAGE:=$(shell echo $(ELBOX_MODEL_NAME)_v$(ELBOX_FIRMWARE_VERSION)_$(BUILDNO))
endif

.PHONY: lang_pack
lang_pack:
ifeq ($(strip $(ELBOX_CREATE_LANGUAGE_PACK)),y)
	$(Q)for i in $(LANGUAGE); do \
		./tools/squashfs-tools/mksquashfs-lzma language/$$i langpack.bin -le; \
		chmod 664 langpack.bin ; \
		./tools/buildimg/v2image -v -i langpack.bin -o images/$(RELIMAGE)_langpack_$$i.bin \
			-d /dev/mtdblock/5 -s $(ELBOX_SIGNATURE) -c -b; \
		rm -f langpack.bin ; \
		done
endif

magic_release: kernel_image_lzma squashfs
	@echo -e "\033[32m$(MYNAME) make F/W image (magic release image)!\033[0m"
	$(Q)./tools/buildimg/packimgs -v -o raw.img -i $(KERNEL_IMG) -i $(ROOTFS_IMG)
	$(Q)./tools/buildimg/v2image -v -i raw.img -o web.img -d /dev/mtdblock/2 -s $(ELBOX_BOARD_NAME)_aLpHa -c
	$(Q)rm -f raw.img
	$(Q)./tools/release.sh web.img $(RELIMAGE)_magic.bin

release: kernel_image_lzma squashfs
	@echo -e "\033[32m$(MYNAME) make F/W image (release image)!\033[0m"
	$(Q)./tools/buildimg/packimgs -v -o raw.img -i $(KERNEL_IMG) -i $(ROOTFS_IMG)
	$(Q)./tools/buildimg/v2image -v -i raw.img -o web.img -d /dev/mtdblock/2 -s $(ELBOX_SIGNATURE) -c
	$(Q)rm -f raw.img
	$(Q)./tools/release.sh web.img $(RELIMAGE).bin

tftpimage: kernel_image_lzma squashfs
	@echo -e "\033[32m$(MYNAME) make F/W image (tftp image)!\033[0m"
	$(Q)rm -f $(TFTPIMG)
	$(Q)./tools/buildimg/packimgs -v -o raw.img -i $(KERNEL_IMG) -i $(ROOTFS_IMG)
	$(Q)./tools/tftpimage.sh $(TFTPIMG)

else
release tftpimage:
	@echo -e "\033[32m$(MYNAME): Can not build image, ROOTFS is not created yet !\033[0m"
endif

.PHONY: kernel release tftpimage
