/*
 ***************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 *
 * (c) Copyright 2002-2006, Ralink Technology, Inc.
 *
 * All rights reserved. Ralink's source code is an unpublished work and the
 * use of a copyright notice does not imply otherwise. This source code
 * contains confidential trade secret material of Ralink Tech. Any attempt
 * or participation in deciphering, decoding, reverse engineering or in any
 * way altering the source code is stricitly prohibited, unless the prior
 * written consent of Ralink Technology, Inc. is obtained.
 ***************************************************************************

    Module Name:
    rt_timer.c

    Abstract:

    Revision History:
    Who         When            What
    --------    ----------      ----------------------------------------------
    Name        Date            Modification logs
    Steven Liu  2007-07-04      Initial version
*/

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/timer.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/rt2880/surfboardint.h>/*Erick for SDK2.10*/
#include <linux/interrupt.h>
#include "rt_timer.h"
#include "chip_reg_map.h"


struct timer_list wdg_timer;
static struct timer0_data tmr0;

/* 
 * System Clock = CPU Clock/2
 * For user easy configuration, We assume the unit of watch dog timer is 10ms, 
 * so we need to calculate the TMR1LOAD value.
 *
 * 233MHz: Unit=1/(116M/16384)=0.0001412 Sec, 10ms / 0.0001412 = 71 
 * 250MHz: Unit=1/(125M/16384)=0.0001311 Sec, 10ms / 0.0001311 = 76
 * 266MHz: Unit=1/(133M/16384)=0.0001231 Sec, 10ms / 0.0001231 = 81
 * 280MHz: Unit=1/(140M/16384)=0.0001170 Sec, 10ms / 0.0001170 = 85
 * 300MHz: Unit=1/(150M/16384)=0.0001092 Sec, 10ms / 0.0001092 = 92
 *
 * Example:
 *
 * In 266MHZ CPU, 5sec=500ms=>TMR1LOAD=500*81=40500
 * We need to fill out 40500 in TMR1LOAD register.
 *
 */
#ifdef CONFIG_RALINK_RT2880_VER_A
static int wdg_load_value = CONFIG_RALINK_WDG_TIMER * 76;
#elif CONFIG_RALINK_RT2880_VER_B
static int wdg_load_value = CONFIG_RALINK_WDG_TIMER * 81;
#endif


void set_timer_ebl(unsigned int timer, unsigned int ebl)
{
	unsigned int result;

	result=sysRegRead(timer);

	if(ebl==1){
		result |= (1<<7);
	}else {
		result &= ~(1<<7);
	}

	sysRegWrite(timer,result);
}


void set_timer_clock_prescale(unsigned int timer, enum timer_clock_freq prescale)
{
	unsigned int result;

	result=sysRegRead(timer);
	result &= ~0xF;
	result=result | (prescale&0xF);
	sysRegWrite(timer,result);

}

void set_timer_mode(unsigned int timer, enum timer_mode mode)
{
	unsigned int result;

	result=sysRegRead(timer);
	result &= ~(0x3<<4);
	result=result | (mode << 4);
	sysRegWrite(timer,result);

}

void setup_timer(struct timer_list * timer,
		void (*function)(unsigned long),
		unsigned long data)
{
	timer->function = function;
	timer->data = data;
	init_timer(timer);
}

void start_refresh_timer_load(unsigned long unused)
{
	sysRegWrite(TMR1LOAD, wdg_load_value);

	wdg_timer.expires = jiffies + HZ * CONFIG_RALINK_WDG_REFRESH_INTERVAL/100;
	add_timer(&wdg_timer);
}


int request_tmr_service(int interval, void (*function)(unsigned long), unsigned long data)
{
	unsigned int reg_val;
        unsigned long flags;

        spin_lock_irqsave(tmr0.tmr0_lock, flags);
       
	//Set Callback function
        tmr0.data = data;
        tmr0.tmr0_callback_function = function;
	
	//Timer 0 Interrupt Status Enable
	reg_val = sysRegRead(INTENA);
        reg_val |= 1;
        sysRegWrite(INTENA, reg_val);

	//Set Timer0 Mode
        set_timer_mode(TMR0CTL, PERIODIC);
	
	//Set Period Interval
	//266MHz: Unit=1/(133M/16384)=0.0001231 Sec, 1ms / 0.0001231 = 8
        set_timer_clock_prescale(TMR0CTL,SYS_CLK_DIV16384);
	sysRegWrite(TMR0LOAD, interval*8);

	//Enable Timer0
	set_timer_ebl(TMR0CTL,1);
       
	spin_unlock_irqrestore(tmr0.tmr0_lock, flags);

        return 0;
}

int unregister_tmr_service(void)
{
        unsigned int reg_val=0;
        unsigned long flags=0;

        spin_lock_irqsave(tmr0.tmr0_lock, flags);

	//Disable Timer0
        set_timer_ebl(TMR0CTL,0);
	
	//Timer0 Interrupt Status Disable
        reg_val = sysRegRead(INTENA);
        reg_val &= 0xfffe;
        sysRegWrite(INTENA, reg_val);

	//Unregister Callback Function
        tmr0.data = 0;
        tmr0.tmr0_callback_function = NULL;

        spin_unlock_irqrestore(tmr0.tmr0_lock, flags);

        return 0;
}


void rt2880tmr0_irq_handler(int irq, void *dev_id, struct pt_regs * regs)
{
        unsigned long flags;
        unsigned int reg_val;

        spin_lock_irqsave(tmr0.tmr0_lock, flags);

	//Writing '1' to TMRSTAT[0]:TMR0INT to clear the interrupt
        reg_val = sysRegRead(TMRSTAT);
        reg_val |= 1; 
        sysRegWrite(TMRSTAT, reg_val);

	//execute callback function
        if ( tmr0.tmr0_callback_function != NULL) {
                (tmr0.tmr0_callback_function)(tmr0.data);
        }

        spin_unlock_irqrestore(tmr0.tmr0_lock, flags);
        
}


int32_t __init timer_init_module(void)
{
    printk("Load RT2880 Timer Module(Wdg/Soft)\n");
   
    // initialize WDG timer (Timer1)
    setup_timer(&wdg_timer, start_refresh_timer_load, 0);
    set_timer_mode(TMR1CTL,WATCHDOG);
    set_timer_clock_prescale(TMR1CTL,SYS_CLK_DIV16384);
    start_refresh_timer_load(wdg_load_value);
    set_timer_ebl(TMR1CTL,1);

    // initialize Soft Timer (Timer0)
    spin_lock_init(&tmr0.tmr0_lock);
/*Erick for SDK2.10*/
    if(request_irq(SURFBOARDINT_TIMER0, rt2880tmr0_irq_handler, SA_INTERRUPT,
		    "rt2880_timer0", NULL)){
    	return 1;
    }


    return 0;
}

void __exit timer_cleanup_module(void)
{
    printk("Unload RT2880 Timer Module(Wdg/Soft)\n");

    set_timer_ebl(TMR1CTL,0);
    del_timer_sync(&wdg_timer);
    
    unregister_tmr_service();
}

module_init(timer_init_module);
module_exit(timer_cleanup_module);

EXPORT_SYMBOL(request_tmr_service);
EXPORT_SYMBOL(unregister_tmr_service);

MODULE_DESCRIPTION("Ralink Timer Module(Wdg/Soft)");
MODULE_AUTHOR("Steven/Bob");
MODULE_LICENSE("GPL");
