/*
 ***************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 *
 * (c) Copyright, Ralink Technology, Inc.
 *
 * All rights reserved. Ralink's source code is an unpublished work and the
 * use of a copyright notice does not imply otherwise. This source code
 * contains confidential trade secret material of Ralink Tech. Any attemp
 * or participation in deciphering, decoding, reverse engineering or in any
 * way altering the source code is stricitly prohibited, unless the prior
 * written consent of Ralink Technology, Inc. is obtained.
 ***************************************************************************
 *
 * $Id: rt2880_gpio.c,v 1.4 2007-06-29 02:26:38 winfred Exp $
 */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <linux/proc_fs.h>
#include "rt2880_gpio.h"
#include <asm/rt2880/surfboardint.h>/*Erick for SDK2.10*/

#ifdef  CONFIG_DEVFS_FS
#include <linux/devfs_fs_kernel.h>
static  devfs_handle_t devfs_handle;
#endif

#define NAME			"rt2880_gpio"
#define RT2880GPIO_DEVNAME	"gpio"
int rt2880gpio_major = 220;
u32 rt2880gpio_intp = 0;
int rt2880gpio_irqnum = 0;
u32 rt2880gpio_edge = 0;
#define RT2880GPIO_MAX_INFO (RT2880GPIO_DATA_LEN)
rt2880gpio_reg_info rt2880gpio_info[RT2880GPIO_MAX_INFO];

MODULE_DESCRIPTION("Ralink RT2880 SoC GPIO Driver");
MODULE_AUTHOR("Winfred Lu <winfred_lu@ralinktech.com.tw>");
MODULE_LICENSE("GPL");
rt2880gpio_reg_info info;

/***************************************************************************/
/* Added by Paul to support system reset */
static struct proc_dir_entry * system_reset = NULL;

static int proc_read_system_reset(char * buf, char ** start, off_t offset, int len, int * eof, void * data)
{
	char * p = buf;
	p += sprintf(p, "n/a\n");
	*eof = 1;
	return p - buf;
}

static int proc_write_system_reset(struct file * file, const char * buf, unsigned long count, void * data)
{
	printk("System resetting !!!\n");
	*(volatile u32 *)(RT2880_SYSCTL_ADDR+0x34) = 1;
	return count;
}

static int __init system_reset_init(struct proc_dir_entry * parent)
{
	system_reset = create_proc_entry("system_reset", 0644, parent);
	if (system_reset == NULL)
	{
		printk("%s: Unable to create system_reset entry!!\n",__FUNCTION__);
		return -ENOMEM;
	}
	system_reset->data = 0;
	system_reset->read_proc = proc_read_system_reset;
	system_reset->write_proc = proc_write_system_reset;
	system_reset->owner = THIS_MODULE;
	return 0;
}

static void __exit system_reset_exit(struct proc_dir_entry * parent)
{
	if (system_reset)
	{
		remove_proc_entry("system_reset", parent);
		system_reset = NULL;
	}
}

/***************************************************************************/



int rt2880gpio_ioctl(struct inode *inode, struct file *file, unsigned int req,
		unsigned long arg)
{
	unsigned long idx, tmp;
	rt2880gpio_reg_info info;

	idx = (req >> RT2880GPIO_DATA_LEN) & 0xFFL;
	req &= RT2880GPIO_DATA_MASK;

	switch(req) {
    case RT2880GPIO_SET_DIR:
        *(volatile u32 *)(RT2880_REG_PIODIR) = cpu_to_le32(arg);
        break;
	case RT2880GPIO_SET_DIR_BIT:
	 if (0L <= idx && idx < RT2880GPIO_SET1_FIRST_PIN) {
			tmp = cpu_to_le32(*(volatile u32 *)(RT2880_REG_PIODIR));
			if (arg & 1L) {
			tmp |= (1L << idx);
			}
			else {
			tmp &= ~(1L << idx);
			}
	      *(volatile u32 *)(RT2880_REG_PIODIR) = cpu_to_le32(tmp);
	  }else if (RT2880GPIO_SET1_FIRST_PIN <= idx && idx < RT2880GPIO_SET2_FIRST_PIN) {
			tmp = cpu_to_le32(*(volatile u32 *)(RT2880_REG_PIO1DIR));
			if (arg & 1L) {
			tmp |= (1L << (idx-RT2880GPIO_SET1_FIRST_PIN));
			}
			else {
			tmp &= ~(1L << (idx-RT2880GPIO_SET1_FIRST_PIN));
			}
	      *(volatile u32 *)(RT2880_REG_PIO1DIR) = cpu_to_le32(tmp);
	  }else	if (RT2880GPIO_SET2_FIRST_PIN <= idx && idx < RT2880GPIO_DATA_MAX_LEN) {
			tmp = cpu_to_le32(*(volatile u32 *)(RT2880_REG_PIO2DIR));
			if (arg & 1L) {
			tmp |= (1L << (idx-RT2880GPIO_SET2_FIRST_PIN));
			}
			else {
			tmp &= ~(1L << (idx-RT2880GPIO_SET2_FIRST_PIN));
			}
	      *(volatile u32 *)(RT2880_REG_PIO2DIR) = cpu_to_le32(tmp);
	  }
	  else
			return -EINVAL;	
		break;
	case RT2880GPIO_READ: //RT2880GPIO_READ_INT
		tmp = le32_to_cpu(*(volatile u32 *)(RT2880_REG_PIODATA));
		put_user(tmp, (int __user *)arg);
		break;
	case RT2880GPIO_WRITE: //RT2880GPIO_WRITE_INT
		*(volatile u32 *)(RT2880_REG_PIODATA) = cpu_to_le32(arg);
		break;
	case RT2880GPIO_READ_BIT:
		if (0L <= idx && idx < RT2880GPIO_SET1_FIRST_PIN) {
		tmp = le32_to_cpu(*(volatile u32 *)(RT2880_REG_PIODATA));
			tmp = (tmp >> idx) & 1L;
			put_user(tmp, (int __user *)arg);
		}else if (RT2880GPIO_SET1_FIRST_PIN <= idx && idx < RT2880GPIO_SET2_FIRST_PIN) {
		    tmp = le32_to_cpu(*(volatile u32 *)(RT2880_REG_PIO1DATA));
			tmp = (tmp >> (idx-RT2880GPIO_SET1_FIRST_PIN)) & 1L;
			put_user(tmp, (int __user *)arg);
		}else if (RT2880GPIO_SET2_FIRST_PIN <= idx && idx < RT2880GPIO_DATA_MAX_LEN) {
		    tmp = le32_to_cpu(*(volatile u32 *)(RT2880_REG_PIO2DATA));
			tmp = (tmp >> (idx-RT2880GPIO_SET2_FIRST_PIN)) & 1L;
			put_user(tmp, (int __user *)arg);
		}
		else
			return -EINVAL;
		break;
	case RT2880GPIO_WRITE_BIT:
		if (0L <= idx && idx < RT2880GPIO_DATA_LEN) {
			tmp =le32_to_cpu(*(volatile u32 *)(RT2880_REG_PIODATA));
			if (arg & 1L)
				tmp |= (1L << idx);
			else
				tmp &= ~(1L << idx);
			*(volatile u32 *)(RT2880_REG_PIODATA)= cpu_to_le32(tmp);
		}else if (RT2880GPIO_SET1_FIRST_PIN <= idx && idx < RT2880GPIO_SET2_FIRST_PIN) {
			tmp =le32_to_cpu(*(volatile u32 *)(RT2880_REG_PIO1DATA));
			if (arg & 1L)
				tmp |= (1L << (idx-RT2880GPIO_SET1_FIRST_PIN));
			else
				tmp &= ~(1L << (idx-RT2880GPIO_SET1_FIRST_PIN));
			*(volatile u32 *)(RT2880_REG_PIO1DATA)= cpu_to_le32(tmp);
		}else if (RT2880GPIO_SET2_FIRST_PIN <= idx && idx < RT2880GPIO_DATA_MAX_LEN) {
			tmp =le32_to_cpu(*(volatile u32 *)(RT2880_REG_PIO2DATA));
			if (arg & 1L)
				tmp |= (1L << (idx-RT2880GPIO_SET2_FIRST_PIN));
			else
				tmp &= ~(1L << (idx-RT2880GPIO_SET2_FIRST_PIN));
			*(volatile u32 *)(RT2880_REG_PIO2DATA)= cpu_to_le32(tmp);
		}
		else
			return -EINVAL;
		break;
	case RT2880GPIO_READ_BYTE:
		tmp = le32_to_cpu(*(volatile u32 *)(RT2880_REG_PIODATA));
		if (0L <= idx && idx < RT2880GPIO_DATA_LEN/8) {
			tmp = (tmp >> idx*8) & 0xFFL;
			put_user(tmp, (int __user *)arg);
		}
		else
			return -EINVAL;
		break;
	case RT2880GPIO_WRITE_BYTE:
		if (0L <= idx && idx < RT2880GPIO_DATA_LEN/8) {
			tmp =le32_to_cpu(*(volatile u32 *)(RT2880_REG_PIODATA));
			tmp &= ~(0xFFL << idx*8);
			tmp |= ((arg & 0xFFL) << idx*8);
			*(volatile u32 *)(RT2880_REG_PIODATA)= cpu_to_le32(tmp);
		}
		else
			return -EINVAL;
		break;
	case RT2880GPIO_ENABLE_INTP:
		*(volatile u32 *)(RT2880_REG_INTENA) = cpu_to_le32(0x8);
		break;
	case RT2880GPIO_DISABLE_INTP:
		*(volatile u32 *)(RT2880_REG_INTDIS) = cpu_to_le32(0x8);
		break;
	case RT2880GPIO_REG_IRQ:
		copy_from_user(&info, (rt2880gpio_reg_info *)arg, sizeof(info));
		if (0 <= info.irq && info.irq < RT2880GPIO_MAX_INFO) {
			rt2880gpio_info[info.irq].pid = info.pid;
			tmp = le32_to_cpu(*(volatile u32 *)(RT2880_REG_PIORENA));
			tmp |= (0x1 << info.irq);
			*(volatile u32 *)(RT2880_REG_PIORENA) = cpu_to_le32(tmp);
			tmp = le32_to_cpu(*(volatile u32 *)(RT2880_REG_PIOFENA));
			tmp |= (0x1 << info.irq);
			*(volatile u32 *)(RT2880_REG_PIOFENA) = cpu_to_le32(tmp);
		}
		else
			printk(KERN_ERR NAME ": irq number(%d) out of range\n",
					info.irq);
		break;
	default:
		return -ENOIOCTLCMD;
	}
	return 0;
}

int rt2880gpio_open(struct inode *inode, struct file *file)
{
	MOD_INC_USE_COUNT;
	return 0;
}

int rt2880gpio_release(struct inode *inode, struct file *file)
{
	MOD_DEC_USE_COUNT;
	return 0;
}

struct file_operations rt2880gpio_fops =
{
	owner:		THIS_MODULE,
	ioctl:		rt2880gpio_ioctl,
	open:		rt2880gpio_open,
	release:	rt2880gpio_release,
};

int __init rt2880gpio_init(void)
{
	unsigned int i;
	u32 gpioSetting=0L;
	SET_MODULE_OWNER(&rt2880gpio_fops);

#ifdef  CONFIG_DEVFS_FS
	if (devfs_register_chrdev(rt2880gpio_major, RT2880GPIO_DEVNAME,
				&rt2880gpio_fops)) {
		printk(KERN_ERR NAME ": unable to register character device\n");
		return -EIO;
	}
	devfs_handle = devfs_register(NULL, RT2880GPIO_DEVNAME,
			DEVFS_FL_DEFAULT, rt2880gpio_major, 0,
			S_IFCHR | S_IRUGO | S_IWUGO, &rt2880gpio_fops, NULL);
#else
	int r = 0;
	r = register_chrdev(rt2880gpio_major, RT2880GPIO_DEVNAME,
			&rt2880gpio_fops);
	if (r < 0) {
		printk(KERN_ERR NAME ": unable to register character device\n");
		return r;
	}
	if (rt2880gpio_major == 0) {
		rt2880gpio_major = r;
		printk(KERN_DEBUG NAME ": got dynamic major %d\n", r);
	}
#endif
#ifdef  CONFIG_RT2880_GPIOMODE_UART1
	//set UARTF_PCM_MODE to UART mode
	*(volatile u32 *)(RT2880_REG_SYSCTL) = 
		cpu_to_le32(*(volatile u32 *)(RT2880_REG_SYSCTL) & 0xFFFFFFBF);
#endif		
	//config these pins to gpio mode
#ifdef	CONFIG_RT2880_GPIOMODE_I2C
	gpioSetting |= RT2880_GPIOMODE_I2C;
#endif
#ifdef	CONFIG_RT2880_GPIOMODE_SPI
	gpioSetting |= RT2880_GPIOMODE_SPI;
#endif
#ifdef	CONFIG_RT2880_GPIOMODE_UART1
	gpioSetting |= RT2880_GPIOMODE_UARTF;
#endif
#ifdef	CONFIG_RT2880_GPIOMODE_UART2
	gpioSetting |= RT2880_GPIOMODE_UARTL;
#endif
#ifdef	CONFIG_RT2880_GPIOMODE_JTAG
	gpioSetting |= RT2880_GPIOMODE_JTAG;
#endif
#ifdef	CONFIG_RT2880_GPIOMODE_MDIO
	gpioSetting |= RT2880_GPIOMODE_MDIO;
#endif
#ifdef	CONFIG_RT2880_GPIOMODE_SDRAM
	gpioSetting |= RT2880_GPIOMODE_SDRAM;
#endif
#ifdef	CONFIG_RT2880_GPIOMODE_PCI
	gpioSetting |= RT2880_GPIOMODE_PCI;
#endif
	if(gpioSetting) {
	*(volatile u32 *)(RT2880_REG_GPIOMODE) =
		cpu_to_le32(gpioSetting);
	}
	else {
	*(volatile u32 *)(RT2880_REG_GPIOMODE) =
		cpu_to_le32(RT2880_GPIOMODE_DFT);
	}	
	//enable gpio interrupt
	*(volatile u32 *)(RT2880_REG_INTENA) = cpu_to_le32(0x8);
	for (i = 0; i < RT2880GPIO_MAX_INFO; i++) {
		rt2880gpio_info[i].irq = i;
		rt2880gpio_info[i].pid = 0;
	}
	printk("Ralink RT2880 gpio driver initialized\n");
	system_reset_init(proc_root_driver);
	return 0;
}

void __exit rt2880gpio_exit(void)
{
#ifdef  CONFIG_DEVFS_FS
	devfs_unregister_chrdev(rt2880gpio_major, RT2880GPIO_DEVNAME);
	devfs_unregister(devfs_handle);
#else
	unregister_chrdev(rt2880gpio_major, RT2880GPIO_DEVNAME);
#endif
	system_reset_exit(proc_root_driver);
	//config these pins to normal mode
	*(volatile u32 *)(RT2880_REG_GPIOMODE) &= ~RT2880_GPIOMODE_DFT;
	//disable gpio interrupt
	*(volatile u32 *)(RT2880_REG_INTDIS) = cpu_to_le32(0x8);
	printk("Ralink RT2880 gpio driver exited\n");
}

/*
 * send a signal(SIGUSR1) to the registered user process whenever any gpio
 * interrupt comes
 * (called by interrupt handler)
 */
void rt2880gpio_notify_user(int usr)
{
	struct task_struct *p = NULL;

	if (rt2880gpio_irqnum < 0 || RT2880GPIO_MAX_INFO <= rt2880gpio_irqnum) {
		printk(KERN_ERR NAME ": gpio irq number out of range\n");
		return;
	}

	//don't send any signal if pid is 0 or 1
	if ((int)rt2880gpio_info[rt2880gpio_irqnum].pid < 2)
		return;
	p = find_task_by_pid(rt2880gpio_info[rt2880gpio_irqnum].pid);
	if (NULL == p) {
		printk(KERN_ERR NAME ": no registered process to notify\n");
		return;
	}

	if (usr == 1) {
		printk(KERN_NOTICE NAME ": sending a SIGUSR1 to process %d\n",
				rt2880gpio_info[rt2880gpio_irqnum].pid);
		send_sig(SIGUSR1, p, 0);
	}
	else if (usr == 2) {
		printk(KERN_NOTICE NAME ": sending a SIGUSR2 to process %d\n",
				rt2880gpio_info[rt2880gpio_irqnum].pid);
		send_sig(SIGUSR2, p, 0);
	}
}

/*
 * 1. save the PIOINT and PIOEDGE value
 * 2. clear PIOINT by writing 1
 * (called by interrupt handler)
 */
void rt2880gpio_save_clear_intp(void)
{
	rt2880gpio_intp = le32_to_cpu(*(volatile u32 *)(RT2880_REG_PIOINT));
	rt2880gpio_edge = le32_to_cpu(*(volatile u32 *)(RT2880_REG_PIOEDGE));
	*(volatile u32 *)(RT2880_REG_PIOINT) = cpu_to_le32(0x00FFFFFF);
	*(volatile u32 *)(RT2880_REG_PIOEDGE) = cpu_to_le32(0x00FFFFFF);
}

void rt2880gpio_irq_handler(unsigned int irq, struct irqaction *irqaction)
{
	extern unsigned long volatile jiffies;
	struct gpio_time_record {
		unsigned long falling;
		unsigned long rising;
	};
	static struct gpio_time_record record[RT2880GPIO_MAX_INFO];
	unsigned long now;
	int i;

	rt2880gpio_save_clear_intp();
	now = jiffies;
	for (i = 0; i < RT2880GPIO_MAX_INFO; i++) {
		if (! (rt2880gpio_intp & (1 << i)))
			continue;
		rt2880gpio_irqnum = i;
		if (rt2880gpio_edge & (1 << i)) { //rising edge
			if (record[i].rising != 0 && time_before_eq(now,
						record[i].rising + 30L)) {
				/*
				 * If the interrupt comes in a short period,
				 * it might be floating. We ignore it.
				 */
			}
			else {
				record[i].rising = now;
				if (time_before(now, record[i].falling + 200L)) {
					//one click
					rt2880gpio_notify_user(1);
				}
				else {
					//press for several seconds
					rt2880gpio_notify_user(2);
				}
			}
		}
		else { //falling edge
			record[i].falling = now;
		}
		break;
	}
}

struct irqaction rt2880gpio_irqaction = {
	rt2880gpio_irq_handler,
	SA_INTERRUPT,
	0,
	"rt2880_gpio",
	NULL,
	NULL
};

void __init rt2880gpio_init_irq(void)
{
	setup_irq(SURFBOARDINT_GPIO, &rt2880gpio_irqaction);
}

module_init(rt2880gpio_init);
module_exit(rt2880gpio_exit);

