
/*
 ***************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 *
 * (c) Copyright 2002, Ralink Technology, Inc.
 *
 * All rights reserved. Ralink's source code is an unpublished work and the
 * use of a copyright notice does not imply otherwise. This source code
 * contains confidential trade secret material of Ralink Tech. Any attemp
 * or participation in deciphering, decoding, reverse engineering or in any
 * way altering the source code is stricitly prohibited, unless the prior
 * written consent of Ralink Technology, Inc. is obtained.
 ***************************************************************************

    Module Name:
    ap_wds.c

    Abstract:
    Support WDS function.

    Revision History:
    Who       When            What
    ------    ----------      ----------------------------------------------
    Fonchi    02-13-2007      created
*/

#include "rt_config.h"


#define VAILD_KEY_INDEX( _X ) ((((_X) >= 0) && ((_X) < 4)) ? (TRUE) : (FALSE))


static INT WdsVirtualIFSendPackets(IN PNDIS_PACKET pSkb, IN PNET_DEV dev);
static INT WdsVirtualIF_open(IN PNET_DEV dev);
static INT WdsVirtualIF_close(IN PNET_DEV dev);
static INT WdsVirtualIF_ioctl(IN PNET_DEV net_dev, IN OUT struct ifreq *rq, IN INT cmd);

extern INT rt2860_ioctl(PNET_DEV net_dev, struct ifreq *rq, int cmd);

static INT WdsVirtualIFSendPackets(
	IN PNDIS_PACKET pSkb, 
	IN PNET_DEV dev)
{
	UCHAR i;
	VIRTUAL_ADAPTER *pVirtualAd = dev->priv;
	RTMP_ADAPTER *pAd = (PRTMP_ADAPTER) pVirtualAd->RtmpDev->priv;
	PNDIS_PACKET pPacket = (PNDIS_PACKET) pSkb;

	DBGPRINT(RT_DEBUG_INFO, ("%s --->\n", __FUNCTION__));

#ifdef RALINK_ATE
	if (pAd->ate.Mode != ATE_STOP)
	{
		RELEASE_NDIS_PACKET(pAd, pPacket, NDIS_STATUS_FAILURE);
		return 0;
	}
#endif // RALINK_ATE //

	if ((RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_BSS_SCAN_IN_PROGRESS)) ||
		(RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_RADIO_OFF))          ||
		(RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_RESET_IN_PROGRESS)))
	{
		RELEASE_NDIS_PACKET(pAd, pPacket, NDIS_STATUS_FAILURE);
		return 0;
	}
	
	if (!(pVirtualAd->RtmpDev->flags & IFF_UP))
	{
		RELEASE_NDIS_PACKET(pAd, pPacket, NDIS_STATUS_FAILURE);
		return 0;
	}

	for (i = 0; i < MAX_WDS_ENTRY; i++)
	{
		if (ValidWdsEntry(pAd, i) && (pAd->WdsTab.WdsEntry[i].dev == dev))
		{
			pAd->RalinkCounters.PendingNdisPacketCount ++;
			RTMP_SET_PACKET_SOURCE(pSkb, PKTSRC_NDIS);
			RTMP_SET_PACKET_MOREDATA(pSkb, FALSE);
			RTMP_SET_PACKET_NET_DEVICE_WDS(pSkb, i);
			RTPKT_TO_OSPKT(pSkb)->dev = pVirtualAd->RtmpDev;
			
			DBGPRINT(RT_DEBUG_INFO, ("VirtualIFSendPackets(WDS) - packet to (%d)\n", i));

			return rt2860_packet_xmit(pSkb);
		}
	}

	DBGPRINT(RT_DEBUG_INFO, ("VirtualIFSendPackets - needn't to send or net_device not exist.\n"));
	RELEASE_NDIS_PACKET(pAd, pPacket, NDIS_STATUS_FAILURE);

	DBGPRINT(RT_DEBUG_INFO, ("%s <---\n", __FUNCTION__));
	return 0;
}

static INT WdsVirtualIF_open(
	IN	PNET_DEV dev)
{
	VIRTUAL_ADAPTER	*pVirtualAd = dev->priv;
#ifdef RTL865X_SOC
	RTMP_ADAPTER	*pAd = pVirtualAd->RtmpDev->priv;
	INT				index;
	unsigned int 	linkid;
#endif

	DBGPRINT(RT_DEBUG_TRACE, ("%s: ===> VirtualIF_open\n", dev->name));

	netif_start_queue(pVirtualAd->VirtualDev);
	
	return 0;
}

static INT WdsVirtualIF_close(
	IN PNET_DEV dev)
{
	VIRTUAL_ADAPTER	*pVirtualAd = dev->priv;

	DBGPRINT(RT_DEBUG_TRACE, ("%s: ===> VirtualIF_close\n", dev->name));

	netif_stop_queue(pVirtualAd->VirtualDev);
	
	return 0;
}

static INT WdsVirtualIF_ioctl(
	IN PNET_DEV net_dev, 
	IN OUT struct ifreq *rq, 
	IN INT cmd)
{
	VIRTUAL_ADAPTER *pVirtualAd = net_dev->priv;
	RTMP_ADAPTER *pAd = pVirtualAd->RtmpDev->priv;
	// struct iwreq *wrq = (struct iwreq *) rq;
	
	if(!RTMP_TEST_FLAG(pAd, fRTMP_ADAPTER_INTERRUPT_IN_USE))
	{
		DBGPRINT(RT_DEBUG_TRACE, ("VirtualIF_ioctl(%s)::Network is down!\n", net_dev->name));
		return -ENETDOWN;
	}

	return rt2860_ioctl(net_dev, rq, cmd);
}

/*
	==========================================================================
	Description:
		Check the WDS Entry is valid or not.
	==========================================================================
 */
BOOLEAN ValidWdsEntry(
	IN PRTMP_ADAPTER pAd,
	IN UCHAR WdsIndex)
{
	BOOLEAN result;
	PMAC_TABLE_ENTRY pMacEntry;

	do
	{
		if (WdsIndex >= MAX_WDS_ENTRY)
		{
			result = FALSE;
			break;
		}

		if (pAd->WdsTab.WdsEntry[WdsIndex].Valid != TRUE)
		{
			result = FALSE;
			break;
		}

		if ((pAd->WdsTab.WdsEntry[WdsIndex].MacTabMatchWCID==0)
			|| (pAd->WdsTab.WdsEntry[WdsIndex].MacTabMatchWCID >= MAX_LEN_OF_MAC_TABLE))
		{
			result = FALSE;
			break;
		}

		pMacEntry = &pAd->MacTab.Content[pAd->WdsTab.WdsEntry[WdsIndex].MacTabMatchWCID];
		if (pMacEntry->ValidAsWDS != TRUE)
		{
			result = FALSE;
			break;
		}
			
		result = TRUE;
	} while(FALSE);

	return result;
}

/*
	==========================================================================
	Description:
		Delete all WDS Entry in pAd->MacTab
	==========================================================================
 */
BOOLEAN MacTableDeleteWDSEntry(
	IN PRTMP_ADAPTER pAd,
	IN USHORT wcid,
	IN PUCHAR pAddr)
{
	MAC_TABLE_ENTRY *pEntry;

	if (wcid >= MAX_LEN_OF_MAC_TABLE)
		return FALSE;

	// delete one WDS entry
	NdisAcquireSpinLock(&pAd->WdsTabLock);

	pEntry = &pAd->MacTab.Content[wcid];
	if (pEntry->MatchWDSTabIdx < MAX_WDS_ENTRY)
	{
		UCHAR WdsTabIdx = pEntry->MatchWDSTabIdx;
		if (pAd->WdsTab.WdsEntry[WdsTabIdx].Valid == TRUE) 
		{
			pAd->WdsTab.WdsEntry[WdsTabIdx].Valid = FALSE;
			NdisZeroMemory(pAd->WdsTab.WdsEntry[WdsTabIdx].PeerWdsAddr, MAC_ADDR_LEN);
			pAd->WdsTab.Size --;
		}
	}

	MacTableDeleteEntry(pAd, wcid, pAddr);

	NdisReleaseSpinLock(&pAd->WdsTabLock);

	return TRUE;
}

/*
================================================================
Description : because WDS and CLI share the same WCID table in ASIC. 
WDS entry also insert to pAd->MacTab.content[].  Such is marked as ValidAsWDS = TRUE.
Also fills the pairwise key.
Because front MAX_AID_BA entries have direct mapping to BAEntry, which is only used as CLI. So we insert WDS
from index MAX_AID_BA.
================================================================
*/
MAC_TABLE_ENTRY *MacTableInsertWDSEntry(
	IN  PRTMP_ADAPTER   pAd, 
	IN  PUCHAR pAddr)
{
	int i;
	UINT WdsTabIdx = 0;
	PMAC_TABLE_ENTRY pEntry = NULL;
	HTTRANSMIT_SETTING HTPhyMode;

	// if FULL, return
	if (pAd->MacTab.Size >= MAX_LEN_OF_MAC_TABLE)
		return NULL;

	if((pEntry = WdsTableLookup(pAd, pAddr, TRUE)) != NULL)
		return pEntry;

	// allocate one WDS entry
	NdisAcquireSpinLock(&pAd->WdsTabLock);

	do
	{
		for (i = 0; i < MAX_WDS_ENTRY; i++)
		{
			if ((pAd->WdsTab.Mode >= WDS_LAZY_MODE) && !WDS_IF_UP_CHECK(pAd, i))
				continue;

			if (pAd->WdsTab.WdsEntry[i].Valid == FALSE)
			{
				pAd->WdsTab.WdsEntry[i].Valid = TRUE;
				pAd->WdsTab.Size ++;
				COPY_MAC_ADDR(pAd->WdsTab.WdsEntry[i].PeerWdsAddr, pAddr);
				WdsTabIdx = i;
				break;
			}
			else if (MAC_ADDR_EQUAL(pAd->WdsTab.WdsEntry[i].PeerWdsAddr, pAddr))
			{
				WdsTabIdx = i;
				break;
			}
			
		}

		if (i == MAX_WDS_ENTRY)
		{
			DBGPRINT(RT_DEBUG_ERROR, ("%s: Unable to allocate WdsEntry.\n", __FUNCTION__));
			pEntry = NULL;
			break;
		}

		// allocate one MAC entry
		pEntry = MacTableInsertEntry(pAd, pAddr, (WdsTabIdx + MIN_NET_DEVICE_FOR_WDS), TRUE);
		if (pEntry)
		{
			pEntry->PortSecured = WPA_802_1X_PORT_SECURED;

			// specific Max Tx Rate for Wds link.
			NdisZeroMemory(&HTPhyMode, sizeof(HTTRANSMIT_SETTING));
			switch (pAd->WdsTab.PhyMode)
			{
				case 0xff: // user doesn't specific a Mode for WDS link.
				case MODE_OFDM: // specific OFDM mode.
					HTPhyMode.field.MODE = MODE_OFDM;
					HTPhyMode.field.MCS = 7;
					pEntry->RateLen = 8;
					break;

				case MODE_CCK:
					HTPhyMode.field.MODE = MODE_CCK;
					HTPhyMode.field.MCS = 3;
					pEntry->RateLen = 4;
					break;

				case MODE_HTMIX:
					HTPhyMode.field.MCS = 7;
					HTPhyMode.field.ShortGI = pAd->CommonCfg.HTPhyMode.field.ShortGI;
					HTPhyMode.field.BW = pAd->CommonCfg.HTPhyMode.field.BW;
					HTPhyMode.field.STBC = pAd->CommonCfg.HTPhyMode.field.STBC;
					HTPhyMode.field.MODE = MODE_HTMIX;
					pEntry->RateLen = 12;
					break;

				case MODE_HTGREENFIELD:
					HTPhyMode.field.MCS = 7;
					HTPhyMode.field.ShortGI = pAd->CommonCfg.HTPhyMode.field.ShortGI;
					HTPhyMode.field.BW = pAd->CommonCfg.HTPhyMode.field.BW;
					HTPhyMode.field.STBC = pAd->CommonCfg.HTPhyMode.field.STBC;
					HTPhyMode.field.MODE = MODE_HTGREENFIELD;
					pEntry->RateLen = 12;
					break;

				default:
					break;
			}

			pEntry->MaxHTPhyMode.word = HTPhyMode.word;
			pEntry->MinHTPhyMode.word = pAd->CommonCfg.MinHTPhyMode.word;
			pEntry->HTPhyMode.word = pEntry->MaxHTPhyMode.word;

			if (pAd->WdsTab.PhyMode >= MODE_HTMIX)
			{
				if (pAd->CommonCfg.RegTransmitSetting.field.MCS != MCS_AUTO)
				{
					printk("@@@ pAd->CommonCfg.RegTransmitSetting.field.MCS = %d\n",
						pAd->CommonCfg.RegTransmitSetting.field.MCS);
					if (pAd->CommonCfg.RegTransmitSetting.field.MCS == 32)
					{
						// Fix MCS as HT Duplicated Mode
						pEntry->MaxHTPhyMode.field.BW = 1;
						pEntry->MaxHTPhyMode.field.MODE = MODE_HTMIX;
						pEntry->MaxHTPhyMode.field.STBC = 0;
						pEntry->MaxHTPhyMode.field.ShortGI = 0;
						pEntry->MaxHTPhyMode.field.MCS = 32;
					}
					else if (pEntry->MaxHTPhyMode.field.MCS > pAd->CommonCfg.HTPhyMode.field.MCS)
					{
						// STA supports fixed MCS 
						pEntry->MaxHTPhyMode.field.MCS = pAd->CommonCfg.HTPhyMode.field.MCS;
					}
				}

				pEntry->HTCapability.MCSSet[0] = pAd->CommonCfg.HtCapability.MCSSet[0];
				pEntry->HTCapability.MCSSet[1] = pAd->CommonCfg.HtCapability.MCSSet[1];
			}

			pAd->WdsTab.WdsEntry[WdsTabIdx].MacTabMatchWCID = (UCHAR)pEntry->Aid;
			pEntry->MatchWDSTabIdx = WdsTabIdx;

			//AsicAddKeyEntry(pAd, pEntry->Aid, BSS0, 0, &pAd->WdsTab.Wpa_key, TRUE, TRUE);
			AsicUpdateWdsEncryption(pAd, pEntry->Aid);

			DBGPRINT(RT_DEBUG_TRACE, ("MacTableInsertWDSEntry - allocate entry #%d, Total= %d\n",i, pAd->MacTab.Size));
			break;
		}
	}while(FALSE);

	NdisReleaseSpinLock(&pAd->WdsTabLock);

	return pEntry;
}

MAC_TABLE_ENTRY *WdsTableLookupByWcid(
	IN PRTMP_ADAPTER pAd,
	IN UCHAR wcid,
	IN PUCHAR pAddr,
	IN BOOLEAN bResetIdelCount)
{
	//USHORT HashIdx;
	ULONG WdsIndex;
	PMAC_TABLE_ENTRY pCurEntry = NULL;
	PMAC_TABLE_ENTRY pEntry = NULL;

	if (wcid <=0 || wcid >= MAX_LEN_OF_MAC_TABLE )
		return NULL;

	NdisAcquireSpinLock(&pAd->WdsTabLock);
	NdisAcquireSpinLock(&pAd->MacTabLock);

	do
	{
		//HashIdx = MAC_ADDR_HASH_INDEX(pAddr);
		//pCurEntry = pAd->MacTab.Hash[wcid];
		pCurEntry = &pAd->MacTab.Content[wcid];

		WdsIndex = 0xff;
		if ((pCurEntry) && (pCurEntry->ValidAsWDS == TRUE))
		{
			WdsIndex = pCurEntry->MatchWDSTabIdx;
		}

		if (WdsIndex == 0xff)
			break;

		if (pAd->WdsTab.WdsEntry[WdsIndex].Valid != TRUE)
			break;

		if (MAC_ADDR_EQUAL(pCurEntry->Addr, pAddr))
		{
			if(bResetIdelCount)
				pCurEntry->NoDataIdleCount = 0;
			pEntry = pCurEntry;
			break;
		}
	} while(FALSE);

	NdisReleaseSpinLock(&pAd->MacTabLock);
	NdisReleaseSpinLock(&pAd->WdsTabLock);

	return pEntry;
}

MAC_TABLE_ENTRY *WdsTableLookup(
	IN PRTMP_ADAPTER pAd,
	IN PUCHAR pAddr,
	IN BOOLEAN bResetIdelCount)
{
	USHORT HashIdx;
	ULONG WdsIndex;
	PMAC_TABLE_ENTRY pCurEntry = NULL;
	PMAC_TABLE_ENTRY pEntry = NULL;

	NdisAcquireSpinLock(&pAd->WdsTabLock);
	NdisAcquireSpinLock(&pAd->MacTabLock);

	do
	{
		HashIdx = MAC_ADDR_HASH_INDEX(pAddr);
		pCurEntry = pAd->MacTab.Hash[HashIdx];

		WdsIndex = 0xff;
		if ((pCurEntry) && (pCurEntry->ValidAsWDS == TRUE))
		{
			WdsIndex = pCurEntry->MatchWDSTabIdx;
		}

		if (WdsIndex == 0xff)
			break;

		if (pAd->WdsTab.WdsEntry[WdsIndex].Valid != TRUE)
			break;

		if (MAC_ADDR_EQUAL(pCurEntry->Addr, pAddr))
		{
			if(bResetIdelCount)
				pCurEntry->NoDataIdleCount = 0;
			pEntry = pCurEntry;
			break;
		}
	} while(FALSE);

	NdisReleaseSpinLock(&pAd->MacTabLock);
	NdisReleaseSpinLock(&pAd->WdsTabLock);

	return pEntry;
}

/*
	==========================================================================
	Description:
		This routine is called by APMlmePeriodicExec() every second to check if
		1. any associated client in PSM. If yes, then TX MCAST/BCAST should be
		   out in DTIM only
		2. any client being idle for too long and should be aged-out from MAC table
		3. garbage collect PSQ
	==========================================================================
*/
VOID WdsTableMaintenance(
	IN PRTMP_ADAPTER pAd)
{
	UCHAR idx;

	if (pAd->WdsTab.Mode != WDS_LAZY_MODE)
		return;

	for (idx = 0; idx < pAd->WdsTab.Size; idx++)
	{
		UCHAR wcid = pAd->WdsTab.WdsEntry[idx].MacTabMatchWCID;
		PMAC_TABLE_ENTRY pEntry = &pAd->MacTab.Content[wcid];

		if(pEntry->ValidAsWDS != TRUE)
			continue;

		NdisAcquireSpinLock(&pAd->WdsTabLock);
		NdisAcquireSpinLock(&pAd->MacTabLock);
		pEntry->NoDataIdleCount += 5;
		NdisReleaseSpinLock(&pAd->MacTabLock);
		NdisReleaseSpinLock(&pAd->WdsTabLock);

		// delete those MAC entry that has been idle for a long time
		if (pEntry->NoDataIdleCount >= MAC_TABLE_AGEOUT_TIME)
		{
			DBGPRINT(RT_DEBUG_TRACE, ("ageout %02x:%02x:%02x:%02x:%02x:%02x from WDS #%d after %d-sec silence\n",
					pEntry->Addr[0],pEntry->Addr[1],pEntry->Addr[2],pEntry->Addr[3],
					pEntry->Addr[4],pEntry->Addr[5], idx, MAC_TABLE_AGEOUT_TIME));
			MacTableDeleteWDSEntry(pAd, pEntry->Aid, pEntry->Addr);
		}
	}

}

// Register WDS interface
VOID RT2860_WDS_Init(
	IN PRTMP_ADAPTER pAd,
	IN PNET_DEV net_dev)
{
	INT index;
	PNET_DEV dev;
	VIRTUAL_ADAPTER *pWdsAd;
	INT i = 0;
	CHAR slot_name[IFNAMSIZ];
	PNET_DEV device;

	/* sanity check to avoid redundant virtual interfaces are created */
	if (pAd->flg_wds_init != FALSE)
		return;
	pAd->flg_wds_init = TRUE;

	for(index = 0; index < MAX_WDS_ENTRY; index++)
	{
#if LINUX_VERSION_CODE <= 0x20402		// Red Hat 7.1
		dev = alloc_netdev(sizeof(VIRTUAL_ADAPTER), "eth%d", ether_setup);
#else
		dev = alloc_etherdev(sizeof(VIRTUAL_ADAPTER));
#endif
		{// find available 
			for (i = 0; i < 32; i++)
			{
				sprintf(slot_name, "wds%d", i);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
				device = dev_get_by_name(slot_name);
#else
				for (device = dev_base; device != NULL; device = device->next)
				{
					if (strncmp(device->name, slot_name, 6) == 0)
					{
						break;
					}
				}
#endif
				if(device == NULL)
					break;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
				else
					dev_put(device);
#endif
			}

			if(i != 32)
			{
				PMAC_TABLE_ENTRY pWdsEntry;
				sprintf(dev->name, "wds%d", i);
				
				pWdsEntry = &pAd->MacTab.Content[pAd->WdsTab.WdsEntry[index].MacTabMatchWCID];
				DBGPRINT(RT_DEBUG_TRACE, ("Register WDS(virtual) interface(wds0%d)-%02X:%02X:%02X:%02X:%02X:%02X\n",
					i, pWdsEntry->Addr[0], pWdsEntry->Addr[1], pWdsEntry->Addr[2], pWdsEntry->Addr[3], pWdsEntry->Addr[4], pWdsEntry->Addr[5]));
			}
			else
			{
				DBGPRINT(RT_DEBUG_ERROR, ("Ethernet interface number overflow...\n"));
				break;
			}
		}
		
		ether_setup(dev);
		pWdsAd = dev->priv;

		pWdsAd->VirtualDev  = dev;
		pWdsAd->RtmpDev     = net_dev;

		NdisMoveMemory(&dev->dev_addr, &net_dev->dev_addr, MAC_ADDR_LEN);

		dev->hard_start_xmit    = (HARD_START_XMIT_FUNC)WdsVirtualIFSendPackets;
		dev->stop               = WdsVirtualIF_close;
		dev->open               = WdsVirtualIF_open;
		dev->do_ioctl           = WdsVirtualIF_ioctl;
		dev->priv_flags			= INT_WDS;

		pAd->WdsTab.PhyMode = 0xff;

		// Register this device
		register_netdevice(dev);
		pAd->WdsTab.WdsEntry[index].dev = dev;
	}
}

VOID RT2860_WDS_Close(
	IN PRTMP_ADAPTER pAd)
{
	UINT index;

	for(index = 0; index < MAX_WDS_ENTRY; index++)
	{
		// free resources of BA
		BASessionTearDownALL(pAd, index);

		if (pAd->WdsTab.WdsEntry[index].dev)
		{
			dev_close(pAd->WdsTab.WdsEntry[index].dev);
		}
	}
	return;
}


VOID RT2860_WDS_Remove(
	IN PRTMP_ADAPTER pAd)
{
	UINT index;

	for(index = 0; index < MAX_WDS_ENTRY; index++)
	{
		if (pAd->WdsTab.WdsEntry[index].dev)
	    {
			unregister_netdev(pAd->WdsTab.WdsEntry[index].dev);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
			free_netdev(pAd->WdsTab.WdsEntry[index].dev);
#else
			kfree(pAd->WdsTab.WdsEntry[index].dev);
#endif
		}
	}
}

VOID AsicUpdateWdsRxWCIDTable(
	IN PRTMP_ADAPTER pAd)
{
	UINT index;
	PMAC_TABLE_ENTRY pEntry = NULL;

	for(index = 0; index < MAX_WDS_ENTRY; index++)
	{
		if (pAd->WdsTab.WdsEntry[index].Valid != TRUE)
			continue;

		pEntry = MacTableInsertWDSEntry(pAd, pAd->WdsTab.WdsEntry[index].PeerWdsAddr);
		if ((pEntry) && (pEntry->ValidAsWDS))
		{
			// It's critical to do this
			AsicUpdateRxWCIDTable(pAd, pEntry->Aid, pEntry->Addr);
			AsicUpdateWdsEncryption(pAd, pEntry->Aid);
		}
	}

	return;
}

VOID AsicUpdateWdsEncryption(
	IN PRTMP_ADAPTER pAd,
	IN UCHAR wcid)
{
	//UINT index;
	PMAC_TABLE_ENTRY pEntry = NULL;

	do
	{
		if (wcid >= MAX_WDS_ENTRY)
			break;

		pEntry = &pAd->MacTab.Content[wcid];
		if (pAd->WdsTab.WdsEntry[pEntry->MatchWDSTabIdx].Valid != TRUE)
			break;

		if (pEntry->ValidAsWDS != TRUE)
			break;

		if (pAd->WdsTab.WepStatus == Ndis802_11WEPEnabled)
		{
			// update WCID attribute table and IVEIV table for this entry
			INT DefaultKeyId = pAd->ApCfg.MBSSID[MAIN_MBSSID].DefaultKeyId;

			if (!VAILD_KEY_INDEX(DefaultKeyId))
				break;

			RTMPAddWcidAttributeEntry(
				pAd,
				MAIN_MBSSID,
				pAd->ApCfg.MBSSID[MAIN_MBSSID].DefaultKeyId, // The value may be not zero
				pAd->SharedKey[MAIN_MBSSID][DefaultKeyId].CipherAlg,
				pEntry);
		}
		else if (((pAd->WdsTab.WepStatus == Ndis802_11Encryption2Enabled)
				|| (pAd->WdsTab.WepStatus == Ndis802_11Encryption3Enabled))
			&& (pAd->WdsTab.Wpa_key.KeyLen > 0))
		{
			AsicAddPairwiseKeyEntry(
				pAd,
				pEntry->Addr,
				pEntry->Aid,
				&pAd->WdsTab.Wpa_key);

			// update WCID attribute table and IVEIV table for this entry
			RTMPAddWcidAttributeEntry(
				pAd, 
				MAIN_MBSSID, 
				0, // The value may be not zero
				pAd->WdsTab.Wpa_key.CipherAlg, 
				pEntry);
		}
	} while (FALSE);

	return;
}

VOID WdsPeerBeaconProc(
	IN PRTMP_ADAPTER pAd,
	IN PMAC_TABLE_ENTRY pEntry,
	IN USHORT CapabilityInfo,
	IN UCHAR MaxSupportedRateIn500Kbps,
	IN UCHAR MaxSupportedRateLen,
	IN BOOLEAN bWmmCapable,
	IN ULONG ClientRalinkIe,
	IN HT_CAPABILITY_IE *pHtCapability,
	IN UCHAR HtCapabilityLen)
{
	UCHAR MaxSupportedRate = RATE_11;

	switch (MaxSupportedRateIn500Kbps)
	{
		case 108: MaxSupportedRate = RATE_54;   break;
		case 96:  MaxSupportedRate = RATE_48;   break;
		case 72:  MaxSupportedRate = RATE_36;   break;
		case 48:  MaxSupportedRate = RATE_24;   break;
		case 36:  MaxSupportedRate = RATE_18;   break;
		case 24:  MaxSupportedRate = RATE_12;   break;
		case 18:  MaxSupportedRate = RATE_9;    break;
		case 12:  MaxSupportedRate = RATE_6;    break;
		case 22:  MaxSupportedRate = RATE_11;   break;
		case 11:  MaxSupportedRate = RATE_5_5;  break;
		case 4:   MaxSupportedRate = RATE_2;    break;
		case 2:   MaxSupportedRate = RATE_1;    break;
		default:  MaxSupportedRate = RATE_11;   break;
	}

	if (pEntry && pEntry->ValidAsWDS)
	{
		pEntry->MaxSupportedRate = min(pAd->CommonCfg.MaxTxRate, MaxSupportedRate);
		pEntry->RateLen = MaxSupportedRateLen;
		if (pEntry->MaxSupportedRate < RATE_FIRST_OFDM_RATE)
		{
			pEntry->MaxHTPhyMode.field.MODE = MODE_CCK;
			pEntry->MaxHTPhyMode.field.MCS = pEntry->MaxSupportedRate;
			pEntry->MinHTPhyMode.field.MODE = MODE_CCK;
			pEntry->MinHTPhyMode.field.MCS = pEntry->MaxSupportedRate;
			pEntry->HTPhyMode.field.MODE = MODE_CCK;
			pEntry->HTPhyMode.field.MCS = pEntry->MaxSupportedRate;
		}
		else
		{
			pEntry->MaxHTPhyMode.field.MODE = MODE_OFDM;
			pEntry->MaxHTPhyMode.field.MCS = OfdmRateToRxwiMCS[pEntry->MaxSupportedRate];
			pEntry->MinHTPhyMode.field.MODE = MODE_OFDM;
			pEntry->MinHTPhyMode.field.MCS = OfdmRateToRxwiMCS[pEntry->MaxSupportedRate];
			pEntry->HTPhyMode.field.MODE = MODE_OFDM;
			pEntry->HTPhyMode.field.MCS = OfdmRateToRxwiMCS[pEntry->MaxSupportedRate];
		}
		pEntry->CapabilityInfo = CapabilityInfo;

		if (ClientRalinkIe & 0x00000004)
			CLIENT_STATUS_SET_FLAG(pEntry, fCLIENT_STATUS_RALINK_CHIPSET);
		else
			CLIENT_STATUS_CLEAR_FLAG(pEntry, fCLIENT_STATUS_RALINK_CHIPSET);
			
		if (pAd->CommonCfg.bAggregationCapable)
		{
			if ((pAd->CommonCfg.bPiggyBackCapable) && (ClientRalinkIe & 0x00000003) == 3)
			{
				CLIENT_STATUS_SET_FLAG(pEntry, fCLIENT_STATUS_AGGREGATION_CAPABLE);
				CLIENT_STATUS_SET_FLAG(pEntry, fCLIENT_STATUS_PIGGYBACK_CAPABLE);
				//RTMPSetPiggyBack(pAd, TRUE);
				DBGPRINT(RT_DEBUG_INFO, ("ASSOC -PiggyBack= 1\n"));
			}
			else if (ClientRalinkIe & 0x00000001)
			{
				CLIENT_STATUS_SET_FLAG(pEntry, fCLIENT_STATUS_AGGREGATION_CAPABLE);
				CLIENT_STATUS_CLEAR_FLAG(pEntry, fCLIENT_STATUS_PIGGYBACK_CAPABLE);
			}
			else
			{
				CLIENT_STATUS_CLEAR_FLAG(pEntry, fCLIENT_STATUS_AGGREGATION_CAPABLE);
				CLIENT_STATUS_CLEAR_FLAG(pEntry, fCLIENT_STATUS_PIGGYBACK_CAPABLE);
			}
		}
		else
		{
			CLIENT_STATUS_CLEAR_FLAG(pEntry, fCLIENT_STATUS_AGGREGATION_CAPABLE);
			if ((pAd->CommonCfg.bPiggyBackCapable) && (ClientRalinkIe & 0x00000002) == 2)
			{				
				CLIENT_STATUS_SET_FLAG(pEntry, fCLIENT_STATUS_PIGGYBACK_CAPABLE);
				//RTMPSetPiggyBack(pAd, TRUE);
				DBGPRINT(RT_DEBUG_TRACE, ("ASSOC -PiggyBack2= 1\n"));
			}
			else
				CLIENT_STATUS_CLEAR_FLAG(pEntry, fCLIENT_STATUS_PIGGYBACK_CAPABLE);
		}

		// If this Entry supports 802.11n, upgrade to HT rate.
		if ((HtCapabilityLen != 0) 
			&& (pAd->CommonCfg.PhyMode >= PHY_11ABGN_MIXED))
		{
			UCHAR	j, bitmask;//k,bitmask;
			CHAR    i;

			if ((pHtCapability->HtCapInfo.GF) && (pAd->CommonCfg.DesiredHtPhy.GF))
			{
				pEntry->MaxHTPhyMode.field.MODE = MODE_HTGREENFIELD;
			}
			else
			{	
				pEntry->MaxHTPhyMode.field.MODE = MODE_HTMIX;
				pAd->MacTab.fAnyStationNonGF = TRUE;
				pAd->CommonCfg.AddHTInfo.AddHtInfo2.NonGfPresent = 1;
			}

			if ((pHtCapability->HtCapInfo.ChannelWidth) && (pAd->CommonCfg.DesiredHtPhy.ChannelWidth))
			{
				pEntry->MaxHTPhyMode.field.BW= BW_40;
				pEntry->MaxHTPhyMode.field.ShortGI = ((pAd->CommonCfg.DesiredHtPhy.ShortGIfor40)&(pHtCapability->HtCapInfo.ShortGIfor40));
			}
			else
			{	
				pEntry->MaxHTPhyMode.field.BW = BW_20;
				pEntry->MaxHTPhyMode.field.ShortGI = ((pAd->CommonCfg.DesiredHtPhy.ShortGIfor20)&(pHtCapability->HtCapInfo.ShortGIfor20));
				pAd->MacTab.fAnyStation20Only = TRUE;
			}

			// find max fixed rate
			for (i=15; i>=0; i--)
			{	
				j = i/8;	
				bitmask = (1<<(i-(j*8)));
				if ( (pAd->CommonCfg.DesiredHtPhy.MCSSet[j]&bitmask) && (pHtCapability->MCSSet[j]&bitmask))
				{
					pEntry->MaxHTPhyMode.field.MCS = i;
					break;
				}
				if (i==0)
					break;
			}

			if ((pEntry->MaxHTPhyMode.field.MCS > pAd->CommonCfg.HTPhyMode.field.MCS) && (pAd->CommonCfg.HTPhyMode.field.MCS != MCS_AUTO))
				pEntry->MaxHTPhyMode.field.MCS = pAd->CommonCfg.HTPhyMode.field.MCS;
			pEntry->MaxHTPhyMode.field.STBC = (pHtCapability->HtCapInfo.RxSTBC & (pAd->CommonCfg.DesiredHtPhy.TxSTBC));
			pEntry->MpduDensity = pHtCapability->HtCapParm.MpduDensity;
			pEntry->MaxRAmpduFactor = pHtCapability->HtCapParm.MaxRAmpduFactor;
			pEntry->MmpsMode = (UCHAR)pHtCapability->HtCapInfo.MimoPs;
			pEntry->AMsduSize = (UCHAR)pHtCapability->HtCapInfo.AMsduSize;				
			pEntry->HTPhyMode.word = pEntry->MaxHTPhyMode.word;
			if (pAd->CommonCfg.DesiredHtPhy.AmsduEnable)
				CLIENT_STATUS_SET_FLAG(pEntry, fCLIENT_STATUS_AMSDU_INUSED);
			if (pHtCapability->HtCapInfo.ShortGIfor20)
				CLIENT_STATUS_SET_FLAG(pEntry, fCLIENT_STATUS_SGI20_CAPABLE);
			if (pHtCapability->HtCapInfo.ShortGIfor40)
				CLIENT_STATUS_SET_FLAG(pEntry, fCLIENT_STATUS_SGI40_CAPABLE);
			if (pHtCapability->HtCapInfo.TxSTBC)
				CLIENT_STATUS_SET_FLAG(pEntry, fCLIENT_STATUS_TxSTBC_CAPABLE);
			if (pHtCapability->HtCapInfo.RxSTBC)
				CLIENT_STATUS_SET_FLAG(pEntry, fCLIENT_STATUS_RxSTBC_CAPABLE);
			if (pHtCapability->ExtHtCapInfo.PlusHTC)				
				CLIENT_STATUS_SET_FLAG(pEntry, fCLIENT_STATUS_HTC_CAPABLE);
			if (pAd->CommonCfg.bRdg && pHtCapability->ExtHtCapInfo.RDGSupport)
				CLIENT_STATUS_SET_FLAG(pEntry, fCLIENT_STATUS_RDG_CAPABLE);	
			if (pHtCapability->ExtHtCapInfo.MCSFeedback == 0x03)
				CLIENT_STATUS_SET_FLAG(pEntry, fCLIENT_STATUS_MCSFEEDBACK_CAPABLE);

			NdisMoveMemory(&pEntry->HTCapability, pHtCapability, sizeof(HT_CAPABILITY_IE));
		}
		else
		{
			pAd->MacTab.fAnyStationIsLegacy = TRUE;
		}

		if (bWmmCapable || (pEntry->MaxHTPhyMode.field.MODE >= MODE_HTMIX))
		{
			CLIENT_STATUS_SET_FLAG(pEntry, fCLIENT_STATUS_WMM_CAPABLE);
		}
		else
		{
			CLIENT_STATUS_CLEAR_FLAG(pEntry, fCLIENT_STATUS_WMM_CAPABLE);
		}

		pEntry->HTPhyMode.word = pEntry->MaxHTPhyMode.word;
	}
}

VOID APWdsInitialize(
	IN PRTMP_ADAPTER pAd)
{
	INT i;

	pAd->WdsTab.Mode = 0;
	pAd->WdsTab.Size = 0;
    pAd->WdsTab.WepStatus = 0;
	NdisZeroMemory(&pAd->WdsTab.Wpa_key, sizeof(pAd->WdsTab.Wpa_key));
	pAd->WdsTab.PhyMode = 0;
	for (i = 0; i < MAX_WDS_ENTRY; i++)
	{
		pAd->WdsTab.WdsEntry[i].Valid = FALSE;
		pAd->WdsTab.WdsEntry[i].MacTabMatchWCID = 0;
	}
	return;	
}

