/*
 ***************************************************************************
 * Ralink Tech Inc.
 * 4F, No. 2 Technology 5th Rd.
 * Science-based Industrial Park
 * Hsin-chu, Taiwan, R.O.C.
 *
 * (c) Copyright 2002-2007, Ralink Technology, Inc.
 *
 * All rights reserved.	Ralink's source	code is	an unpublished work	and	the
 * use of a	copyright notice does not imply	otherwise. This	source code
 * contains	confidential trade secret material of Ralink Tech. Any attemp
 * or participation	in deciphering,	decoding, reverse engineering or in	any
 * way altering	the	source code	is stricitly prohibited, unless	the	prior
 * written consent of Ralink Technology, Inc. is obtained.
 ***************************************************************************

	Module Name:
	cmm_mat_iparp.c

	Abstract:
		MAT convert engine subroutine for ip base protocols, currently now we 
	just handle IP/ARP protocols.

	Revision History:
	Who			When			What
	--------	----------		----------------------------------------------
	Shiang      02/26/07      Init version
*/
#ifdef MAT_SUPPORT

#include "rt_config.h"

static NDIS_STATUS MATProto_IP_Init(VOID);
static NDIS_STATUS MATProto_IP_Exit(VOID);
static PUCHAR MATProto_IP_Rx(PRTMP_ADAPTER pAd, PNDIS_PACKET pSkb, PUCHAR pLayerHdr, UINT infIdx);
static PUCHAR MATProto_IP_Tx(PRTMP_ADAPTER pAd, PNDIS_PACKET pSkb, PUCHAR pLayerHdr, UINT infIdx);

static NDIS_STATUS MATProto_ARP_Init(VOID);
static NDIS_STATUS MATProto_ARP_Exit(VOID);
static PUCHAR MATProto_ARP_Rx(PRTMP_ADAPTER pAd, PNDIS_PACKET pSkb, PUCHAR pLayerHdr, UINT infIdx);
static PUCHAR MATProto_ARP_Tx(PRTMP_ADAPTER pAd, PNDIS_PACKET pSkb,PUCHAR pLayerHdr, UINT infIdx);

#define IPV4_ADDR_LEN 4

#define NEED_UPDATE_IPMAC_TB(Mac, IP) (IS_UCAST_MAC(Mac) && IS_GOOD_IP(IP))


typedef struct _IPMacMappingEntry
{
	UINT  ipAddr;	// In network order
	UCHAR macAddr[MAC_ADDR_LEN];
	UINT  lastTime;
	struct _IPMacMappingEntry *pNext;
}IPMacMappingEntry, *PIPMacMappingEntry;


typedef struct _IPMacMappingTable
{
	BOOLEAN			valid;
	IPMacMappingEntry *hash[MAT_MAX_HASH_ENTRY_SUPPORT+1]; //0~63 for specific station, 64 for broadcast MacAddress
}IPMacMappingTable;

static IPMacMappingTable IPMacTable=
{
	.valid = FALSE,
}; // Used for IP, ARP protocol

struct _MATProtoEntry MATProtoIPHandle =
{
	.init = MATProto_IP_Init,
	.tx = MATProto_IP_Tx,
	.rx = MATProto_IP_Rx,
	.exit = MATProto_IP_Exit,
};

struct _MATProtoEntry MATProtoARPHandle =
{
	.init = MATProto_ARP_Init,
	.tx = MATProto_ARP_Tx,
	.rx = MATProto_ARP_Rx,
	.exit =MATProto_ARP_Exit,
};

VOID dumpIPMacTb(int index)
{
	IPMacMappingEntry *pHead;
	int startIdx, endIdx;

	if (!IPMacTable.valid)
	{
		printk("%s():IPMacTable not init yet, so cannot do dump!\n", __FUNCTION__);
		return;
	}
	
	
	if(index < 0)
	{	// dump all.
		startIdx = 0;
		endIdx = MAT_MAX_HASH_ENTRY_SUPPORT;
	}
	else
	{	// dump specific hash index.
		startIdx = endIdx = index;
	}

	printk("%s(): Now dump IPMac Table!\n", __FUNCTION__);
	for(; startIdx<= endIdx; startIdx++)
	{
		pHead = IPMacTable.hash[startIdx];
	while(pHead)
	{
			printk("IPMac[%d]:\n", startIdx);
			printk("\t:IP=0x%x,Mac=%02x:%02x:%02x:%02x:%02x:%02x, jiffies=0x%x, next=%p\n", 
				pHead->ipAddr, pHead->macAddr[0],pHead->macAddr[1],pHead->macAddr[2],
				pHead->macAddr[3],pHead->macAddr[4],pHead->macAddr[5], pHead->lastTime,
					pHead->pNext);
		pHead = pHead->pNext;
	}
	}
	printk("----EndOfDump!\n");
	
}


static inline NDIS_STATUS getDstIPFromIpPkt(
	IN PUCHAR pIpHdr, 
	IN UINT *dstIP)
{
	
	if (!pIpHdr)
		return FALSE;
	
	NdisMoveMemory(dstIP, (pIpHdr + 16), 4); //shift 16 for IP header len before DstIP.
//	DBGPRINT(RT_DEBUG_TRACE, ("%s(): Get the dstIP=0x%x\n", __FUNCTION__, *dstIP));
	
	return TRUE;
}

static inline NDIS_STATUS getSrcIPFromIpPkt(
	IN PUCHAR pIpHdr,
	IN UINT   *pSrcIP)
{
	
	if (!pIpHdr)
		return FALSE;
	
	NdisMoveMemory(pSrcIP, (pIpHdr + 12), 4); //shift 12 for IP header len before DstIP.
//	DBGPRINT(RT_DEBUG_TRACE, ("%s(): Get the srcIP=0x%x\n", __FUNCTION__, *pSrcIP));
	
	return TRUE;
	
}

static NDIS_STATUS IPMacTableUpdate(
	IN PRTMP_ADAPTER	pAd,
	IN PUCHAR			pMacAddr,
	IN UINT			ipAddr)
{
	UINT 				hashIdx;
	IPMacMappingEntry	*pEntry = NULL, *pPrev = NULL, *pNewEntry =NULL;

	if (!IPMacTable.valid)
		return FALSE;

#if 0
	//We didn't need to maintain the Mac address of ourself.
	if(IS_EQUAL_MAC(pMacAddr, pAd->CurrentAddress))
		return 0;
#endif
	
	hashIdx = MAT_IP_ADDR_HASH_INDEX(ipAddr);

    pEntry = pPrev = IPMacTable.hash[hashIdx];
	while(pEntry)
	{
		// Find a existed IP-MAC Mapping entry
		if (ipAddr == pEntry->ipAddr)
    	{
/*			DBGPRINT(RT_DEBUG_TRACE, ("%s(): Got the Mac(%02x:%02x:%02x:%02x:%02x:%02x) of mapped IP(%d.%d.%d.%d)\n",
					__FUNCTION__, pEntry->macAddr[0],pEntry->macAddr[1],pEntry->macAddr[2],
					pEntry->macAddr[3],pEntry->macAddr[4],pEntry->macAddr[5],
					(ipAddr>>24) & 0xff, (ipAddr>>16) & 0xff, (ipAddr>>8) & 0xff, ipAddr & 0xff)); 
*/
			// compare is useless. So we directly copy it into the entry.
			NdisMoveMemory(pEntry->macAddr, pMacAddr, 6);
			pEntry->lastTime = jiffies;
			
	        return TRUE;
		}
        else
        {	// handle the age-out situation
        	if ((jiffies - pEntry->lastTime) > MAT_TB_ENTRY_AGEOUT_TIME)
        	{
        		// Remove the aged entry
        		if (pEntry == IPMacTable.hash[hashIdx])
				{
					IPMacTable.hash[hashIdx]= pEntry->pNext;
					pPrev = IPMacTable.hash[hashIdx];
        		}
				else 
				{	
	        		pPrev->pNext = pEntry->pNext;
				}
				MATDBEntryFree((PUCHAR)pEntry);
				
				pEntry = (pPrev == NULL ? NULL: pPrev->pNext);
        	} 
			else
			{
				pPrev = pEntry;
	            pEntry = pEntry->pNext;
        	}
        }
	}


	// Allocate a new IPMacMapping entry and insert into the hash
	pNewEntry = (IPMacMappingEntry *)MATDBEntryAlloc(sizeof(IPMacMappingEntry));
	if (pNewEntry != NULL)
	{	
		pNewEntry->ipAddr = ipAddr;
		NdisMoveMemory(pNewEntry->macAddr, pMacAddr, 6);
		pNewEntry->pNext = NULL;
		pNewEntry->lastTime = jiffies;

		if (IPMacTable.hash[hashIdx] == NULL)
		{	// Hash list is empty, directly assign it.
			IPMacTable.hash[hashIdx] = pNewEntry;
		} 
		else 
		{
			// Ok, we insert the new entry into the root of hash[hashIdx]
			pNewEntry->pNext = IPMacTable.hash[hashIdx];
			IPMacTable.hash[hashIdx] = pNewEntry;
		}
		//dumpIPMacTb(hashIdx); //for debug
		

		return TRUE;
	}
	
	return FALSE;
}

static PUCHAR IPMacTableLookUp(
	IN	PRTMP_ADAPTER	pAd,
	IN	UINT			ipAddr)
{
    UINT 				hashIdx;
    IPMacMappingEntry	*pEntry = NULL;

    if (!IPMacTable.valid)
        return NULL;
	
	// Use hash to find out the location of that entry and get the Mac address.
	hashIdx = MAT_IP_ADDR_HASH_INDEX(ipAddr);

//	spin_lock_irqsave(&IPMacTabLock, irqFlag);
    pEntry = IPMacTable.hash[hashIdx];
	while(pEntry)
	{
		if (pEntry->ipAddr == ipAddr)
        {
/*			DBGPRINT(RT_DEBUG_TRACE, ("%s(): dstMac=%02x:%02x:%02x:%02x:%02x:%02x for mapped dstIP(%d.%d.%d.%d)\n", 
					__FUNCTION__, pEntry->macAddr[0],pEntry->macAddr[1],pEntry->macAddr[2],
					pEntry->macAddr[3],pEntry->macAddr[4],pEntry->macAddr[5],
					(ipAddr>>24) & 0xff, (ipAddr>>16) & 0xff, (ipAddr>>8) & 0xff, ipAddr & 0xff)); 
*/
			
			//Update the lastTime to prevent the aging before pDA processed!
			pEntry->lastTime = jiffies; 
			
			return pEntry->macAddr;
        }
        else
            pEntry = pEntry->pNext;
	}
	
	// We didn't find any matched Mac address, our policy is treat it as 
	// broadcast packet and send to all.
	return IPMacTable.hash[IPMAC_TB_HASH_INDEX_OF_BCAST]->macAddr;
	
}

static NDIS_STATUS IPMacTable_RemoveAll(VOID)
{
	IPMacMappingEntry *pEntry;
	INT		i;

	if (!IPMacTable.valid)
		return TRUE;
	
	for (i=0; i<IPMAC_TB_HASH_ENTRY_NUM; i++)
	{
		while((pEntry = IPMacTable.hash[i]) != NULL)
		{
			IPMacTable.hash[i] = pEntry->pNext;
			MATDBEntryFree((PUCHAR)pEntry);
		}
	}

	IPMacTable.valid = FALSE;

	return TRUE;
}

static NDIS_STATUS IPMacTable_init(VOID)
{
	IPMacMappingEntry *pEntry = NULL;
	UCHAR BcastAddr[]= {0xff, 0xff, 0xff,0xff,0xff,0xff};
	
	NdisZeroMemory(&IPMacTable, sizeof(IPMacTable));
	
	//Set the last hash entry (hash[64]) as our default broadcast Mac address
	pEntry = (IPMacMappingEntry *)MATDBEntryAlloc(sizeof(IPMacMappingEntry));
	if (!pEntry)
		return FALSE;
	
	pEntry->ipAddr = 0;
	NdisMoveMemory(pEntry->macAddr, BcastAddr, 6);
	pEntry->pNext = NULL;
	IPMacTable.hash[IPMAC_TB_HASH_INDEX_OF_BCAST] = pEntry;
	
	IPMacTable.valid = TRUE;

	DBGPRINT(RT_DEBUG_TRACE, ("%s(): IPMacTable_init success!\n", __FUNCTION__));
	return TRUE;
}

static NDIS_STATUS MATProto_ARP_Exit(VOID)
{
	INT status;
		
	status = IPMacTable_RemoveAll();

	return status;
}

static PUCHAR MATProto_ARP_Rx(
	IN PRTMP_ADAPTER	pAd, 
	IN PNDIS_PACKET		pSkb,
	IN PUCHAR			pLayerHdr,
	IN UINT				infIdx)
{
	PUCHAR pArpHdr = NULL, pMacAddr = NULL;
	PUCHAR	tgtMac, tgtIP;
	BOOLEAN isUcastMac, isGoodIP;

	
	pArpHdr = pLayerHdr;

//dumpPkt(RTPKT_TO_OSPKT(pSkb)->data, RTPKT_TO_OSPKT(pSkb)->len);		
	// We just take care about the target(Mac/IP address) fields.
	tgtMac = pArpHdr + 18;
	tgtIP = tgtMac + 6;
		
	// isUcastMac = !(00:00:00:00:00:00|| mcastMac);
	isUcastMac = ((tgtMac[0]|tgtMac[1]|tgtMac[2]|tgtMac[3]|tgtMac[4]|tgtMac[5])!=0);
	isUcastMac &= ((tgtMac[0] && 0x1)==0);

	// isGoodIP = ip address is not 0.0.0.0
	isGoodIP = (*(UINT *)tgtIP != 0);

	DBGPRINT(RT_DEBUG_INFO, ("%s(): ARP Pkt=>tgtIP=%d.%d.%d.%d, tgtMac=%02x:%02x:%02x:%02x:%02x:%02x\n",
			__FUNCTION__, tgtIP[0], tgtIP[1], tgtIP[2], tgtIP[3],
			tgtMac[0],tgtMac[1],tgtMac[2],tgtMac[3],tgtMac[4],tgtMac[5]));
		
	if (isUcastMac && isGoodIP)
		pMacAddr= IPMacTableLookUp(pAd, *(UINT *)tgtIP);
		
	// For need replaced mac, we need to replace the targetMAC as correct one to make 
	// the real receiver can receive that.
	if (isUcastMac && pMacAddr)
		NdisMoveMemory(tgtMac, pMacAddr, MAC_ADDR_LEN);

	if (pMacAddr == NULL)
		pMacAddr = IPMacTable.hash[IPMAC_TB_HASH_INDEX_OF_BCAST]->macAddr;
	
	return pMacAddr;
}

static PUCHAR MATProto_ARP_Tx(
	IN PRTMP_ADAPTER 	pAd,
	IN PNDIS_PACKET		pSkb,
	IN PUCHAR 			pLayerHdr,
	IN UINT				infIdx)
{
	PUCHAR	pSMac, pSIP;
	BOOLEAN isUcastMac, isGoodIP;
	struct arphdr *arpHdr;
	PUCHAR pPktHdr;
	PNDIS_PACKET newSkb = NULL;

	pPktHdr = RTPKT_TO_OSPKT(pSkb)->data;
	
	arpHdr = (struct arphdr *)pLayerHdr;

	// Check the arp header.
	// We just handle ether type hardware address and IPv4 internet 
	// address type and opcode is  ARP reuqest/response.
	if ((arpHdr->ar_hrd != OS_HTONS(ARPHRD_ETHER)) || (arpHdr->ar_pro != OS_HTONS(ETH_P_IP)) ||
		(arpHdr->ar_op != OS_HTONS(ARPOP_REPLY) && arpHdr->ar_op != OS_HTONS(ARPOP_REQUEST)))
		return NULL;

	// We just take care about the sender(Mac/IP address) fields.
	pSMac =(PUCHAR)(pLayerHdr + 8);
	pSIP = (PUCHAR)(pSMac + MAC_ADDR_LEN);
	
	isUcastMac = IS_UCAST_MAC(pSMac);
	isGoodIP = IS_GOOD_IP(*(UINT *)pSIP);
	
/*	
	printk("%s(): ARP Pkt=>senderIP=%d.%d.%d.%d, senderMac=%02x:%02x:%02x:%02x:%02x:%02x\n",
			__FUNCTION__, pSIP[0], pSIP[1], pSIP[2], pSIP[3],
			pSMac[0],pSMac[1],pSMac[2],pSMac[3],pSMac[4],pSMac[5]);	
*/
	if (isUcastMac && isGoodIP)
		IPMacTableUpdate(pAd, pSMac, *(UINT *)pSIP);

	// For outgoing unicast mac, we need to replace the senderMAC as ourself to make 
	// the receiver can send to us.
	if (isUcastMac)
	{
		if(skb_cloned(RTPKT_TO_OSPKT(pSkb))) 
		{
			newSkb = (PNDIS_PACKET)skb_copy(RTPKT_TO_OSPKT(pSkb), GFP_ATOMIC);
			if(newSkb) {
				if (IS_VLAN_PACKET(RTPKT_TO_OSPKT(newSkb)->data))
					pSMac = (PUCHAR)(RTPKT_TO_OSPKT(newSkb)->data + MAT_VLAN_ETH_HDR_LEN + 8);
				else
					pSMac = (PUCHAR)(RTPKT_TO_OSPKT(newSkb)->data + MAT_ETHER_HDR_LEN + 8);
			}
		}
		
#ifdef CONFIG_AP_SUPPORT
#ifdef APCLI_SUPPORT
		NdisMoveMemory(pSMac, pAd->ApCfg.ApCliTab[infIdx].CurrentAddress, MAC_ADDR_LEN);
#endif // APCLI_SUPPORT //
#endif // CONFIG_AP_SUPPORT //

#ifdef CONFIG_STA_SUPPORT
#endif // CONFIG_STA_SUPPORT //
	}

	return (PUCHAR)newSkb;
}


static NDIS_STATUS MATProto_ARP_Init(VOID)
{
	BOOLEAN status = FALSE;

	status = (IPMacTable.valid==TRUE) ? TRUE : IPMacTable_init();

	if (status)
		MATProtoARPHandle.pMgtTb =&IPMacTable;
	else 
		MATProtoARPHandle.pMgtTb = NULL;

	return status;
}


static NDIS_STATUS MATProto_IP_Exit(VOID)
{
	INT status;
		
	status = IPMacTable_RemoveAll();

	return status;
}

static PUCHAR MATProto_IP_Rx(
	IN PRTMP_ADAPTER	pAd, 
	IN PNDIS_PACKET		pSkb,
	IN PUCHAR 			pLayerHdr,
	IN UINT				infIdx)
{
	PUCHAR pMacAddr;
	UINT   dstIP;
	
	// Fetch the IP addres from the packet header.
	getDstIPFromIpPkt(pLayerHdr, &dstIP);

	DBGPRINT(RT_DEBUG_INFO, ("%s():Receive a IP pkt=>dstIP=0x%x!\n", __FUNCTION__, dstIP));
	pMacAddr = IPMacTableLookUp(pAd, dstIP); 
	
	return pMacAddr;
}

static UCHAR  DHCP_MAGIC[]= {0x63, 0x82, 0x53, 0x63};
static PUCHAR MATProto_IP_Tx(
	IN PRTMP_ADAPTER 	pAd,
	IN PNDIS_PACKET		pSkb,
	IN PUCHAR 			pLayerHdr,
	IN UINT 			infIdx)
{
	PUCHAR pSrcMac;
	PUCHAR pSrcIP;
	BOOLEAN needUpdate;
	PUCHAR pPktHdr;

	pPktHdr = RTPKT_TO_OSPKT(pSkb)->data;
	
	pSrcMac = pPktHdr + 6;
	pSrcIP = pLayerHdr + 12;

	DBGPRINT(RT_DEBUG_INFO, ("%s(): Transmit a IP Pkt=>srcIP=0x%x\n", __FUNCTION__, *(UINT *)pSrcIP));

	needUpdate = NEED_UPDATE_IPMAC_TB(pSrcMac, (*(UINT *)pSrcIP));
	if (needUpdate)
		IPMacTableUpdate(pAd, pSrcMac, *(UINT *)pSrcIP);

	//For UDP packet, we need to check about the DHCP packet, to modify the flag of DHCP discovey/request as broadcast.
	if (*(pLayerHdr + 9) == 0x11)
	{
		PUCHAR udpHdr;
		UINT16 srcPort, dstPort;
		
		udpHdr = pLayerHdr + 20;
		srcPort = OS_NTOHS(*((UINT16 *)udpHdr));
		dstPort = OS_NTOHS(*((UINT16 *)(udpHdr+2)));
		
		if (srcPort==68 && dstPort==67) //It's a DHCP packet
		{
			PUCHAR bootpHdr;
			UINT16 bootpFlag;	
			
			bootpHdr = udpHdr + 8;
			bootpFlag = OS_NTOHS(*((UINT16 *)(bootpHdr+10)));
			DBGPRINT(RT_DEBUG_TRACE, ("is bootp packet! bootpFlag=0x%x\n", bootpFlag));
			if (bootpFlag != 0x8000) //check if it's a broadcast request.
			{
				PUCHAR dhcpHdr;
				
				dhcpHdr = bootpHdr + 236;
				
				DBGPRINT(RT_DEBUG_TRACE, ("the DHCP flag is a unicast, dhcp_magic=%02x:%02x:%02x:%02x\n", dhcpHdr[0], dhcpHdr[1], dhcpHdr[2], dhcpHdr[3]));
				if (NdisEqualMemory(dhcpHdr, DHCP_MAGIC, 4))
				{	
					DBGPRINT(RT_DEBUG_TRACE, ("dhcp magic macthed!\n"));	
					bootpFlag = OS_HTONS(0x8000);
					NdisMoveMemory((bootpHdr+10), &bootpFlag, 2);	//Set the bootp flag as broadcast
					NdisZeroMemory((udpHdr+6), 2); //modify the UDP chksum as zero
				}
			}	
		}
	}

	return NULL;
}

static NDIS_STATUS MATProto_IP_Init(VOID)
{
	BOOLEAN status = FALSE;

	status = (IPMacTable.valid==TRUE) ? TRUE : IPMacTable_init();

	if (status)
		MATProtoIPHandle.pMgtTb =&IPMacTable;
	else 
		MATProtoIPHandle.pMgtTb = NULL;
	
	return status;
}

void IPintToIPstr(int ipint, char Ipstr[20])
{
     int temp = 0;
     
     temp = ipint & 0x000FF;
     sprintf(Ipstr, "%d.", temp);
     temp = (ipint>>8) & 0x000FF;
     sprintf(Ipstr, "%s%d.", Ipstr, temp);
     temp = (ipint>>16) & 0x000FF;
     sprintf(Ipstr, "%s%d.", Ipstr, temp);
     temp = (ipint>>24) & 0x000FF;
     sprintf(Ipstr, "%s%d", Ipstr, temp);
}

VOID getIPMacTbInfo(char *pOutBuf)
{
	IPMacMappingEntry *pHead;
	int startIdx, endIdx;
    char Ipstr[20] = {0};

	if (!IPMacTable.valid)
	{
        DBGPRINT(RT_DEBUG_TRACE, ("%s():IPMacTable not init yet!\n", __FUNCTION__));
		return;
	}
		
	// dump all.
	startIdx = 0;
	endIdx = MAT_MAX_HASH_ENTRY_SUPPORT;

    sprintf(pOutBuf, "\n");
    sprintf(pOutBuf+strlen(pOutBuf), "%-18s%-20s\n", "IP", "MAC");
	for(; startIdx< endIdx; startIdx++)
	{
		pHead = IPMacTable.hash[startIdx];
    	while(pHead)
    	{
    	    if (strlen(pOutBuf) > (IW_PRIV_SIZE_MASK - 30))
                break;
            NdisZeroMemory(Ipstr, 20);
            IPintToIPstr(pHead->ipAddr, Ipstr);
			sprintf(pOutBuf+strlen(pOutBuf), "%-18s%02x:%02x:%02x:%02x:%02x:%02x\n",
				Ipstr, pHead->macAddr[0],pHead->macAddr[1],pHead->macAddr[2],
				pHead->macAddr[3],pHead->macAddr[4],pHead->macAddr[5]);
    		pHead = pHead->pNext;
    	}
	}	
}
#endif   // endif of "MAT_SUPPORT"

