#include <rt_config.h>

// Radar detection and carrier detection for 2880-SW-MCU
#ifdef CONFIG_AP_SUPPORT

#ifdef DFS_SUPPORT
void APRadarDetectionCB(PRTMP_ADAPTER pAd)
{
}
#endif // DFS_SUPPORT //

#ifdef CARRIER_DETECTION_SUPPORT
void APCarrierDetectionCB(PRTMP_ADAPTER pAd)
{
}
#endif // CARRIER_DETECTION_SUPPORT //

void McuRadarInit(PRTMP_ADAPTER pAd)
{
	pAd->ApCfg.McuRadarCmd = DETECTION_STOP;
	pAd->ApCfg.McuRadarState = FREE_FOR_TX;
	pAd->ApCfg.McuRadarEvent = 0;
	pAd->ApCfg.McuRadarTick = 0;
	pAd->ApCfg.McuRadarCTSPeriod = 0;
	pAd->ApCfg.McuRadarDfsActiveTime = 0;
	pAd->ApCfg.McuRadarCtsProtect = 0;
	pAd->ApCfg.McuRadarDetectCount = 0;
	memset(pAd->ApCfg.Detect1msResult, 0, sizeof(pAd->ApCfg.Detect1msResult));
}

#if defined(DFS_SUPPORT) || defined(CARRIER_DETECTION_SUPPORT)

void TimerCB(PRTMP_ADAPTER pAd)
{
	ULONG Value;

	pAd->ApCfg.McuRadarTick++;
	if (pAd->ApCfg.McuRadarTick >= pAd->ApCfg.McuRadarCTSPeriod)
	{
		// Start of period
		pAd->ApCfg.McuRadarTick = 0;
		pAd->ApCfg.McuRadarDetectCount = 0;
		pAd->ApCfg.McuRadarState = WAIT_CTS_BEING_SENT;

#if 0
		// to prevent Tx disabled due to CTS kick delay
		if (pAd->ApCfg.McuRadarEvent & RADAR_EVENT_CTS_SENT)
		{
			pAd->ApCfg.McuRadarEvent &= ~RADAR_EVENT_CTS_SENT;
			pAd->ApCfg.McuRadarDetectCount = 1;

			// Restore BBP
#ifdef DFS_DEBUG				
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R65, pAd->ApCfg._R65);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, pAd->ApCfg._R66);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R113, pAd->ApCfg._R113);
#else
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R65, 0x2c);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, 0x38);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R113, 0x01);
#endif
			// EnableNormalTx
			RTMP_IO_READ32(pAd, MAC_SYS_CTRL, &Value);
			Value |= 0x04;
			RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, Value);
		}
#endif // #if 0, Try remove

		pAd->ApCfg.McuRadarEvent = 0; // clear event

		if (pAd->ApCfg.McuRadarCtsProtect == 1)
		{
			// no CTS protect
			pAd->ApCfg.McuRadarState = DO_DETECTION;
			MCURadarDetect(pAd);
		}
		else if (pAd->ApCfg.McuRadarCtsProtect == 2 || pAd->ApCfg.McuRadarCtsProtect == 3)
		{
			
			// Carrier detection always send 1 CTS
			//if (pAd->ApCfg.McuRadarCtsProtect == 3 && pAd->ApCfg.McuRadarCmd == RADAR_DETECTION)
			if (pAd->ApCfg.McuRadarCtsProtect == 3)
			{
				RTMP_IO_READ32(pAd, 0x7784, &Value);
				Value &= 0xffff00ff;
				RTMP_IO_WRITE32(pAd, 0x7784, Value);
				
				// kick CTS once
				RTMP_IO_WRITE32(pAd, PBF_CTRL, 0x40);
				do
				{
					RTMP_IO_READ32(pAd, PBF_CTRL, &Value);
				}while(Value);
			}


			RTMP_IO_READ32(pAd, 0x7784, &Value);
			Value &= 0xffff00ff;
			Value |= 0x0000fe00;
			RTMP_IO_WRITE32(pAd, 0x7784, Value);
			// kick CTS twice
			if (pAd->ApCfg.McuRadarCmd == RADAR_DETECTION)
			{
				RTMP_IO_WRITE32(pAd, PBF_CTRL, 0x40);
			}
			else // CARRIER_DETECTION
			{
				RTMP_IO_WRITE32(pAd, PBF_CTRL, 0x80);
			}

			return;
		}
		else 
		{
			// stop radar/carrier detection
			return;
		}
	}

	if (pAd->ApCfg.McuRadarState == WAIT_CTS_BEING_SENT)
	{
		// check event or timeout
		if (pAd->ApCfg.McuRadarEvent & RADAR_EVENT_CTS_SENT)
		{
			pAd->ApCfg.McuRadarEvent &= ~RADAR_EVENT_CTS_SENT;
			pAd->ApCfg.McuRadarDetectCount = 1;
			pAd->ApCfg.McuRadarState = DO_DETECTION;
			// MCURadarDetect(pAd); // move to AsicSendCommandToMcu
			return;
		}
		
		if (pAd->ApCfg.McuRadarDetectCount >= (5-1)) // timeout 5ms, start from 0
		{
			pAd->ApCfg.McuRadarState = FREE_FOR_TX; // timeout 5ms, give up
		}
		else
		{
		}
		
		pAd->ApCfg.McuRadarDetectCount++;
	}
	else if (pAd->ApCfg.McuRadarState == DO_DETECTION)
	{
		if (pAd->ApCfg.McuRadarDetectCount >= (pAd->ApCfg.McuRadarDfsActiveTime - 1))
		{
			UCHAR BBPR115;

			//DetectResult = McuReadBBP(115)
			BBP_IO_READ8_BY_REG_ID(pAd, BBP_R115, &BBPR115);
			if (BBPR115 & 0x1)
			{
				POS_COOKIE pObj;
	
    			pObj = (POS_COOKIE) pAd->OS_Cookie;
#ifdef DFS_DEBUG
				printk("B115 = %x\n", BBPR115);
#endif
				// pulse radar detected
				if (pAd->ApCfg.McuRadarCmd == RADAR_DETECTION)
				{
					//RTMP_IO_WRITE32(pAd, 0x7024, 0x40);
					tasklet_hi_schedule(&pObj->pulse_radar_detect_task);
				}
				else // CARRIER_DETECTION
				{
					//RTMP_IO_WRITE32(pAd, 0x7024, 0x04);
					tasklet_hi_schedule(&pObj->carrier_sense_task);
				}
			}
			else if (BBPR115 & 0x2)
			{
				POS_COOKIE pObj;
	
    			pObj = (POS_COOKIE) pAd->OS_Cookie;
#ifdef DFS_DEBUG
				printk("B115 = %x\n", BBPR115);
#endif
				// width radar detected
				//RTMP_IO_WRITE32(pAd, 0x7024, 0x80);
				tasklet_hi_schedule(&pObj->width_radar_detect_task);
			}
			else
			{
			}
			pAd->ApCfg.McuRadarState = FREE_FOR_TX;
			// Restore BBP
#ifdef DFS_DEBUG				
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R65, pAd->ApCfg._R65);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, pAd->ApCfg._R66);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R113, pAd->ApCfg._R113);
#else
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R65, 0x2c);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, 0x38);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R113, 0x01);
#endif
#ifdef DFS_DEBUG
			if ((pAd->ApCfg.McuRadarDebug & 1) == 0)
				_RTMP_IO_WRITE32(0xa0300630, 0x2000);
#endif
			// EnableNormalTx
			RTMP_IO_READ32(pAd, MAC_SYS_CTRL, &Value);
			Value |= 0x04;
			RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, Value);

			

		}
		else
		{
		}
		pAd->ApCfg.McuRadarDetectCount++;
	}
	else // FREE_FOR_TX
	{
		if (pAd->ApCfg.McuRadarEvent & RADAR_EVENT_CTS_SENT)
		{
			pAd->ApCfg.McuRadarEvent &= ~RADAR_EVENT_CTS_SENT;
			pAd->ApCfg.McuRadarDetectCount = 1;

			// Restore BBP
#ifdef DFS_DEBUG				
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R65, pAd->ApCfg._R65);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, pAd->ApCfg._R66);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R113, pAd->ApCfg._R113);
#else
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R65, 0x2c);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, 0x38);
			BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R113, 0x01);
#endif


#ifdef DFS_DEBUG
			if ((pAd->ApCfg.McuRadarDebug & 1) == 0)
				_RTMP_IO_WRITE32(0xa0300630, 0x2000);
#endif
			// EnableNormalTx
			RTMP_IO_READ32(pAd, MAC_SYS_CTRL, &Value);
			Value |= 0x04;
			RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, Value);
		}
	}	
}

void MCURadarDetect(PRTMP_ADAPTER pAd)
{
	ULONG Value;
			//DisableNormalTx
			RTMP_IO_READ32(pAd, MAC_SYS_CTRL, &Value);
			Value &= (ULONG)~(0x04);
			RTMP_IO_WRITE32(pAd, MAC_SYS_CTRL, Value);
#ifdef DFS_DEBUG
			if ((pAd->ApCfg.McuRadarDebug & 1) == 0)
				if (pAd->ApCfg.McuRadarCtsProtect != 1)
					_RTMP_IO_WRITE32(0xa030062c, 0x2000);
			
			
#endif

			//change BBP now
			if (pAd->ApCfg.McuRadarCmd == CARRIER_DETECTION)
			{
				UCHAR BBPR4;
				
				BBP_IO_READ8_BY_REG_ID(pAd, BBP_R66, &BBPR4);
				if (BBPR4 & 0x10) // BW 40
				{
					if (pAd->CommonCfg.Channel > 14)
					{
						// BW 40, A band
						BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, 0x70);
					}
					else
					{
						// BW 40, G band
						BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, 0x60);
					}
					BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R69, 0x30);
					BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R70, 0x30);
					BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R73, 0x08);
				}
				else // BW 20
				{
					if (pAd->CommonCfg.Channel > 14)
					{
						// BW 20, A band
						BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, 0x50);
					}
					else
					{
						// BW 20, G band
						BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, 0x40);
					}
					BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R69, 0x28);
					BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R70, 0x28);
					BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R73, 0x08);
				}
				
			}
			else // RADAR_DETECTION
			{
#ifdef DFS_DEBUG				
				BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R65, pAd->ApCfg.R65);
				BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, pAd->ApCfg.R66);
				BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R113, pAd->ApCfg.R113);
#else
				BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R65, 0x1d);
				BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R66, 0x60);
				BBP_IO_WRITE8_BY_REG_ID(pAd, BBP_R113, 0x21);
#endif
			}		
			return;
}

#endif // #if defined(DFS_SUPPORT) || defined(CARRIER_DETECTION_SUPPORT)


#endif // CONFIG_AP_SUPPORT //


