/*
 *	Handle incoming frames
 *	Linux ethernet bridge
 *
 *	Authors:
 *	Lennert Buytenhek		<buytenh@gnu.org>
 *
 *	$Id: br_input.c,v 1.1.1.1 2006-10-13 02:42:51 steven Exp $
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 */

#include <linux/kernel.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/if_bridge.h>
#include <linux/netfilter_bridge.h>
#ifdef CONFIG_SOFTWARE_TURBO
#include <linux/hw_tcpip.h>
#endif
#include "br_private.h"
#ifdef CONFIG_BRIDGE_IGMP_SNOOP
#include <linux/ip.h>
#include <linux/in.h>
#include <linux/igmp.h>
#endif //CONFIG_BRIDGE_IGMP_SNOOP
#ifdef CONFIG_BRIDGE_PORT_ACCESS_HOST_LIMITED
	#include <linux/ip.h>
	#include <linux/in.h>
	#include <linux/tcp.h>
	#include <linux/udp.h>
	#include <linux/inetdevice.h>
	#include <linux/list.h>
	#include <linux/sysctl.h>
	#include <linux/module.h>
#endif	//CONFIG_BRIDGE_PORT_ACCESS_HOST_LIMITED
#ifdef CONFIG_BRIDGE_IGMPP_PROCFS
    #ifndef CONFIG_BRIDGE_PORT_ACCESS_HOST_LIMITED
		#include <linux/ip.h>
		#include <linux/in.h>
    #endif
    #include <linux/igmp.h>
#endif

unsigned char bridge_ula[6] = { 0x01, 0x80, 0xc2, 0x00, 0x00, 0x00 };

#ifdef CONFIG_BRIDGE_PORT_ACCESS_HOST_LIMITED
static int	bridge_port_access;
static char	bridge_port_access_if[32];					
static struct ctl_table_header	*bridge_port_access_header;
static struct ctl_table_header 	*bridge_port_access_if_header;

static ctl_table	bridge_port_access_entry[]= {
	{ 	.ctl_name		=	BRIDGE_PORT_ACCESS_ENTRY, 
		.procname		=	"br_port_access", 
		.mode			=	0644, 
		.data			=	&bridge_port_access, 
		.maxlen			=	sizeof(int), 
		.proc_handler	=	&proc_dointvec, 
	}, 
	{ 	.ctl_name	=	0
	}
};
static ctl_table	bridge_port_access_core[]= {
	{	.ctl_name	=	NET_CORE, 
		.mode		=	0555, 
		.child		=	bridge_port_access_entry, 
		.procname	=	"core", 
	}, 
	{	.ctl_name	=	0
	}
};
static ctl_table	bridge_port_access_net[]= {
	{	.ctl_name	=	CTL_NET, 
		.procname	=	"net", 
		.child		=	bridge_port_access_core, 
		.mode		=	0555, 
	}, 
	{	.ctl_name	=	0
	}
};


static ctl_table	bridge_port_access_if_entry[]= {
	{ 	.ctl_name		=	BRIDGE_PORT_ACCESS_IF_ENTRY, 
		.procname		=	"br_port_access_if", 
		.mode			=	0644, 
		.data			=	bridge_port_access_if, 
		.maxlen			=	sizeof(bridge_port_access_if), 
		.proc_handler	=	&proc_dostring, 
	}, 
	{ 	.ctl_name	=	0
	}
};
static ctl_table	bridge_port_access_if_core[]= {
	{	.ctl_name	=	NET_CORE, 
		.mode		=	0555, 
		.child		=	bridge_port_access_if_entry, 
		.procname	=	"core", 
	}, 
	{	.ctl_name	=	0
	}
};
static ctl_table	bridge_port_access_if_net[]= {
	{	.ctl_name	=	CTL_NET, 
		.procname	=	"net", 
		.child		=	bridge_port_access_if_core, 
		.mode		=	0555, 
	}, 
	{	.ctl_name	=	0
	}
};

#endif	//CONFIG_BRIDGE_PORT_ACCESS_HOST_LIMITED

#ifdef CONFIG_BRIDGE_MAT

#define MULTICAST   0xE0

static int Wlan_Rx(struct sk_buff *skb)
{
    unsigned int                    dip=0;
#if MAT_DBG
    unsigned int                    sip=0;
#endif /*MAT_DBG*/
    struct net_bridge               *br;
    struct net_bridge_port          *p;
    struct iphdr                    iphd, *iph;
    struct arphdr	                *arphd, arphdh;
    struct postarphdr               *postarphd, postarphdh;
    struct net_bridge_fdb_entry*    dst=NULL;
    unsigned char                   multicastMAC[6] = {0x01, 0x00, 0x5E, 0x00, 0x00, 0x00};

	/*--- Add your filter below ---*/

    p = skb->dev->br_port;
    br = p->br;
    if(ntohs(skb->mac.ethernet->h_proto)==ETH_P_IP){
        /*IP based packets*/
        iph = skb_header_pointer(skb, 0, sizeof(iphd), &iphd);
        memcpy(&dip, &(iph->daddr), sizeof(unsigned int));
#if MAT_DBG
        memcpy(&sip, &(iph->saddr), sizeof(unsigned int));
        printk("Wlan_Rx: ETH_P_IP from %d.%d.%d.%d ", sip&0x000000ff, sip>>8&0x000000ff, sip>>16&0x000000ff, sip>>24&0x000000ff);
        printk("to %d.%d.%d.%d.\n", dip&0x000000ff, dip>>8&0x000000ff, dip>>16&0x000000ff, dip>>24&0x000000ff);
#endif /*MAT_DBG*/
        if ((dip&0x000000f0)==MULTICAST){
            /*Multicast case*/
#if MAT_DBG
            printk("Wlan_Rx: Multicast case.\n");
#endif /*MAT_DBG*/
            multicastMAC[3] = dip>>8&0x000000ff;
            multicastMAC[4] = dip>>16&0x000000ff;
            multicastMAC[5] = dip>>24&0x000000ff;
            memcpy(skb->mac.ethernet->h_dest, multicastMAC, ETH_ALEN);
        }else{
            /*Not Multicast case*/
            dst = br_fdb_ip_get(br, dip);
            if(dst!=NULL){
#if MAT_DBG
                printk("dst != NULL. %.2x:%.2x:%.2x:%.2x:%.2x:%.2x.\n",
                dst->addr.addr[0], dst->addr.addr[1], dst->addr.addr[2],
                dst->addr.addr[3], dst->addr.addr[4], dst->addr.addr[5]);
#endif /*MAT_DBG*/
                memcpy(skb->mac.ethernet->h_dest, dst->addr.addr, ETH_ALEN);
            }else{
#if MAT_DBG
                printk("dst = NULL.\n");
#endif /*MAT_DBG*/
            }
        }
    }else if(ntohs(skb->mac.ethernet->h_proto)==ETH_P_ARP){
        /*ARP packets*/
        arphd = skb_header_pointer(skb, 0, sizeof(arphdh), &arphdh);
        postarphd = skb_header_pointer(skb, sizeof(arphdh), sizeof(postarphdh), &postarphdh);
        memcpy(&dip, postarphd->ar_tip, sizeof(unsigned int));
        if(ntohs(arphd->ar_op)!=ARPOP_REQUEST){
#if MAT_DBG
            printk("Wlan_Rx: arp(%d) %d.%d.%d.%d(%8x).\n", ntohs(arphd->ar_op), dip&0x000000ff, dip>>8&0x000000ff, dip>>16&0x000000ff, dip>>24&0x000000ff, dip);
#endif /*MAT_DBG*/
            dst = br_fdb_ip_get(br, dip);
            if(dst!=NULL){
#if MAT_DBG
                printk("dst != NULL. %.2x:%.2x:%.2x:%.2x:%.2x:%.2x.\n",
                dst->addr.addr[0], dst->addr.addr[1], dst->addr.addr[2],
                dst->addr.addr[3], dst->addr.addr[4], dst->addr.addr[5]);
#endif /*MAT_DBG*/
                memcpy(skb->mac.ethernet->h_dest, dst->addr.addr, ETH_ALEN);
                memcpy(postarphd->ar_tha, dst->addr.addr, ETH_ALEN);
            }else{
#if MAT_DBG
                printk("dst = NULL.\n");
#endif /*MAT_DBG*/
            }
        }else{
#if MAT_DBG
            printk("Wlan_Rx: arp(%d) req %d.%d.%d.%d(%8x).\n", ntohs(arphd->ar_op), dip&0x000000ff, dip>>8&0x000000ff, dip>>16&0x000000ff, dip>>24&0x000000ff, dip);
#endif /*MAT_DBG*/
        }
    }

	return 0;
}

static int Eth_Rx(struct sk_buff *skb)
{
    struct net_bridge_port *p;
    unsigned int    sip=0;
#if MAT_DBG
    unsigned int    dip=0;
#endif /*MAT_DBG*/
    struct iphdr    iphd, *iph;
    struct arphdr	    arphdh;
    struct postarphdr   *postarphd, postarphdh;

	/*--- Add your filter below ---*/
	
    p = skb->dev->br_port;
	if (p->state == BR_STATE_LEARNING || p->state == BR_STATE_FORWARDING){
        if(ntohs(skb->mac.ethernet->h_proto)==ETH_P_IP){
            /*IP based packets*/
            iph = skb_header_pointer(skb, 0, sizeof(iphd), &iphd);
            memcpy(&sip, &(iph->saddr), sizeof(unsigned int));
#if MAT_DBG
            memcpy(&dip, &(iph->daddr), sizeof(unsigned int));
            printk("Eth_Rx: ETH_P_IP from %d.%d.%d.%d ", sip&0x000000ff, sip>>8&0x000000ff, sip>>16&0x000000ff, sip>>24&0x000000ff);
            printk("to %d.%d.%d.%d\n", dip&0x000000ff, dip>>8&0x000000ff, dip>>16&0x000000ff, dip>>24&0x000000ff);
#endif /*MAT_DBG*/
            br_fdb_insert(p->br, p, skb->mac.ethernet->h_source, sip, 0);
        }else if(ntohs(skb->mac.ethernet->h_proto)==ETH_P_ARP){
            /*ARP packets*/
            postarphd = skb_header_pointer(skb, sizeof(arphdh), sizeof(postarphdh), &postarphdh);
            memcpy(&sip, postarphd->ar_sip, sizeof(unsigned int));
#if MAT_DBG
            memcpy(&dip, &(iph->daddr), sizeof(unsigned int));
            printk("Eth_Rx: ETH_P_ARP from %d.%d.%d.%d ", sip&0x000000ff, sip>>8&0x000000ff, sip>>16&0x000000ff, sip>>24&0x000000ff);
            printk("to %d.%d.%d.%d\n", dip&0x000000ff, dip>>8&0x000000ff, dip>>16&0x000000ff, dip>>24&0x000000ff);
#endif /*MAT_DBG*/
            br_fdb_insert(p->br, p, skb->mac.ethernet->h_source, sip, 0);
        }
    }

	return 0;
}

static int frame_Rx(struct sk_buff *skb)
{
	if(!strcmp(skb->dev->name, WLAN)){
		if(Wlan_Rx(skb)==1)
			return 1;
	}else if(!strcmp(skb->dev->name, ETHERNET)){
		if(Eth_Rx(skb)==1)
			return 1;
	}
	/*--- Add your filter below ---*/

	return 0;
}
#endif /*CONFIG_BRIDGE_MAT*/

static int br_pass_frame_up_finish(struct sk_buff *skb)
{
        if(skb->protocol==0x8100){
	   unsigned short protocol_id;
	   
	   skb_pull(skb,4);
	   memcpy(&protocol_id,skb->data,2);
	   skb->protocol = protocol_id;
	}
#ifdef CONFIG_SOFTWARE_TURBO
	if (hw_tcpip && hw_tcpip->sw_net_rx_action)
		if (hw_tcpip->sw_net_rx_action(skb)) {
			return 0;
		}
#endif
	netif_rx(skb);

	return 0;
}

static void br_pass_frame_up(struct net_bridge *br, struct sk_buff *skb)
{
	struct net_device *indev;

	br->statistics.rx_packets++;
	br->statistics.rx_bytes += skb->len;

	indev = skb->dev;
	skb->dev = &br->dev;
	skb->pkt_type = PACKET_HOST;
	skb_push(skb, ETH_HLEN);
	skb->protocol = eth_type_trans(skb, &br->dev);

	NF_HOOK(PF_BRIDGE, NF_BR_LOCAL_IN, skb, indev, NULL,
			br_pass_frame_up_finish);
}
#ifdef CONFIG_BRIDGE_IGMP_SNOOP
int search_group_ap_IP(struct port_igmp_ap_table_t *pt, uint32_t ip_addr)
{
     int i;
    for (i=0; i< 30; ++i){
        if (pt->group_list[i].used == 1 )
            if (pt->group_list[i].ip_addr == ip_addr)
				return i;
    }
    return (-1);
}

static int search_ap_list_MAC(struct port_igmp_ap_table_t *pt, int groupIndex, unsigned char* mac_addr)
{
    int i;
	for (i=0; i<HOSTLIST_NUMBER; ++i){
		if (pt->group_list[groupIndex].host_list[i].used==1)
			if (pt->group_list[groupIndex].host_list[i].mac_addr[0] == *mac_addr				)
			  if (pt->group_list[groupIndex].host_list[i].mac_addr[1] == *(mac_addr+1)			)
			    if (pt->group_list[groupIndex].host_list[i].mac_addr[2] == *(mac_addr+2)		)
			      if (pt->group_list[groupIndex].host_list[i].mac_addr[3] == *(mac_addr+3)		)
			        if (pt->group_list[groupIndex].host_list[i].mac_addr[4] == *(mac_addr+4)	)
			         if (pt->group_list[groupIndex].host_list[i].mac_addr[5] == *(mac_addr+5)	)
						return i;
	}
	return (-1);
}

static int get_ap_MAC(unsigned char * mac_addr, struct net_bridge *br, uint32_t ip_addr)
{
	struct br_mac_ap_table_t *tlist;
	int find = -1;
	read_lock(&br->br_mac_table_ap_rwlock);
	list_for_each_entry(tlist,&(br->br_mac_ap_table.list), list){
		if ( tlist->ip_addr == ip_addr){
			find = 0;
			int i;
			for (i =0; i<6; i++)
				mac_addr[i] = tlist->mac_addr[i];
			break;
		}
	}
	read_unlock(&br->br_mac_table_ap_rwlock);
	return find;
}

static int get_ap_element_from_MAC_pool(struct port_igmp_ap_table_t *pt ,int gIdx)
{
    int i;
    for(i=0; i<HOSTLIST_NUMBER; ++i){
        if (pt->group_list[gIdx].host_list[i].used == 0) return i;
    }
    return (-1);
}

static int ap_add_MAC_2_pool(struct port_igmp_ap_table_t *pt, int gIdx, int mIdx, unsigned char* mac_addr)
{
	pt->group_list[gIdx].host_list[mIdx].used =1;
	int i;
	for (i=0; i<6; i++)
		pt->group_list[gIdx].host_list[mIdx].mac_addr[i] = *(mac_addr+i);
	return 0;
}

static int check_ap_GROUP_pool(struct port_igmp_ap_table_t *pt)
{
	int i;
	for( i=0; i<30; ++i){
		if( pt->group_list[i].used == 0) return i;
	}
	return -1;
}

static int add_ap_GROUP(struct port_igmp_ap_table_t *pt, int gIdx, uint32_t ip_addr)
{
	pt->group_list[gIdx].used = 1;
	pt->group_list[gIdx].ip_addr = ip_addr;
	pt->group_list[gIdx].in_special_query = 0;
	pt->group_list[gIdx].port_aging_time = 125;
	pt->group_list[gIdx].max_response_time = 15;
	return 0;
}

static void add_to_port_igmp_ap_table_t(	struct net_bridge_port *p,	struct port_igmp_ap_table_t *pt,
						uint32_t ip32_addr, 	unsigned char * mac_addr)
{
	/* search group IP */
	int groupIdx,ipIdx;
	groupIdx = search_group_ap_IP(pt, ip32_addr);//check the group ip exist or not 
	if(groupIdx >= 0){
		/* search list MAC */
		ipIdx = search_ap_list_MAC(pt, groupIdx, mac_addr);//check the group's sta_mac
		
		if(ipIdx >= 0 ){ 			
			pt->group_list[groupIdx].port_aging_time = 125;
			pt->group_list[groupIdx].in_special_query = 0;
			pt->group_list[groupIdx].max_response_time = 15;
			printk("[BR_IGMP_AP_PROC]-> MAChas been existed !!\n");	

		}else{
			/* check MAC pool */
			int macPoolIdx;
			macPoolIdx = get_ap_element_from_MAC_pool(pt, groupIdx);
			if (macPoolIdx >=0 ){
				/* add MAC to pool */
				ap_add_MAC_2_pool(pt, groupIdx, macPoolIdx, mac_addr);				
				printk("[BR_IGMP_AP_PROC]-> MAC: %X:%X:%X:%X:%X:%X add !!\n",
						*mac_addr,*(mac_addr+1),*(mac_addr+2),*(mac_addr+3),*(mac_addr+4),*(mac_addr+5) );
				pt->group_list[groupIdx].port_aging_time = 125; 
				pt->group_list[groupIdx].in_special_query = 0;
				pt->group_list[groupIdx].max_response_time = 15;
			}
			else{
				printk("pool------error\n");
			}
		}			
	}else{ 
	
		/* check group pool */
		int groupPoolIdx;
		groupPoolIdx = check_ap_GROUP_pool(pt); 
		if(groupPoolIdx >= 0){
			add_ap_GROUP(pt, groupPoolIdx, ip32_addr);
			uint8_t *ip8_addr;
			trans_32to8(&ip32_addr, &ip8_addr);
			printk(KERN_INFO "[BR_IGMP_AP_PROC]-> Group IP: %u.%u.%u.%u add !!\n",
												*ip8_addr, *(ip8_addr+1), *(ip8_addr+2), *(ip8_addr+3));
			
			ap_add_MAC_2_pool(pt, groupPoolIdx, 0, mac_addr);
			printk("[BR_IGMP_AP_PROC]-> MAC: %X:%X:%X:%X:%X:%X add !!\n",
						*mac_addr,*(mac_addr+1),*(mac_addr+2),*(mac_addr+3),*(mac_addr+4),*(mac_addr+5) );
				
		}else {
				printk("group pool is full------input\n");
			
		}
	}

	return;
}
static void IGMP_TABLE_CHECK(struct net_bridge *br ,struct sk_buff *skb2, __u32 igmp_group_ip)
{
	
	uint32_t ip32 = (uint32_t) skb2->nh.iph->saddr;//sta's ip
	struct br_mac_ap_table_t *tlist;
	int find = 0;
	struct net_bridge_fdb_entry *hit_fdb_entry;
	unsigned char sta_mac_addr[6];
	uint32_t host_ip32_addr;	

	
//==================================================================================//	
	list_for_each_entry(tlist,&(br->br_mac_ap_table.list), list){
		if ( tlist->ip_addr == ip32){
			find =1;
			int i;
			for (i =0; i<6; i++)
				tlist->mac_addr[i] = skb2->mac.ethernet->h_source[i];
			break;
		}
	}
	if (find == 0 ){
		struct br_mac_ap_table_t * new_entry;
		new_entry = (struct br_mac_ap_table_t *)kmalloc(sizeof(struct br_mac_ap_table_t), GFP_ATOMIC);
		if (new_entry != NULL){
			int i;
			for (i =0; i<6; i++)
				new_entry->mac_addr[i] = skb2->mac.ethernet->h_source[i];
			new_entry->ip_addr = ip32;
			list_add(&(new_entry->list), &(br->br_mac_ap_table.list));
		}else{		
			printk("[BR_MAC_AP_PROC]-> alloc new br_mac_ap_table_t fail !!\n");			
		}
	}
//======================================================================================//	
	if(get_ap_MAC(sta_mac_addr,br,ip32)!=0) 
		{
		printk("get_ap_MAC------error\n");
		}
	hit_fdb_entry = br_fdb_get(br, sta_mac_addr); 
	if (hit_fdb_entry != NULL){
		if(atomic_read(&br->br_igmp_ap_table_enable) == 1){			
			if(atomic_read(&hit_fdb_entry->dst->ap_wireless_interface) == 1){
				
			write_lock(&hit_fdb_entry->dst->port_igmp_ap_table_rwlock);
			//send group ip and sta's mac to add table			
			add_to_port_igmp_ap_table_t(hit_fdb_entry->dst,&hit_fdb_entry->dst->port_igmp_ap_table,igmp_group_ip,sta_mac_addr);
			write_unlock(&hit_fdb_entry->dst->port_igmp_ap_table_rwlock);
			struct net_bridge_port *p;
				p = br->port_list;
				while( p != NULL){
				if(hit_fdb_entry->dst->port_no != p->port_no){
					write_lock(&p->port_igmp_ap_table_rwlock);
					del_to_port_igmp_ap_table_t(&p->port_igmp_ap_table, igmp_group_ip, sta_mac_addr);
					write_unlock(&p->port_igmp_ap_table_rwlock);
						}
				p=p->next;
					}
				}
			else{
				printk("wireless not enable-----------\n");
				br_fdb_put(hit_fdb_entry);
				return;
				}
			}
		br_fdb_put(hit_fdb_entry);
	}	
	else{
		printk("error-------------input\n");
		return;
		}
return;

}

static int remove_ap_MAC_from_pool(struct port_igmp_ap_table_t *pt, int gIdx, int mIdx)
{
	pt->group_list[gIdx].host_list[mIdx].used = 0;
	int i;
	for (i=0; i<6; i++)
		pt->group_list[gIdx].host_list[mIdx].mac_addr[i] = 0;
	return 0;
}

static int remove_ap_GROUP_from_pool(struct port_igmp_ap_table_t *pt, int gIdx)
{
	
	pt->group_list[gIdx].used = 0;
	pt->group_list[gIdx].ip_addr = 0;	
	pt->group_list[gIdx].max_response_time=0;
	pt->group_list[gIdx].port_aging_time=0;
	pt->group_list[gIdx].in_special_query=0;
	return 0;
}
static int check_ap_GROUP_is_empty_and_remove(struct port_igmp_ap_table_t *pt, int gIdx)
{
	int i,used=0;
	for (i=0; i<HOSTLIST_NUMBER; ++i){ 
		if (pt->group_list[gIdx].host_list[i].used == 1)
			++used;
	}
	if (used == 0){		
		remove_ap_GROUP_from_pool(pt, gIdx);	
		return 0;
	}
	else 
		return 0;
}
static void del_to_port_igmp_ap_table_t(struct port_igmp_ap_table_t *pt, uint32_t igmp_group_ip, unsigned char * sta_mac_addr)
{
	/* search group IP */
	int groupIdx,ipIdx;
	groupIdx = search_group_ap_IP(pt, igmp_group_ip);
	if(groupIdx >= 0){	
		/* search list MAC */
		ipIdx = search_ap_list_MAC(pt, groupIdx, sta_mac_addr);
		if(ipIdx >= 0){
			/* remove MAC and check group member*/
			remove_ap_MAC_from_pool(pt, groupIdx, ipIdx);
			printk("[BR_IGMP_AP_PROC]-> MAC: %X:%X:%X:%X:%X:%X remove !!\n",
						*sta_mac_addr,*(sta_mac_addr+1),*(sta_mac_addr+2),*(sta_mac_addr+3),*(sta_mac_addr+4),*(sta_mac_addr+5) );
			check_ap_GROUP_is_empty_and_remove(pt, groupIdx); 
		}else{			
			printk(KERN_INFO "[BR_IGMP_AP_PROC]-> MAC: %X:%X:%X:%X:%X:%X does't exist !!\n",
						*sta_mac_addr,*(sta_mac_addr+1),*(sta_mac_addr+2),*(sta_mac_addr+3),*(sta_mac_addr+4),*(sta_mac_addr+5) );		
		}
	}else{		
		uint8_t *ip8_addr;
		trans_32to8(&igmp_group_ip, &ip8_addr);
		printk(KERN_INFO "[BR_IGMP_AP_PROC]-> Group IP: %u.%u.%u.%u does't exist !!\n",
											*ip8_addr, *(ip8_addr+1), *(ip8_addr+2), *(ip8_addr+3));		
	}

	return;
}


static void IGMP_AP_TABLE_LEAVE(struct net_bridge *br ,struct sk_buff *skb2, __u32 igmp_group_ip)
{

		uint32_t ip32 = (uint32_t) skb2->nh.iph->saddr;//sta's ip.
		unsigned char sta_mac_addr[6];
		struct net_bridge_fdb_entry *hit_fdb_entry;
		if(get_ap_MAC(sta_mac_addr,br,ip32)!=0) 
		{
		printk("get_ap_MAC------error\n");
		}
		hit_fdb_entry = br_fdb_get(br, sta_mac_addr);//use sta's mac to find it in which port.
		if (hit_fdb_entry != NULL){
			if(atomic_read(&hit_fdb_entry->dst->ap_wireless_interface) == 1){
				write_lock(&hit_fdb_entry->dst->port_igmp_ap_table_rwlock);				
				del_to_port_igmp_ap_table_t(&hit_fdb_entry->dst->port_igmp_ap_table, igmp_group_ip, sta_mac_addr);
				write_unlock(&hit_fdb_entry->dst->port_igmp_ap_table_rwlock);
				}
			else{// wireless interface not enable in this port
					br_fdb_put(hit_fdb_entry);
					printk("in IGMP_AP_TABLE_LEAVE wireless not enable-------------- \n");
					return;
				}
			br_fdb_put(hit_fdb_entry);
			}
		else	{// not find the sta in any port
			printk("error IGMP_AP_TABLE_LEAVE hit_fdb_entry----------- \n");
			return;
			}
return;
}

void br_timer_check_table(unsigned long __data)
{
	struct net_bridge *br = (struct net_bridge *)__data;
	struct port_igmp_ap_table_t *pt;
	struct net_bridge_port *p;
	p = br->port_list;
	atomic_set(&br->br_igmp_AP_query_enable, 0);
	if(br->igmp_switch_query_timer!=0){		
	br->igmp_switch_query_timer = br->igmp_switch_query_timer - 5;
		}
	else{
		while( p != NULL){		
			write_lock(&p->port_igmp_ap_table_rwlock);		
			if(p->port_igmp_ap_table.enable == 1){
		 	 int i;
  		  	for (i=0; i< 30; ++i){
  		  	if(p->port_igmp_ap_table.group_list[i].used == 1){
		  		//aging >0 and flag = 0		  	
				if ((p->port_igmp_ap_table.group_list[i].port_aging_time > 0)&&(p->port_igmp_ap_table.group_list[i].in_special_query == 0) ){
					p->port_igmp_ap_table.group_list[i].port_aging_time = p->port_igmp_ap_table.group_list[i].port_aging_time - 5;
					}
				//aging =0 so set flag =0 and reponse time = 10
		  		else if ((p->port_igmp_ap_table.group_list[i].port_aging_time == 0)&&(p->port_igmp_ap_table.group_list[i].in_special_query == 0) ){        		
         	       p->port_igmp_ap_table.group_list[i].in_special_query = 1;
					p->port_igmp_ap_table.group_list[i].max_response_time = 15;
					atomic_set(&br->br_igmp_AP_query_enable, 1); // send igmp query
				}
				//aging =0 and flag =1 and response >0 
				else if ((p->port_igmp_ap_table.group_list[i].port_aging_time == 0)&&(p->port_igmp_ap_table.group_list[i].in_special_query == 1)&&(p->port_igmp_ap_table.group_list[i].max_response_time>0) ){
					p->port_igmp_ap_table.group_list[i].max_response_time = p->port_igmp_ap_table.group_list[i].max_response_time -5;
				
				}
				//aging time ==0 flag ==1 response time ==0
				else if ((p->port_igmp_ap_table.group_list[i].port_aging_time == 0)&&(p->port_igmp_ap_table.group_list[i].in_special_query == 1)&&(p->port_igmp_ap_table.group_list[i].max_response_time<=0) ){
					int j,k; 
					for(j=0;j<8;j++){
					   p->port_igmp_ap_table.group_list[i].host_list[j].used = 0;
					for(k=0;k<6;k++){
						p->port_igmp_ap_table.group_list[i].host_list[j].mac_addr[k] = 0;
						}
					}			
					remove_ap_GROUP_from_pool(&br->port_list->port_igmp_ap_table,i);//remove group 
				}
  			  }	
			}
		  }
			write_unlock(&p->port_igmp_ap_table_rwlock);
			p=p->next;
    	 }
 	 }		
mod_timer(& br->igmptimer, jiffies + 5*HZ);	
}

void br_igmp_timer_enable(struct net_bridge *br)
{
	struct net_bridge_port *p;
	struct timer_list *timer = &br->igmptimer;	
	init_timer(timer);
	timer->data = (unsigned long) br;
	timer->function = br_timer_check_table;	
	timer->expires = jiffies + 5*HZ;
	add_timer(timer);	
}
void br_igmp_timer_disable(struct net_bridge *br)
{
	del_timer(&br->igmptimer);
}
#endif //CONFIG_BRIDGE_IGMP_SNOOP

#ifdef CONFIG_BRIDGE_IGMPP_PROCFS
/* snoop_MAC() => If IP address that existed in br_mac_table ,replace it,
 * else create a new list entry and add it to list.
 */
static void snoop_MAC(struct net_bridge *br ,struct sk_buff *skb2)
{
	uint32_t ip32 = (uint32_t) skb2->nh.iph->saddr;

	struct br_mac_table_t *tlist;
	int find = 0;;
	list_for_each_entry(tlist,&(br->br_mac_table.list), list){
		if ( tlist->ip_addr == ip32){
			find =1;
			int i;
			for (i =0; i<6; i++)
				tlist->mac_addr[i] = skb2->mac.ethernet->h_source[i];
			break;
		}
	}
	if (find == 0 ){
		struct br_mac_table_t * new_entry;
		new_entry = (struct br_mac_table_t *)kmalloc(sizeof(struct br_mac_table_t), GFP_ATOMIC);
		if (new_entry != NULL){
			int i;
			for (i =0; i<6; i++)
				new_entry->mac_addr[i] = skb2->mac.ethernet->h_source[i];
			new_entry->ip_addr = ip32;
			list_add(&(new_entry->list), &(br->br_mac_table.list));
		}else{
			#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
			printk(KERN_INFO "[BR_MAC_PROC]-> alloc new br_mac_table_t fail !!\n");
			#endif
		}
	}
}
#endif 

static int br_handle_frame_finish(struct sk_buff *skb)
{
	struct net_bridge *br;
	unsigned char *dest;
	struct net_bridge_fdb_entry *dst;
	struct net_bridge_port *p;
	int passedup;

	dest = skb->mac.ethernet->h_dest;

	p = skb->dev->br_port;
	if (p == NULL)
		goto err_nolock;

	br = p->br;
	read_lock(&br->lock);
	if (skb->dev->br_port == NULL)
		goto err;

	passedup = 0;
	if (br->dev.flags & IFF_PROMISC) {
		struct sk_buff *skb2;

		skb2 = skb_clone(skb, GFP_ATOMIC);
		if (skb2 != NULL) {
			passedup = 1;
			br_pass_frame_up(br, skb2);
		}
	}

	if (dest[0] & 1) {
	#ifdef CONFIG_BRIDGE_IGMPP_PROCFS
		read_lock(&br->br_mac_table_rwlock); //reader spin lock
		if (atomic_read(&br->br_mac_table_enable) == 1 )
			if(skb->nh.iph->protocol == IPPROTO_IGMP){ // IGMP protocol number: 0x02
				struct sk_buff *skb2;
				if ((skb2 = skb_clone(skb, GFP_ATOMIC)) != NULL) {
					skb_pull(skb2, skb2->nh.iph->ihl<<2);
					struct igmphdr *ih = (struct igmphdr *) skb2->data;
					if (ih->type == IGMP_HOST_MEMBERSHIP_REPORT ||		// IGMPv1 REPORT
						ih->type == IGMPV2_HOST_MEMBERSHIP_REPORT ||	// IGMPv2 REPORT
						ih->type == IGMPV3_HOST_MEMBERSHIP_REPORT	)	// IGMPv3 REPORT
					{
						snoop_MAC(br, skb2);
					}
					kfree_skb(skb2);
				}else{
					#ifdef CONFIG_BRIDGE_IGMPP_PROCFS_DEBUG
					printk(KERN_INFO "[BR_MAC_PROC]-> alloc new sk_buff fail !!\n");
					#endif
				}
			}
		read_unlock(&br->br_mac_table_rwlock); //reader spin lock
#endif
#ifdef CONFIG_BRIDGE_IGMP_SNOOP	
		if(br->igmp_enable==1){				
		read_lock(&br->br_mac_table_ap_rwlock); //reader spin lock					
			if(skb->nh.iph->protocol == IPPROTO_IGMP){ // IGMP protocol number: 0x02
				struct sk_buff *skb2;
				
				if ((skb2 = skb_copy(skb, GFP_ATOMIC)) != NULL) {
					skb_pull(skb2, skb2->nh.iph->ihl<<2);
					struct igmphdr *ih = (struct igmphdr *) skb2->data;
					if(ih->type == IGMP_HOST_MEMBERSHIP_REPORT || ih->type == IGMPV2_HOST_MEMBERSHIP_REPORT){						
						if(ih->group!=0xfaffffef)//239.255.255.250 SSDP, not care
							{							
						IGMP_TABLE_CHECK(br,skb2,ih->group);
							}
						}
					if(ih->type == IGMP_HOST_LEAVE_MESSAGE)
						IGMP_AP_TABLE_LEAVE(br,skb2,ih->group); //need group ip and sta 'mac and port
						
					if(ih->type == IGMPV3_HOST_MEMBERSHIP_REPORT){
						struct igmpv3_report *v3report = (struct igmpv3_report *) skb2->data;
							if(ntohs(v3report->ngrec)!=1)
								printk("v3report->ngrec-----error\n");
						
							else{
								if(v3report->grec[0].grec_type==IGMPV3_CHANGE_TO_EXCLUDE){
									if(v3report->grec[0].grec_mca!=0xfaffffef)//239.255.255.250 SSDP, not care
										IGMP_TABLE_CHECK(br,skb2,v3report->grec[0].grec_mca);
									}
								if(v3report->grec[0].grec_type==IGMPV3_CHANGE_TO_INCLUDE){
									if(v3report->grec[0].grec_mca!=0xfaffffef)//239.255.255.250 SSDP, not care
										IGMP_AP_TABLE_LEAVE(br,skb2,v3report->grec[0].grec_mca);
									}
								}
							}	
					if(ih->type == IGMP_HOST_MEMBERSHIP_QUERY){						
						 	br->igmp_switch_query_timer=255;
					
						}
						
					kfree_skb(skb2);				
				}else{				
					printk(KERN_INFO "[BR_MAC_AP_PROC]-> alloc new sk_buff fail !!\n");				
				}
			}
			read_unlock(&br->br_mac_table_ap_rwlock); //reader spin lock
		} 
#endif //CONFIG_BRIDGE_IGMP_SNOOP		

#ifdef CONFIG_BRIDGE_MAT
    if (br->ap_mode == APC){
        struct sk_buff *skb2;
        if ((skb2 = skb_copy(skb, GFP_ATOMIC)) == NULL) {
            br->statistics.tx_dropped++;
            kfree_skb(skb); /*Root cause of memery leak.*/
            return;
        }
        br_flood_forward(br, skb, !passedup);
		if (!passedup) {
			passedup = 1;
			kfree_skb(skb); /*Root cause of memery leak.*/
			br_pass_frame_up(br, skb2);
		}
	}else{
	    br_flood_forward(br, skb, !passedup);
		if (!passedup)
			br_pass_frame_up(br, skb);
	}
#else
		br_flood_forward(br, skb, !passedup);
		if (!passedup)
			br_pass_frame_up(br, skb);
#endif /*CONFIG_BRIDGE_MAT*/

		goto out;
	}

	dst = br_fdb_get(br, dest);
	if (dst != NULL && dst->is_local) {
		if (!passedup)
			br_pass_frame_up(br, skb);
		else
			kfree_skb(skb);
		br_fdb_put(dst);
		goto out;
	}

	if (dst != NULL) {
#ifdef CONFIG_BRIDGE_FORWARD_CTRL
		if (!atomic_read(&br->br_forward))
		{
			kfree_skb(skb);
			br_fdb_put(dst);
			goto out;
		}
#endif
#ifdef CONFIG_BRIDGE_PORT_FORWARD
		if (!p->port_forwarding) {
	        kfree_skb(skb);
			br_fdb_put(dst);
			goto out;
		}
#endif
#ifdef CONFIG_BRIDGE_REJECT_FORWARD_LIST
		if((p->rejected_port_list) && 
			(dst->dst->port_no <= (sizeof(p->rejected_port_list) << 3) - 1) &&
			(p->rejected_port_list >> dst->dst->port_no) & 0x01) {
			kfree_skb(skb);
			br_fdb_put(dst);
			goto out;
		}
#endif
		br_forward(dst->dst, skb);
		br_fdb_put(dst);
		goto out;
	}

	br_flood_forward(br, skb, 0);

out:
	read_unlock(&br->lock);
	return 0;

err:
	read_unlock(&br->lock);
err_nolock:
	kfree_skb(skb);
	return 0;
}

void br_handle_frame(struct sk_buff *skb)
{
	struct net_bridge *br;
	unsigned char *dest;
	struct net_bridge_port *p;

	dest = skb->mac.ethernet->h_dest;

	p = skb->dev->br_port;
	if (p == NULL)
		goto err_nolock;

	br = p->br;
	read_lock(&br->lock);
	if (skb->dev->br_port == NULL)
		goto err;

	if (!(br->dev.flags & IFF_UP) ||
	    p->state == BR_STATE_DISABLED)
		goto err;

	if (skb->mac.ethernet->h_source[0] & 1)
		goto err;

#ifdef CONFIG_BRIDGE_MAT
    /*Somebody take care the skb*/
if (br->ap_mode == APC){
	if(frame_Rx(skb)==1)
		goto err;
}else{
	if (p->state == BR_STATE_LEARNING ||
	    p->state == BR_STATE_FORWARDING)
		br_fdb_insert(br, p, skb->mac.ethernet->h_source, 0, 0);
}
#else
	if (p->state == BR_STATE_LEARNING ||
	    p->state == BR_STATE_FORWARDING)
		br_fdb_insert(br, p, skb->mac.ethernet->h_source, 0);
#endif /*CONFIG_BRIDGE_MAT*/

	if (br->stp_enabled &&
	    !memcmp(dest, bridge_ula, 5) &&
	    !(dest[5] & 0xF0))
		goto handle_special_frame;

	if (p->state == BR_STATE_FORWARDING) {
	#ifdef	CONFIG_BRIDGE_PORT_ACCESS_HOST_LIMITED
		if(bridge_port_access && *bridge_port_access_if) {
			uint8_t			tmp_protocol;	/* packet's protocol */
			int				ip_len;			/* length of ip-header */
			unsigned char	*tmp_mac_raw;	/*  */
			
			tmp_mac_raw=(unsigned char *)skb->mac.raw;
			tmp_mac_raw+=12;
			/* Blocking LLD2 packets from ra1 interface */
			if(*tmp_mac_raw==0x88 && *(tmp_mac_raw+1)==0xd9 
				&& !strcmp(p->dev->name, bridge_port_access_if)) {
			/*	printk(KERN_WARNING "DROP LLD2 packets!\n");*/
				goto	err;
			}
			ip_len=4*(skb->nh.iph->ihl);
			tmp_protocol=skb->nh.iph->protocol;
			
			if(tmp_protocol==IPPROTO_UDP) {
				struct udphdr	*udp;		/*  */
				uint16_t		tmp_dport;	/*  */
				
				udp=(struct udphdr *)((void *)skb->nh.iph+ip_len);
				tmp_dport=ntohs(udp->dest);
				
				if(!strcmp(p->dev->name, bridge_port_access_if)) {
					/* Blocking UPNP packets from ra1 interface */
					if(tmp_dport==1900 || tmp_dport==49152 || 
						tmp_dport==49153 || tmp_dport==49154) {
				/*		printk(KERN_WARNING "DROP SSDP/UPNP packets!\n");*/
						goto	err;
					}
				}
			}
		}

		if (!memcmp(p->br->dev.dev_addr, dest,ETH_ALEN)) {
			if(bridge_port_access) {
				struct net_bridge	*br;
				struct in_device	*in__;
				uint32_t			local;			/* RG's LAN ip */
				uint32_t			tmp_dip;		/* packet's dest ip */
				uint8_t				tmp_protocol;	/* packet's protocol */
				int					ip_len;			/* length of ip-header */
				
				br=p->br;
				in__=(struct in_device *)(br->dev.ip_ptr);
				ip_len=4*(skb->nh.iph->ihl);
				tmp_protocol=skb->nh.iph->protocol;
				tmp_dip=ntohl(skb->nh.iph->daddr);
				
				if(in__ && in__->ifa_list) {
					/* Find out the br0's ip */
					local=htonl(in__->ifa_list->ifa_local);
					
					/* Examining the packet.
					 * Droping packets from ra1 to RG's web pages.
					 */
					if(tmp_protocol==IPPROTO_TCP) {
						struct tcphdr	*tcp;		/*  */
						uint16_t		tmp_dport;	/*  */
						
						/*
						 * FIXME: ntohs(skb->h.th->dest) cannot get the correct value.
						 */
						tcp=(struct tcphdr *)((void *)skb->nh.iph+ip_len);
						tmp_dport=ntohs(tcp->dest);
						
						if(!strcmp(p->dev->name, bridge_port_access_if) && local==tmp_dip && tmp_dport==80) {
						/*	printk(KERN_WARNING "DROP packets to RG's web pages!\n");*/
							goto	err;
						}
					}
				}
			}
		}
#endif //CONFIG_BRIDGE_PORT_ACCESS_HOST_LIMITED

		NF_HOOK(PF_BRIDGE, NF_BR_PRE_ROUTING, skb, skb->dev, NULL,
			br_handle_frame_finish);
		read_unlock(&br->lock);
		return;
	}

err:
	read_unlock(&br->lock);
err_nolock:
	kfree_skb(skb);
	return;

handle_special_frame:
	if (!dest[5]) {
		NF_HOOK(PF_BRIDGE, NF_BR_LOCAL_IN, skb, skb->dev,NULL,
			br_stp_handle_bpdu);
		read_unlock(&br->lock);
		return;
	}

	read_unlock(&br->lock);
	kfree_skb(skb);
}
#ifdef CONFIG_BRIDGE_PORT_ACCESS_HOST_LIMITED
static int	__init	init_bridge_port_access(void)
{
	bridge_port_access_header=register_sysctl_table(bridge_port_access_net, 0);
	bridge_port_access_if_header=register_sysctl_table(bridge_port_access_if_net, 0);
	if(bridge_port_access_header==NULL || bridge_port_access_if_header==NULL) {
		printk(KERN_WARNING "Register Bridge port access fails!\n");
		return -1;
	}
	printk("Register Bridge port access successes!\n");
	return	0;
}

static void	__exit	fini_bridge_port_access(void)
{
	unregister_sysctl_table(bridge_port_access_header);
	unregister_sysctl_table(bridge_port_access_if_header);
}

module_init(init_bridge_port_access);
module_exit(fini_bridge_port_access);
#endif	//CONFIG_BRIDGE_PORT_ACCESS_HOST_LIMITED
