
#include <linux/config.h>
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/timer.h>
#include <linux/netfilter.h>
#include <net/protocol.h>
#include <net/checksum.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv4/ip_nat_rule.h>

//#include <net/route.h>
//#include <linux/netfilter_ipv4/ip_conntrack.h>
//#include <linux/module.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_STUN.h>

/* Called when user tries to insert an entry of this type: 
 * hook_mask is a bitmask of hooks from which it can be called.
 *
 * Return true or false.
 */
int checkentry(const char *tablename,
			const struct ipt_entry *e,
			void *targinfo,
			unsigned int targinfosize,
			unsigned int hook_mask)
{
	//const struct ip_nat_multi_range	*mr=targinfo;
	
	/* checking */
	if(strcmp(tablename, "nat")!=0)
		return	0;
	/*else if(targinfosize)
		return	0;
	else if(hook_mask)
		return	0;*/
	
	return	1;
}

/* Returns verdict. */
unsigned int target(struct sk_buff **pskb,
					unsigned int hooknum,
					const struct net_device *in,
					const struct net_device *out,
					const void *targinfo,
					void *userdata)
{
	struct rt_key	key;
	struct rtable	*rt;
	STUN		stun_state=__NONE;		/* record the state of STUN */
	u_int32_t	newsrc;
	struct ip_conntrack		*ct;
	enum ip_conntrack_info	ctinfo;
	struct ip_nat_multi_range	newrange;
	const struct ip_nat_multi_range		*mr;
	
	IP_NF_ASSERT(hooknum == NF_IP_POST_ROUTING);
	/* FIXME: For the moment, don't do local packets, breaks testsuite for 2.3.49 --RR */
	if((*pskb)->sk)
		return NF_ACCEPT;
		
	ct=ip_conntrack_get(*pskb, &ctinfo);
	mr=targinfo;
	
	/* Setting STUN type */
	if(mr->range[0].flags&IP_NAT_FULL_CONE) {
		stun_state=__FULL_CONE;
	} else if(mr->range[0].flags&IP_NAT_RESTRICTED_CONE) {
		stun_state=__RESTRICTED_CONE;
	} else if(mr->range[0].flags&IP_NAT_PORT_RESTRICTED_CONE) {
		stun_state=__PORT_RESTRICTED_CONE;
	/*} else if(mr->range[0].flags&IP_NAT_SYMMETRIC) {
		stun_state=__SYMMETRIC_NAT;*/
	} else {
		/* not possible */
		stun_state=__ERROR;
		printk(KERN_ERR "Error status happens in STUN!\n");
	}
	key.dst=(*pskb)->nh.iph->daddr;
	key.src=0;
	key.oif=0;
	key.tos=RT_TOS((*pskb)->nh.iph->tos)|RTO_CONN;
#ifdef CONFIG_IP_ROUTE_FWMARK
	key.fwmark = (*pskb)->nfmark;
#endif
	if(ip_route_output_key(&rt, &key)!=0) {
		/* Funky routing can do this. */
		if (net_ratelimit())
			printk("STUN: No route: Rusty's brain broke!\n");
		return NF_DROP;
	}
	if (rt->u.dst.dev!=out) {
		if (net_ratelimit())
			printk("STUN: Route sent us somewhere else.\n");
		return NF_DROP;
	}	
	newsrc=rt->rt_src;
	ip_rt_put(rt);
	
	/* No CONFIG_IP_NF_TARGET_MASQUERADE_MODULE, comment this! */
	/*WRITE_LOCK(&stun_rwlock);
	ct->nat.masq_index=out->ifindex;
	WRITE_UNLOCK(&stun_rwlock);*/
	newrange=(struct ip_nat_multi_range){1, {{mr->range[0].flags|IP_NAT_RANGE_MAP_IPS
		, newsrc, newsrc, mr->range[0].min, mr->range[0].max}}};
	
	/* Recognize IPPROTO_TCP or IPPROTO_UDP packets */
	if(stun_helper) {
		if(ct->tuplehash[0].tuple.dst.protonum==IPPROTO_TCP||
			ct->tuplehash[0].tuple.dst.protonum==IPPROTO_UDP) {
			return	ip_nat_setup_info_for_stun(ct, &newrange, hooknum, stun_state);
		} 
	}
	return	ip_nat_setup_info(ct, &newrange, hooknum);
}

/* Called when entry of this type deleted. */
void destroy(void *targinfo, 
			unsigned int targinfosize)
{
}

static struct ipt_target	target_stun= {
	.list		= {NULL, NULL},
	.name		= "STUN",
	.target		= &target,
	.checkentry	= &checkentry,
	.destroy	= NULL,
	.me			= THIS_MODULE
};

static int __init init(void)
{
	/*register_inetaddr_notifier()*/
	return ipt_register_target(&target_stun);
}

static void __exit fini(void)
{
	/*unregister_inetaddr_notifier();*/
	ipt_unregister_target(&target_stun);
}

module_init(init);
module_exit(fini);
MODULE_LICENSE("GPL");
