/*
 * Copyright (C) 2000 Lennert Buytenhek
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>      
#include <sys/time.h>
#include <asm/param.h>
#include "libbridge.h"
#include "brctl.h"

char *help_message =
"commands:\n"
"\taddbr\t\t<bridge>\t\tadd bridge\n"
"\taddif\t\t<bridge> <device>\tadd interface to bridge\n"
"\tdelbr\t\t<bridge>\t\tdelete bridge\n"
"\tdelif\t\t<bridge> <device>\tdelete interface from bridge\n"
"\tshow\t\t\t\t\tshow a list of bridges\n"
"\tshowmacs\t<bridge>\t\tshow a list of mac addrs\n"
"\tshowstp\t\t<bridge>\t\tshow bridge stp info\n"
"\n"
"\tsetageing\t<bridge> <time>\t\tset ageing time\n"
"\tsetbridgeprio\t<bridge> <prio>\t\tset bridge priority\n"
"\tsetfd\t\t<bridge> <time>\t\tset bridge forward delay\n"
"\tsetgcint\t<bridge> <time>\t\tset garbage collection interval\n"
"\tsethello\t<bridge> <time>\t\tset hello time\n"
"\tsetmaxage\t<bridge> <time>\t\tset max message age\n"
"\tsetpathcost\t<bridge> <port> <cost>\tset path cost\n"
"\tsetportprio\t<bridge> <port> <prio>\tset port priority\n"
"\tstp\t\t<bridge> <state>\tturn stp on/off\n"
#ifdef MULTICAST_BWCTRL
"\tsetbwctrl\t<bridge> <port> <bdwh>\tset multicast bandwidth (kbps)\n"
#endif
#ifdef FORWARD_CTRL
"\tsetfdctrl\t<bridge> <port> <state>\tturn port forwarding on/off\n"
#endif 
#ifdef REJECT_FORWARD_LIST
"\taddrejfwlist\t<bridge> <port1> <port2>add port2 to deny list of port1\n"
"\tdelrejfwlist\t<bridge> <port1> <port2>del port2 from deny list of port1\n"
"\tshwrejfwlist\t<bridge> <port>\t\tshow deny list of port\n"
#endif
#if AP_OPERATION_MODE
"\tapmode\t\t<bridge> <state>\tpass AP operation mode to bridge\n"
#endif
/* Jack add 13/04/07 MULTI_VLAN +++ */
//"\tshowprimaryvid\t\t<bridge> \t show_primary_vid \n"
"\tsetprimaryvid\t\t<bridge> <vid>\tset primary vlan ID\n"
/* Jack add 13/04/07 MULTI_VLAN --- */
/* AP Client function 2007-01-21 start */
#ifdef APC_OPERATION_MODE
"\tapc\t\t<bridge> <state>\tturn AP Clint mode on/off\n"
#endif
/*dhcp server no provide to eth0's pc 2008-01-23 dennis start */
#ifdef DHCP_SERVER_ENABLE
"\tdhcp_server_enable\t\t<bridge> <state>\tturn dhcp_server on/off\n"
#endif
/*dhcp server no provide to eth0's pc 2008-01-23 dennis end */
/* AP Client function 2007-01-21 end */
/* IGMP Snooping dennis 2008-01-29 start */
#ifdef IGMP_SNOOP
"\tigmp_snooping\t\t<bridge> <state>\tturn igmp snooping on/off\n"
#endif
/* IGMP Snooping dennis 2008-01-29 end */
/* e_partition 2008-02-1 start */
#ifdef E_PARTITION
"\te_partition\t\t<bridge> <state>\tturn e_partition on/off\n"
#endif
/* e_partition 2008-02-1 end */
/* ethernet integration dennis2008-02-05 start */
#ifdef ETHLINK
"\tethlink\t\t<bridge> <state>\tturn ethlink on/off\n"
#endif
/* ethernet integration dennis2008-02-05 end */
/*+++ Limited Administration, Builder, 2008/03/11 +++*/
#ifdef LIMITED_ADMIN
/*limited administration type*/
"\tladtype\t\t<bridge> <type>\t\tlimited administration type/off\n"
"\t\t\t\t\t\t0: Limited Adminstrator Disabled\n"
"\t\t\t\t\t\t1: Administrate with VID\n"
"\t\t\t\t\t\t2: Administrate with Limited IP\n"
"\t\t\t\t\t\t3: Both\n"
/*limited administration vlan ID*/
"\tladvid\t\t<bridge> <vlan id>\tadministrator's vlan ID\n"
/*limited administration IP pool*/
"\tladippool\t<bridge> <pool idx> <start ip> <end ip>\n"
#endif /*LIMITED_ADMIN*/
/*+++ Limited Administration, Builder, 2008/03/11 +++*/
;

void help()
{
	fprintf(stderr, help_message);
}

#ifdef RGBIN_BOX
int brctl_main(int argc, char *argv[])
#else
int main(int argc, char *argv[])
#endif
{
	int argindex;
	struct bridge *br;
	struct command *cmd;

	br_init();

	if (argc < 2)
		goto help;

	if ((cmd = br_command_lookup(argv[1])) == NULL) {
		fprintf(stderr, "never heard of command [%s]\n", argv[1]);
		goto help;
	}

	argindex = 2;
	br = NULL;
	if (cmd->needs_bridge_argument) {
		if (argindex >= argc) {
			fprintf(stderr, "this option requires a bridge name as argument\n");
			return 1;
		}

		br = br_find_bridge(argv[argindex]);

		if (br == NULL) {
			fprintf(stderr, "bridge %s doesn't exist!\n", argv[argindex]);
			return 1;
		}

		argindex++;
	}

	if (argc - argindex != cmd->num_string_arguments) {
		fprintf(stderr, "incorrect number of arguments for command\n");
		return 1;
	}

/*+++ Limited Administration, Builder, 2008/03/11 +++*/
#ifdef LIMITED_ADMIN
    cmd->func(br, argv[argindex], argv[argindex+1], argv[argindex+2]);
#else
	cmd->func(br, argv[argindex], argv[argindex+1]);
#endif /*LIMITED_ADMIN*/
/*+++ Limited Administration, Builder, 2008/03/11 +++*/

	return 0;

help:
	help();
	return 1;
}
