#include "defmod.h"

int def_ioctl_get_channel(int skfd, char *ifname)
{
	d_info("%s() calling -> \n", __func__);
    int              ret;
    struct iwreq    wrq;
    int             channel = -1;
    double          freq;
    char buf[8];

    memset(buf, 0, sizeof(buf));

    ret = iw_get_ext(skfd, ifname, SIOCGIWFREQ, &wrq);
    dassert(ret == 0);
    freq = iw_freq2float(&(wrq.u.freq));
    snprintf(buf, sizeof(buf), "%g", freq);
    channel = atoi(buf);
    d_info("%s, channel=%d\n", __func__, channel);
    dassert(channel >=0 );
    return channel;
}

void def_ioctl_get_clients_list(int skfd, const char *ifname, struct dlist_head * head)
{
	d_info("%s() calling -> \n", __func__);
    u_int8_t        buf[24*1024];
    u_int8_t        *ptr;
    struct iwreq    wrq;
    int             len;
    int             ret;
    int             i;
    RT_802_11_MAC_TABLE *mac_table;

    memset(&wrq, 0 ,sizeof(wrq));
    strncpy(wrq.ifr_name, ifname, sizeof(wrq.ifr_name));
    wrq.u.data.pointer = (void *) buf;
    wrq.u.data.length = sizeof(buf);
    ret = ioctl(skfd, RTPRIV_IOCTL_GET_MAC_TABLE, &wrq);
    dassert(ret == 0);

    len = wrq.u.data.length;
    ptr = buf;

    mac_table = (RT_802_11_MAC_TABLE *) ptr;
    for (i=0; i<mac_table->Num; i++)
    {
        struct client_info ci;
        memcpy(&(ci.mac_addr), mac_table->Entry[i].Addr, 6);
        ci.inf      = mac_table->Entry[i].apidx;
        ci.mode     = mac_table->Entry[i].TxRate.field.MODE;
        ci.bw       = mac_table->Entry[i].TxRate.field.BW;
        ci.sgi      = mac_table->Entry[i].TxRate.field.ShortGI;
        ci.mcs      = mac_table->Entry[i].TxRate.field.MCS;
        ci.uptime   = _uptime();
        add_client(head, &ci);
    }

    return;
}

void def_set_mode_rate_phase1(char *mode, char *rate, struct client_info *ci) 
{

    d_info("%s() calling -> \n", __func__);

	char    buf1[MAXLEN], buf2[MAXLEN];
	snprintf(buf1, sizeof(buf1), "N/A");
	snprintf(buf2, sizeof(buf2), "N/A");

	if (ci->mode == 0)
	{
		snprintf(buf1, sizeof(buf1), "11b");
		if (ci->mcs <= 11 && ci->mcs >= 0)
			snprintf(buf2, sizeof(buf2), "%s", B_MCS[ci->mcs]);
	}
	else if (ci->mode == 1)
	{
		snprintf(buf1, sizeof(buf1), "11g");
		if (ci->mcs <= 7 && ci->mcs >= 0)
			snprintf(buf2, sizeof(buf2), "%s", G_MCS[ci->mcs]);
	}
	else if (ci->mode == 2)
	{
		snprintf(buf1, sizeof(buf1), "11n");
		if (ci->bw==0)
		{
			if (ci->sgi==0)
			{
				if (ci->mcs <= 15 && ci->mcs >= 0)
					snprintf(buf2, sizeof(buf2), "%s", N_LGI_20_MCS[ci->mcs]);
			}
			else if (ci->sgi==1)
			{
				if (ci->mcs <= 15 && ci->mcs >= 0)
					snprintf(buf2, sizeof(buf2), "%s", N_SGI_20_MCS[ci->mcs]);
			}
		}
		else if (ci->bw==1)
		{
			if (ci->sgi==0)
			{
				if (ci->mcs <= 15 && ci->mcs >= 0)
					snprintf(buf2, sizeof(buf2), "%s", N_LGI_40_MCS[ci->mcs]);
			}
			else if (ci->sgi==1)
			{
				if (ci->mcs <= 15 && ci->mcs >= 0)
					snprintf(buf2, sizeof(buf2), "%s", N_SGI_40_MCS[ci->mcs]);
			}
		}
	}

	strcpy(mode, buf1);
	strcpy(rate, buf2);

}

