#include "madwifi.h"
#if 1
int madwifi_ioctl_get_channel(int skfd, char *ifname)
{
	d_info("%s() calling -> \n", __func__);
	int ret;
	struct iw_range range;
	struct iwreq wrq;
	int channel = -1;
	double freq;

	ret = iw_get_range_info(skfd, ifname, &range);
	dassert(ret == 0);

	ret = iw_get_ext(skfd, ifname, SIOCGIWFREQ, &wrq);
	dassert(ret == 0);
	freq = iw_freq2float(&(wrq.u.freq));
	channel = iw_freq_to_channel(freq, &range);
	dassert(channel >= 0);
	return channel;
}

void madwifi_ioctl_get_clients_list (int skfd, const char *ifname, struct dlist_head * head)
{
	d_info("%s() calling -> \n", __func__);
	u_int8_t buf[24*1024];
	u_int8_t *ptr;
	struct iwreq wrq;
	int len;
	int ret;

	memset(&wrq, 0, sizeof(wrq));
	strncpy(wrq.ifr_name, ifname, sizeof(wrq.ifr_name));
	wrq.u.data.pointer = (void *) buf;
	wrq.u.data.length = sizeof(buf);
	ret = ioctl(skfd, IEEE80211_IOCTL_STA_INFO, &wrq);
	dassert(ret == 0);

	len = wrq.u.data.length;
	ptr = buf;
	while (len >= sizeof(struct ieee80211req_sta_info))
	{
		struct ieee80211req_sta_info *si;
		struct client_info ci;

		si = (struct ieee80211req_sta_info *) ptr;	
		memcpy(&(ci.mac_addr), si->isi_macaddr, 6);
        ci.inf      = -1; // need patching driver...
		/* STAMODE dennis start */
		if (si->isi_rssi>=50){
			ci.rssi = 100;
		}	
		else  if (si->isi_rssi<=0){	
			ci.rssi = 0;
		}	
		else{
			ci.rssi = si->isi_rssi*2;
			}
		if(si->isi_state & 0x10)   //0x0010 /* power save mode enabled */
			ci.psmode=1;
		else
			ci.psmode=0;
		 /* STAMODE dennis end */
		
		//ci.rssi = si->isi_rssi;
		ci.ssid_index = si->isi_ssid_index;
		ci.rate = (si->isi_rates[si->isi_txrate] & 0x7f)/2;
		/* 802.11b: 1, 2, 5.5, 11 Mbps */
#ifdef ELBOX_PROGS_PRIV_MADWIFI_AP83
		ci.mode = si->isi_stamode;
		ci.ssid_index = 0;
#else
		if (ci.rate == 0 || ci.rate == 1 || ci.rate == 2 || ci.rate == 5 || 
				ci.rate == 11) {
			ci.mode = 0;	/* 802.11b */
		} else {
			ci.mode = 1;	/* 802.11g */
		}
#endif		
		ci.uptime = _uptime();
		/* add this client to list */
		add_client(head, &ci);
		ptr += si->isi_len;
		len -= si->isi_len;
	}
	return;
}

void madwifi_set_mode_rate_phase1(char *mode, char *rate, struct client_info *ci) 
{
	d_info("%s() calling -> \n", __func__);

	char    buf1[MAXLEN], buf2[MAXLEN];
	snprintf(buf1, sizeof(buf1), "N/A");
	snprintf(buf2, sizeof(buf2), "N/A");

	/* mode */
#ifdef ELBOX_PROGS_PRIV_MADWIFI_AP83
	if(ci->mode==0)
		snprintf(buf1, sizeof(buf1), "%s", "11b");
	else if (ci->mode==1)
		snprintf(buf1, sizeof(buf1), "%s", "11g");
	else if (ci->mode==2)	
		snprintf(buf1, sizeof(buf1), "%s", "11n");
	else
		snprintf(buf1, sizeof(buf1), "%s", "11a");
#else
	snprintf(buf1, sizeof(buf1), "%s", (ci->mode) ? "11g" : "11b");
#endif
	/* rate */
	snprintf(buf2, sizeof(buf2), "%d", ci->rate);

	strcpy(mode, buf1);
	strcpy(rate, buf2);
}
#else
int rt2800_ioctl_get_channel(int skfd, char *ifname)
{
	d_info("%s() calling -> \n", __func__);
    int              ret;
    struct iwreq    wrq;
    int             channel = -1;
    double          freq;
    char buf[8];

    memset(buf, 0, sizeof(buf));

    ret = iw_get_ext(skfd, ifname, SIOCGIWFREQ, &wrq);
    dassert(ret == 0);
    freq = iw_freq2float(&(wrq.u.freq));
    snprintf(buf, sizeof(buf), "%g", freq);
    channel = atoi(buf);
    d_info("%s, channel=%d\n", __func__, channel);
    dassert(channel >=0 );
    return channel;
}

void rt2800_ioctl_get_clients_list(int skfd, const char *ifname, struct dlist_head * head)
{
	d_info("%s() calling -> \n", __func__);

    u_int8_t        buf[24*1024];
    u_int8_t        *ptr;
    struct iwreq    wrq;
    int             len;
    int             ret;
    int             i;
    RT_802_11_MAC_TABLE *mac_table;

    memset(&wrq, 0 ,sizeof(wrq));
    strncpy(wrq.ifr_name, ifname, sizeof(wrq.ifr_name));
    wrq.u.data.pointer = (void *) buf;
    wrq.u.data.length = sizeof(buf);
    ret = ioctl(skfd, RTPRIV_IOCTL_GET_MAC_TABLE, &wrq);
    dassert(ret == 0);

    len = wrq.u.data.length;
    ptr = buf;

    mac_table = (RT_802_11_MAC_TABLE *) ptr;
    for (i=0; i<mac_table->Num; i++)
    {
        struct client_info ci;
        memcpy(&(ci.mac_addr), mac_table->Entry[i].Addr, 6);
        //ci.inf      = mac_table->Entry[i].apidx;
        ci.inf      = -1; // need patching driver...
        ci.mode     = mac_table->Entry[i].TxRate.field.MODE;
        ci.bw       = mac_table->Entry[i].TxRate.field.BW;
        ci.sgi      = mac_table->Entry[i].TxRate.field.ShortGI;
        ci.mcs      = mac_table->Entry[i].TxRate.field.MCS;
        ci.uptime   = _uptime();
        /*Client_signal_strength_10071109*//*Added by Erick*/ 
        ci.signal_strength = ((mac_table->Entry[i].rssi)+ 100)*2;
        if (ci.signal_strength > 100) ci.signal_strength = 100;
        if (ci.signal_strength < 1 ) ci.signal_strength = 1 ;
        /*Client_signal_strength_10071109*//*Added by Erick*/ 
        add_client(head, &ci);
    }

    return;
}

void rt2800_set_mode_rate_phase1(char *mode, char *rate, struct client_info *ci) 
{

	d_info("%s() calling -> \n", __func__);

	char    buf1[MAXLEN], buf2[MAXLEN];
	snprintf(buf1, sizeof(buf1), "N/A");
	snprintf(buf2, sizeof(buf2), "N/A");

	if (ci->mode == 0)
	{
		snprintf(buf1, sizeof(buf1), "11b");
		if (ci->mcs <= 11 && ci->mcs >= 0)
			snprintf(buf2, sizeof(buf2), "%s", B_MCS[ci->mcs]);
	}
	else if (ci->mode == 1)
	{
		snprintf(buf1, sizeof(buf1), "11a/g");
		if (ci->mcs <= 7 && ci->mcs >= 0)
			snprintf(buf2, sizeof(buf2), "%s", G_MCS[ci->mcs]);
	}
	else if (ci->mode == 2)
	{
		snprintf(buf1, sizeof(buf1), "11n");
		if (ci->bw==0)
		{
			if (ci->sgi==0)
			{
				if (ci->mcs <= 15 && ci->mcs >= 0)
					snprintf(buf2, sizeof(buf2), "%s", N_LGI_20_MCS[ci->mcs]);
			}
			else if (ci->sgi==1)
			{
				if (ci->mcs <= 15 && ci->mcs >= 0)
					snprintf(buf2, sizeof(buf2), "%s", N_SGI_20_MCS[ci->mcs]);
			}
		}
		else if (ci->bw==1)
		{
			if (ci->sgi==0)
			{
				if (ci->mcs <= 15 && ci->mcs >= 0)
					snprintf(buf2, sizeof(buf2), "%s", N_LGI_40_MCS[ci->mcs]);
			}
			else if (ci->sgi==1)
			{
				if (ci->mcs <= 15 && ci->mcs >= 0)
					snprintf(buf2, sizeof(buf2), "%s", N_SGI_40_MCS[ci->mcs]);
			}
		}
	}

	strcpy(mode, buf1);
	strcpy(rate, buf2);

}
#endif
