#include "rt2800.h"

#ifdef ELBOX_PROGS_PRIV_RT2860V2_STA
unsigned int rt2800_ioctl_get_bridge_status(int skfd, char *ifname)
{
    struct iwreq    wrq;
    unsigned int    data;

    strcpy(wrq.ifr_name, ifname);
    memset(&data, 0, sizeof(NDIS_MEDIA_STATE));
    wrq.u.data.length = sizeof(NDIS_MEDIA_STATE);
    wrq.u.data.pointer = &data;
    wrq.u.data.flags = OID_GEN_MEDIA_CONNECT_STATUS;
    ioctl(skfd, RT_PRIV_IOCTL, &wrq);

    return data;
}
#endif /*ELBOX_PROGS_PRIV_RT2860V2_STA*/

int rt2800_ioctl_get_channel(int skfd, char *ifname)
{
	d_info("%s() calling -> \n", __func__);
    int              ret;
    struct iwreq    wrq;
    int             channel = -1;
    double          freq;
    char buf[8];

    memset(buf, 0, sizeof(buf));

    ret = iw_get_ext(skfd, ifname, SIOCGIWFREQ, &wrq);
    dassert(ret == 0);
    freq = iw_freq2float(&(wrq.u.freq));
    if (freq>2400000000&&freq<2500000000){
        channel = (freq-2407000000)/5000000;
    }else if (freq>3000000000){
        channel = (freq-5000000000)/5000000;
    }else{
    snprintf(buf, sizeof(buf), "%g", freq);
    channel = atoi(buf);
    }
    d_info("%s, channel=%d\n", __func__, channel);
    dassert(channel >=0 );
    return channel;
}

void rt2800_ioctl_get_clients_list(int skfd, const char *ifname, struct dlist_head * head)
{
	d_info("%s() calling -> \n", __func__);

    u_int8_t        buf[24*1024];
    u_int8_t        *ptr;
    struct iwreq    wrq;
    int             len;
    int             ret;
    int             i;
    RT_802_11_MAC_TABLE *mac_table;

    memset(&wrq, 0 ,sizeof(wrq));
    strncpy(wrq.ifr_name, ifname, sizeof(wrq.ifr_name));
    wrq.u.data.pointer = (void *) buf;
    wrq.u.data.length = sizeof(buf);
    ret = ioctl(skfd, RTPRIV_IOCTL_GET_MAC_TABLE, &wrq);
    dassert(ret == 0);

    len = wrq.u.data.length;
    ptr = buf;

    mac_table = (RT_802_11_MAC_TABLE *) ptr;
    for (i=0; i<mac_table->Num; i++)
    {
        struct client_info ci;
        memcpy(&(ci.mac_addr), mac_table->Entry[i].Addr, 6);
        //ci.inf      = mac_table->Entry[i].apidx;
        ci.inf      = -1; // need patching driver...
        ci.mode     = mac_table->Entry[i].TxRate.field.MODE;
        ci.bw       = mac_table->Entry[i].TxRate.field.BW;
        ci.sgi      = mac_table->Entry[i].TxRate.field.ShortGI;
        ci.mcs      = mac_table->Entry[i].TxRate.field.MCS;
        ci.uptime   = _uptime();
        /*Client_signal_strength_10071109*//*Added by Erick*/ 
        ci.signal_strength = ((mac_table->Entry[i].rssi)+ 100)*2;
        if (ci.signal_strength > 100) ci.signal_strength = 100;
        if (ci.signal_strength < 1 ) ci.signal_strength = 1 ;
        /*Client_signal_strength_10071109*//*Added by Erick*/ 
        add_client(head, &ci);
    }

    return;
}

void rt2800_set_mode_rate_phase1(char *mode, char *rate, struct client_info *ci) 
{

	d_info("%s() calling -> \n", __func__);

	char    buf1[MAXLEN], buf2[MAXLEN];
	snprintf(buf1, sizeof(buf1), "N/A");
	snprintf(buf2, sizeof(buf2), "N/A");

	if (ci->mode == 0)
	{
		snprintf(buf1, sizeof(buf1), "11b");
		if (ci->mcs <= 11 && ci->mcs >= 0)
			snprintf(buf2, sizeof(buf2), "%s", B_MCS[ci->mcs]);
	}
	else if (ci->mode == 1)
	{
		snprintf(buf1, sizeof(buf1), "11a/g");
		if (ci->mcs <= 7 && ci->mcs >= 0)
			snprintf(buf2, sizeof(buf2), "%s", G_MCS[ci->mcs]);
	}
	else if (ci->mode == 2)
	{
		snprintf(buf1, sizeof(buf1), "11n");
		if (ci->bw==0)
		{
			if (ci->sgi==0)
			{
				if (ci->mcs <= 15 && ci->mcs >= 0)
					snprintf(buf2, sizeof(buf2), "%s", N_LGI_20_MCS[ci->mcs]);
			}
			else if (ci->sgi==1)
			{
				if (ci->mcs <= 15 && ci->mcs >= 0)
					snprintf(buf2, sizeof(buf2), "%s", N_SGI_20_MCS[ci->mcs]);
			}
		}
		else if (ci->bw==1)
		{
			if (ci->sgi==0)
			{
				if (ci->mcs <= 15 && ci->mcs >= 0)
					snprintf(buf2, sizeof(buf2), "%s", N_LGI_40_MCS[ci->mcs]);
			}
			else if (ci->sgi==1)
			{
				if (ci->mcs <= 15 && ci->mcs >= 0)
					snprintf(buf2, sizeof(buf2), "%s", N_SGI_40_MCS[ci->mcs]);
			}
		}
	}

	strcpy(mode, buf1);
	strcpy(rate, buf2);

}

/*
* Function : rt2800_set_extention_channel			 
* Purpose  : get extention channel according to Ralink Driver 1.6.0.0 spec.
* by Kwest Wan 20071212
*/
#define XML_NODE_BANDWIDTH      "/wireless/bandwidth"
#define XML_NODE_COUNTRYCODE    "/sys/countrycode"
#define XML_RUNTIME_COUNTRYCODE "/runtime/nvram/countrycode"
static int get_channel_num(void)
{
	char buf1[64];
	int chnl_num, code;

	memset(buf1, 0, sizeof(buf1));
	xmldbc_get_wb(NULL, 0, XML_NODE_COUNTRYCODE, buf1, sizeof(buf1));
	if(!strcmp(buf1, ""))
		xmldbc_get_wb(NULL, 0, XML_RUNTIME_COUNTRYCODE, buf1, sizeof(buf1));
	code = atoi(buf1);
	if(code == 840)
		chnl_num = 11;
	else if(code == 826 || code == 152 || code == 392)
		chnl_num = 13;
	else
		chnl_num = 13;

	return chnl_num;
}

int rt2800_set_extention_channel(int channel)
{
	int bw, extchannel;
	int max_num;
	char buf1[64];
	
	memset(buf1, 0, sizeof(buf1));
	xmldbc_get_wb(NULL, 0, XML_NODE_BANDWIDTH, buf1, sizeof(buf1));
	bw = atoi(buf1);
	if(bw == 2 && channel <= 4)
	{
		/*above the channel*/
		extchannel = channel + 4;
	}
	else if(channel > 4 && channel < 8 && bw == 2)
	{
		/* TODO: should check HT_EXTCHA, default 0 */
		extchannel = channel - 4;
	}
	else if(channel >= 8 && bw == 2)
	{
		max_num = get_channel_num();
		if((max_num - channel) < 4)
			extchannel = channel - 4;
		else
			/* TODO: should check HT_EXTCHA, default 0 */
			extchannel = channel - 4;
	}
	else
		extchannel = channel;
	return extchannel;
}

