#define __packed __attribute__((__packed__))

//#include <rtmp_def.h>

#include "wlxmlpatch.h"

#define POLLING_PERIOD			1000	/* polling per second */
#define msleep(x)				usleep((x)*1000)

/* define keywords for EVNS_set  */
#define XML_NODE_CHANNEL		ENVS_set.channelPath
#define XML_NODE_EXT_CHANNEL	ENVS_set.extchnlPath	/* extention channel, Kwest Wan 20071212 */
#define XML_NODE_CLIENT			ENVS_set.clientPath
#define XML_NODE_CLIENT_RECORD	ENVS_set.clientRecPath
#define WLAN_LED_ON				ENVS_set.ledOnCmd	
#define WLAN_LED_BLINK			ENVS_set.ledBlinkCmd
#define WLAN_IF_NAME			ENVS_set.ifName
#define WLAN_MODEL				ENVS.set.model

/* define model keywords for wlxmlpatch  */
#define WLXMLPATCH_MOD_DEFAULT		"DEFAULTMOD"

#ifdef ELBOX_PROGS_PRIV_WLXMLPATCHV2_RALINK_RT2800
#define WLXMLPATCH_MOD_RT2800		"RT2800"
#endif

#ifdef ELBOX_PROGS_PRIV_WLXMLPATCHV2_RALINK_RT61
#define WLXMLPATCH_MOD_RT61			"RT61"
#endif

#ifdef ELBOX_PROGS_PRIV_WLXMLPATCHV2_MADWIFI
#define WLXMLPATCH_MOD_MADWIFI		"MADWIFI"
#endif

/* define function point for wlxmlpatch */
#ifdef ELBOX_PROGS_PRIV_RT2860V2_STA
typedef int (*WLXMLPATCH_GET_BRIDGE_STATUS_FuncP)(int, char*);
#endif /*ELBOX_PROGS_PRIV_RT2860V2_STA*/
typedef int (*WLXMLPATCH_GET_CHANNEL_FuncP)(int, char*);
typedef int (*WLXMLPATCH_GET_EXTCHANNEL_FuncP)(int); /* extention channel, Kwest Wan 20071212 */
typedef void (*WLXMLPATCH_GET_CLIENT_LIST_FuncP)(int, const char *, struct dlist_head *);
typedef void (*WLXMLPATCH_SET_MODE_RATE_FuncP)(char *, char *, struct client_info *);

/* global variable */
static int led_blink=0;
static DLIST_HEAD(clients_head);
static int o_led_ctl = 0;		/* led control.*/
static int signal_terminated = 0;

/* (golbal variable) function pointer  */
WLXMLPATCH_GET_CLIENT_LIST_FuncP	FP_wlxmlGetClistList;
#ifdef ELBOX_PROGS_PRIV_RT2860V2_STA
WLXMLPATCH_GET_BRIDGE_STATUS_FuncP  FP_wlxmlGetBridgeStatus;
#endif /*ELBOX_PROGS_PRIV_RT2860V2_STA*/
WLXMLPATCH_GET_CHANNEL_FuncP		FP_wlxmlGetChannel;
WLXMLPATCH_SET_MODE_RATE_FuncP		FP_wlxmlSetModeRate;
WLXMLPATCH_GET_EXTCHANNEL_FuncP		FP_wlxmlGetExtChannel; /* extention channel, Kwest Wan 20071212 */

/* ENVS_set data structure */
struct environment_set
{
	char cfgPath[MAXLEN];
	char wpsPath[MAXLEN];
	char channelPath[MAXLEN];
	char extchnlPath[MAXLEN]; /* extention channel, Kwest Wan 20071212 */
	char clientPath[MAXLEN];
	char clientRecPath[MAXLEN];
	char ledOnCmd[MAXLEN];
	char ledBlinkCmd[MAXLEN];
	char ifName[32];
	char model[32];
} ENVS_set;


static void signal_handler(int index)
{
	signal_terminated = 1;
}

static void show_usage(int exit_code)
{
	printf( "NAME:\n" 
			"	wlxmlpatch v2.0\n"
			"Usaged: wlxmlpatch [OPTION]\n"
			" -h,\n"
			"	show this help message.\n"
			" -l,\n"
			"	With led control,\n"
			"	and use default sets.\n"
			"	include: 'ifName' ,'nodePath' 'led on command', 'led blink command' and 'model'. \n"
			" -L <ifName> <nodePath> <ledOnCmd> <ledBlinkCmd> <model>,\n"
			"	With led control, \n"
			"	and specify interface name ,node path, led on and blink command, model.\n"
			"	<model>: DEFAULTMOD"
			#ifdef ELBOX_PROGS_PRIV_WLXMLPATCHV2_RALINK_RT2800
			" | RT2800"
			#endif
			#ifdef ELBOX_PROGS_PRIV_WLXMLPATCHV2_RALINK_RT61
			" | RT61"
			#endif
			#ifdef ELBOX_PROGS_PRIV_WLXMLPATCHV2_MADWIFI
			" | MADWIFI"
			#endif
			" \n-S <ifName> <nodePath> <ledOnCmd> <ledBlinkCmd> <model>,\n"
			" \n-S <ifName> <nodePath> <ledOnCmd> <ledBlinkCmd> <model>,\n"
			"	Without led control,\n"
			"	and specify interface name ,node path, led on and blink command, model.\n"
			"	<model>: DEFAULTMOD"
			#ifdef ELBOX_PROGS_PRIV_WLXMLPATCHV2_RALINK_RT2800
			" | RT2800"
			#endif
			#ifdef ELBOX_PROGS_PRIV_WLXMLPATCHV2_RALINK_RT61
			" | RT61"
			#endif
			#ifdef ELBOX_PROGS_PRIV_WLXMLPATCHV2_MADWIFI
			" | MADWIFI"
			#endif
			"\n"
			);
	exit(exit_code);
}
static void setFuncPt(char * model)
{
	/* seting FP_wlxmlGetChannel, FP_wlxmlGetClistList and FP_wlxmlSetModeRate 
	 * to real function pointer by each model. */

	
	#ifdef ELBOX_PROGS_PRIV_WLXMLPATCHV2_RALINK_RT2800
	if ( strcmp(WLXMLPATCH_MOD_RT2800, model) == 0)
	{
#ifdef ELBOX_PROGS_PRIV_RT2860V2_STA
        FP_wlxmlGetBridgeStatus = rt2800_ioctl_get_bridge_status;
#endif /*ELBOX_PROGS_PRIV_RT2860V2_STA*/
        FP_wlxmlGetChannel		= rt2800_ioctl_get_channel;
		FP_wlxmlGetClistList	= rt2800_ioctl_get_clients_list;
		FP_wlxmlSetModeRate		= rt2800_set_mode_rate_phase1;
		FP_wlxmlGetExtChannel	= rt2800_set_extention_channel; /* extention channel, Kwest Wan 20071212 */
	}
	#endif


	#ifdef ELBOX_PROGS_PRIV_WLXMLPATCHV2_RALINK_RT61
    if ( strcmp(WLXMLPATCH_MOD_RT61, model) == 0)
	{
        FP_wlxmlGetChannel		= rt61_ioctl_get_channel;
		FP_wlxmlGetClistList	= rt61_ioctl_get_clients_list;
		FP_wlxmlSetModeRate		= rt61_set_mode_rate_phase1;
	}
	#endif

	#ifdef ELBOX_PROGS_PRIV_WLXMLPATCHV2_MADWIFI
    if ( strcmp(WLXMLPATCH_MOD_MADWIFI, model) == 0)
	{
        FP_wlxmlGetChannel		= madwifi_ioctl_get_channel;
		FP_wlxmlGetClistList	= madwifi_ioctl_get_clients_list;
		FP_wlxmlSetModeRate		= madwifi_set_mode_rate_phase1;
	}
	#endif

    if ( strcmp(WLXMLPATCH_MOD_DEFAULT, model) == 0)
	{
        FP_wlxmlGetChannel		= def_ioctl_get_channel;
		FP_wlxmlGetClistList	= def_ioctl_get_clients_list;
		FP_wlxmlSetModeRate		= def_set_mode_rate_phase1;
	}
}

static void set_envs(char *ifName, char *nodePath, char *ledOnCmd, char *ledBlinkCmd, char *model)
{
	int set = 0;
	char bufStr[MAXLEN];
	memset( bufStr, 0, sizeof(bufStr));

	#ifdef ELBOX_PROGS_PRIV_WLXMLPATCHV2_RALINK_RT2800
	if ( strcmp(WLXMLPATCH_MOD_RT2800, model) == 0 ){
		// set ENVS_set for RT2800
		
		/* config path and wpspath are reserve for future, so i set them staticly. */
		strcpy( ENVS_set.cfgPath,		"/var/run/RT2860.dat");
		strcpy( ENVS_set.wpsPath,		"/wireless/wps/enable");
		/***************************************************************************/

		memset(bufStr, 0, sizeof(bufStr));
		strncpy(bufStr, nodePath, strlen(nodePath) );
		strcat(bufStr, "/channel");
		strcpy(ENVS_set.channelPath, bufStr);

		/* extention channel, Kwest Wan 20071212 */
		memset(bufStr, 0, sizeof(bufStr));
		strncpy(bufStr, nodePath, strlen(nodePath) );
		strcat(bufStr, "/SecondaryChannel");
		strcpy(ENVS_set.extchnlPath, bufStr);
		memset(bufStr, 0, sizeof(bufStr));
		strncpy(bufStr, nodePath, strlen(nodePath) );
		strcat(bufStr, "/client");
		strcpy(ENVS_set.clientPath, bufStr);

		memset(bufStr, 0, sizeof(bufStr));
		strncpy(bufStr, nodePath, strlen(nodePath) );
		strcat(bufStr, "/client_record");
		strcpy( ENVS_set.clientRecPath,	bufStr);

		memset(bufStr, 0, sizeof(bufStr));
		strcpy(bufStr, "usockc /var/run/fresetd_unixsock ");
		strcat(bufStr, ledOnCmd);
		strcpy(ENVS_set.ledOnCmd, bufStr);

		memset( bufStr, 0, sizeof(bufStr));
		strcpy(bufStr, "usockc /var/run/fresetd_unixsock ");
		strcat(bufStr, ledBlinkCmd);
		strcpy( ENVS_set.ledBlinkCmd, bufStr);

		strcpy( ENVS_set.ifName,		ifName);
		strcpy( ENVS_set.model,			WLXMLPATCH_MOD_RT2800);

		set = 1;
	} 

	#endif

	#ifdef ELBOX_PROGS_PRIV_WLXMLPATCHV2_RALINK_RT61
	if ( strcmp(WLXMLPATCH_MOD_RT61, model) == 0){
		// set ENVS_set for RT61
	
		/* config path and wpspath are reserve for future, so i set them staticly. */
		strcpy( ENVS_set.cfgPath,		"/var/run/RT61AP.dat");
		strcpy( ENVS_set.wpsPath,		"/wireless/wps/enable");
		/***************************************************************************/

		memset(bufStr, 0, sizeof(bufStr));
		strncpy(bufStr, nodePath, strlen(nodePath) );
		strcat(bufStr, "/channel");
		strcpy(ENVS_set.channelPath, bufStr);

		memset(bufStr, 0, sizeof(bufStr));
		strncpy(bufStr, nodePath, strlen(nodePath) );
		strcat(bufStr, "/client");
		strcpy(ENVS_set.clientPath, bufStr);

		memset(bufStr, 0, sizeof(bufStr));
		strncpy(bufStr, nodePath, strlen(nodePath) );
		strcat(bufStr, "/client_record");
		strcpy( ENVS_set.clientRecPath,	bufStr);

		memset(bufStr, 0, sizeof(bufStr));
		strcpy(bufStr, "usockc /var/run/fresetd_unixsock ");
		strcat(bufStr, ledOnCmd);
		strcpy(ENVS_set.ledOnCmd, bufStr);

		memset( bufStr, 0, sizeof(bufStr));
		strcpy(bufStr, "usockc /var/run/fresetd_unixsock ");
		strcat(bufStr, ledBlinkCmd);
		strcpy( ENVS_set.ledBlinkCmd, bufStr);

		strcpy( ENVS_set.ifName,		ifName);
		strcpy( ENVS_set.model,			WLXMLPATCH_MOD_RT61);

		set = 1;
	} 

	#endif

	#ifdef ELBOX_PROGS_PRIV_WLXMLPATCHV2_MADWIFI
	if ( strcmp(WLXMLPATCH_MOD_MADWIFI, model) == 0){
		// set ENVS_set for MADWIFI
	
		/* config path and wpspath are reserve for future, so i set them staticly. */
//		strcpy( ENVS_set.cfgPath,		"/var/run/RT61AP.dat");
		strcpy( ENVS_set.wpsPath,		"/wireless/wps/enable");
		/***************************************************************************/

		memset(bufStr, 0, sizeof(bufStr));
		strncpy(bufStr, nodePath, strlen(nodePath) );
		strcat(bufStr, "/channel");
		strcpy(ENVS_set.channelPath, bufStr);

		memset(bufStr, 0, sizeof(bufStr));
		strncpy(bufStr, nodePath, strlen(nodePath) );
		strcat(bufStr, "/client");
		strcpy(ENVS_set.clientPath, bufStr);

		memset(bufStr, 0, sizeof(bufStr));
		strncpy(bufStr, nodePath, strlen(nodePath) );
		strcat(bufStr, "/client_record");
		strcpy( ENVS_set.clientRecPath,	bufStr);

		memset(bufStr, 0, sizeof(bufStr));
		strcpy(bufStr, "usockc /var/run/fresetd_unixsock ");
		strcat(bufStr, ledOnCmd);
		strcpy(ENVS_set.ledOnCmd, bufStr);

		memset( bufStr, 0, sizeof(bufStr));
		strcpy(bufStr, "usockc /var/run/fresetd_unixsock ");
		strcat(bufStr, ledBlinkCmd);
		strcpy( ENVS_set.ledBlinkCmd, bufStr);

		strcpy( ENVS_set.ifName,		ifName);
		strcpy( ENVS_set.model,			WLXMLPATCH_MOD_MADWIFI);

		set = 1;
	} 
	#endif

	if ( set == 0 ){
		// set ENVS_set for default
		
		/* config path and wpspath are reserve for future, so i set them staticly. */
		strcpy( ENVS_set.cfgPath,		"/var/run/RT2860AP.dat");
		strcpy( ENVS_set.wpsPath,		"/wireless/wps/enable");
		/***************************************************************************/

		memset(bufStr, 0, sizeof(bufStr));
		strncpy(bufStr, nodePath, strlen(nodePath) );
		strcat(bufStr, "/channel");
		strcpy(ENVS_set.channelPath, bufStr);

		memset(bufStr, 0, sizeof(bufStr));
		strncpy(bufStr, nodePath, strlen(nodePath) );
		strcat(bufStr, "/client");
		strcpy(ENVS_set.clientPath, bufStr);

		memset(bufStr, 0, sizeof(bufStr));
		strncpy(bufStr, nodePath, strlen(nodePath) );
		strcat(bufStr, "/client_record");
		strcpy( ENVS_set.clientRecPath,	bufStr);

		memset(bufStr, 0, sizeof(bufStr));
		strcpy(bufStr, "usockc /var/run/fresetd_unixsock ");
		strcat(bufStr, ledOnCmd);
		strcpy(ENVS_set.ledOnCmd, bufStr);

		memset( bufStr, 0, sizeof(bufStr));
		strcpy(bufStr, "usockc /var/run/fresetd_unixsock ");
		strcat(bufStr, ledBlinkCmd);
		strcpy( ENVS_set.ledBlinkCmd, bufStr);

		strcpy( ENVS_set.ifName,		ifName);
		strcpy( ENVS_set.model,			WLXMLPATCH_MOD_DEFAULT);
	}

	setFuncPt(ENVS_set.model);
	
}
static void show_ENVS_set()
{
 
 	d_info("%s( )\n",	__func__);
	d_info("ENVS_set{\n");
	d_info("	cfgPath         = %s\n", ENVS_set.cfgPath);
	d_info("	wpsPath         = %s\n", ENVS_set.wpsPath);
	d_info("	channlePath     = %s\n", ENVS_set.channelPath);
	d_info("	clientPath      = %s\n", ENVS_set.clientPath);
	d_info("	clientRecPath   = %s\n", ENVS_set.clientRecPath);
	d_info("	ledOnCmd        = %s\n", ENVS_set.ledOnCmd);
	d_info("	ledBlinkCmd     = %s\n", ENVS_set.ledBlinkCmd);
	d_info("	ifName          = %s\n", ENVS_set.ifName);
	d_info("	model           = %s\n", ENVS_set.model);
	d_info("}\n");

}

static void parse_args(int argc, char * argv[])
{
	int opt;
	while((opt = getopt(argc, argv, "S:L:lh")) > 0)
	{
		switch(opt)
		{
		case 'h':
				show_usage(0);
				//show_ENVS_set();
				break;
		case 'l': 
				o_led_ctl=1;
				set_envs( "ra0", "/runtime/stats/wireless", "RADIO_ON", "RADIO_BLINK", WLXMLPATCH_MOD_DEFAULT);
				show_ENVS_set();
				break;
		case 'L':
				o_led_ctl=1;
				/*********************************************************************************
				 *	set_envs(ifName, nodePath, ledOnCmd, ledBlinkCmd, model)(....)				 */
				set_envs( optarg, argv[optind], argv[optind+1], argv[optind+2], argv[optind+3] );
				/*********************************************************************************/
				show_ENVS_set();
				break;
		case 'S': 
				o_led_ctl=0;
				/*********************************************************************************
				 *	set_envs(ifName, nodePath, ledOnCmd, ledBlinkCmd, model)(....)				 */
				set_envs( optarg, argv[optind], argv[optind+1], argv[optind+2], argv[optind+3] );
				/*********************************************************************************/
				show_ENVS_set();
				break;
		default:
				show_usage(-1);	
				//show_ENVS_set();
				break;
		}
	}
}
long _uptime(void)
{
	struct sysinfo info;
	sysinfo(&info);
	return info.uptime;
}

int add_client(struct dlist_head *head, const struct client_info * client_info)
{
	int ret = -1;
	struct client_info * ci = NULL;

	ci = (struct client_info *) malloc (sizeof(struct client_info));
	dassert(ci);
	if (ci)
	{
		memcpy (ci, client_info, sizeof(struct client_info));
		INIT_DLIST_HEAD(&ci->link);
		dlist_add_tail(&ci->link, head);
		ret = 0;
	}
	return ret;
}

static struct client_info * find_client(struct dlist_head * head, const char * mac_addr)
{
	struct client_info	* ci;
	struct dlist_head	* entry;

	entry = head->next;
	while (entry != head)
	{
		ci = dlist_entry(entry, struct client_info, link);
		if (memcmp(ci->mac_addr, mac_addr, sizeof(ci->mac_addr)) == 0)	return ci;
		entry = entry->next;
	}

	/* failed to find the client */
	return NULL;
};

static int free_all_clients(struct dlist_head * head)
{
	struct client_info	* ci;
	struct dlist_head	* entry;

	while (!dlist_empty(head))
	{
		entry = head->next;
		ci = dlist_entry(entry, struct client_info, link);
		dassert(ci);
		dlist_del(&ci->link);
		free (ci);
	}
	return 0;
}

static int update_clients_time(struct dlist_head * list_new, struct dlist_head * list_old)
{
	struct client_info	* ci_new, * ci_old;
	struct dlist_head	* entry;

	dassert(list_new && list_old);

	entry = list_new->next;
	while (entry != list_new)
	{
		ci_new = dlist_entry(entry, struct client_info, link);
		ci_old = find_client(list_old, ci_new->mac_addr);
		if (ci_old) ci_new->uptime = ci_old->uptime;
		entry = entry->next;
	}
	return 0;
}

static void get_clients_list(int skfd, char *ifname, struct dlist_head *head)
{
	DLIST_HEAD(old_list);
	struct dlist_head * entry;

	while(!dlist_empty(head))
	{
		entry = head->next;
		dlist_del(entry);
		dlist_add(entry, &old_list);
	}

	/* get clients list from driver */
	(*FP_wlxmlGetClistList)(skfd, ifname, head);
	update_clients_time(head, &old_list);

	free_all_clients(&old_list);
	return;
}



static void set_db(const char * node, const char * value)
{
	int ret;
	d_info("%s, node=%s\n", __func__, node);
	d_info("%s, value=%s\n", __func__, value);
	ret = xmldbc_set(NULL, 0, node, value);
	dassert(ret == 0);
}

static void set_mode_rate(const char * mode, const char * rate, int index )
{
	char getPath[MAXLEN];

	memset(getPath, 0, sizeof(getPath));

	snprintf(getPath, sizeof(getPath), "%s:%d/mode", XML_NODE_CLIENT, index);
	set_db(getPath, mode);
	snprintf(getPath, sizeof(getPath), "%s:%d/rate", XML_NODE_CLIENT, index);
	set_db(getPath, rate);
}

static void del_clients_in_xmldb(int index)
{
	char	value[MAXLEN], getPath[MAXLEN];
	int 	max_index = 0;
	memset(value, 0, sizeof(value));
	memset(getPath, 0, sizeof(getPath));

	snprintf(getPath, sizeof(getPath), "%s#", XML_NODE_CLIENT);
	xmldbc_get_wb(NULL, 0, getPath, value, sizeof(value));

	max_index = atoi(value);

	for (; index <= max_index; index++)
	{
		snprintf(getPath, sizeof(getPath), "%s:%d", XML_NODE_CLIENT , index);
		xmldbc_del(NULL, 0, getPath);
	}
}

static int get_cur_list_num(struct dlist_head * head)
{
	int count = 0;
	struct dlist_head * tmp_head = head;
	while ( head->next != tmp_head ) {
		count++;
		head = head->next;
	}
	return count;	
}

static void led_ctl(void)
{
	int cur_client_num = 0;
	cur_client_num = get_cur_list_num(&clients_head);
	d_info("cur_client_num=%d\n", cur_client_num);
	if (cur_client_num > 0)
	{
		if(!led_blink)
		{
			led_blink=1;
			d_info("%s\n",WLAN_LED_BLINK);
			system(WLAN_LED_BLINK);
		}
	}
	else if (led_blink)
	{
		led_blink=0;
		d_info("%s\n",WLAN_LED_ON);
		system(WLAN_LED_ON);
	}
	
}


static int ralink_poll(void)
{
	int		i;
	int		skfd = -1;
	char	value[MAXLEN], getPath[MAXLEN], setPath[MAXLEN];
	char	buf1[MAXLEN], buf2[MAXLEN];
	int		index, max_index;
	struct	dlist_head * entry;
	struct	client_info *ci;
	long	uptime;
	unsigned int conn_status=0;

	memset(value, 0, sizeof(value));
	memset(getPath, 0, sizeof(getPath));
	memset(setPath, 0, sizeof(setPath));
	memset(buf1, 0, sizeof(buf1));
	memset(buf2, 0, sizeof(buf2));
	
	d_info("=================================================================\n", __func__);
	
	/* use iwlib (wireless tools) to open wireless socket and get the info */
	if ((skfd = iw_sockets_open()) < 0)
	{
		d_error("%s: can not open iw socket !!\n", __FUNCTION__);
		return -1;
	}

#ifdef ELBOX_PROGS_PRIV_RT2860V2_STA
    conn_status = FP_wlxmlGetBridgeStatus(skfd, WLAN_IF_NAME);
    if (conn_status==1){
        set_db("/runtime/bridge/status", "connected");
    }else if (conn_status==0){
        set_db("/runtime/bridge/status", "disconnected");
    }
#endif /*#ifdef ELBOX_PROGS_PRIV_RT2860V2_STA*/    

	/* get channel */
	snprintf(value, sizeof(value), "%d", (*FP_wlxmlGetChannel)(skfd, WLAN_IF_NAME)	);
	set_db(XML_NODE_CHANNEL, value);

	/* get extention channel, Kwest Wan 20071206 */
	if(FP_wlxmlGetExtChannel){
		int channel = atoi(value);
		memset(value, 0, sizeof(value));
		snprintf(value, sizeof(value), "%d",(*FP_wlxmlGetExtChannel)(channel));
		set_db(XML_NODE_EXT_CHANNEL, value);
	}

	/* get clients info form driver */
	get_clients_list(skfd, WLAN_IF_NAME, &clients_head);

	/* control the WLAN LED. */
	if (o_led_ctl)
		led_ctl();


	/* Go through all clients */
	index = 1;
	entry = clients_head.next;
	while (entry != &clients_head)
	{
		char mac[18];
		ci = dlist_entry(entry, struct client_info, link);

		/* mac address */
		snprintf(setPath, sizeof(setPath), "%s:%d/mac", XML_NODE_CLIENT , index);
		snprintf(mac, sizeof(mac), "%02X:%02X:%02X:%02X:%02X:%02X",
				ci->mac_addr[0], ci->mac_addr[1], ci->mac_addr[2],
				ci->mac_addr[3], ci->mac_addr[4], ci->mac_addr[5]);
		set_db(setPath, mac);

		/* client histroy record */
		snprintf(getPath, sizeof(getPath), "%s#", XML_NODE_CLIENT_RECORD);
		xmldbc_get_wb(NULL, 0, getPath, value, sizeof(value));
		max_index = atoi(value);

		/* compare current 'client list' and 'client recoreds on xmldb'. 
		 * If existed, do nothing, else appending new node to client records on xmldb */
		for (i=1; i <= max_index; i++)
		{
			snprintf(getPath, sizeof(getPath), "%s:%d/mac", XML_NODE_CLIENT_RECORD, i);
			xmldbc_get_wb(NULL, 0, getPath, value, sizeof(value));
			if (!strncmp(value, mac, sizeof(mac))) break;
		}
		if (i > max_index)
		{
			snprintf(setPath, sizeof(setPath), "%s:%d/mac", XML_NODE_CLIENT_RECORD, i);
			set_db(setPath, mac);
		}

		/* uptime - the stamp of that moment of client joining */
		snprintf(setPath, sizeof(setPath), "%s:%d/uptime", XML_NODE_CLIENT, index);
		snprintf(value, sizeof(value), "%lu", ci->uptime);
		set_db(setPath, value);
		
		/* time - the time had pass through that client connected in seconds */
		uptime = _uptime() - ci->uptime;
		dassert(uptime >= 0);
		snprintf(setPath, sizeof(setPath), "%s:%d/time", XML_NODE_CLIENT, index);
		snprintf(value, sizeof(value), "%lu", uptime);
		set_db(setPath, value);
 /* STAMODE dennis start */
#ifdef ELBOX_PROGS_PRIV_MADWIFI_AP83
		/* rssi dennis add foe 2553 */
		{
		 snprintf(setPath, sizeof(setPath), "%s:%d/rssi", XML_NODE_CLIENT, index);
		 snprintf(value, sizeof(value), "%d", ci->rssi);
		 set_db(setPath, value);
		}
 /* STAMODE dennis end */
#else  /* STAMODE dennis  */
		/*Client_signal_strength_10071109*//*Added by Erick*/
		{
		    snprintf(setPath, sizeof(setPath), "%s:%d/rssi", XML_NODE_CLIENT, index);
		    snprintf(value, sizeof(value), "%d", ci->signal_strength);
		    set_db(setPath, value);
		}
#endif /* STAMODE dennis */
		/* mode and rate */
		(*FP_wlxmlSetModeRate)(buf1, buf2, ci);
		set_mode_rate(buf1, buf2, index);
 /* STAMODE dennis start */
		/*ssid index,  vap index (0: primary, 1~7: multi-ssid)*/
#ifdef ELBOX_PROGS_PRIV_MADWIFI_AP83		
		{
			snprintf(setPath, sizeof(setPath), "%s:%d/ssid_index", XML_NODE_CLIENT, index);
		    snprintf(value, sizeof(value), "%d", ci->ssid_index);
		    set_db(setPath, value);
		}

		{
			snprintf(setPath, sizeof(setPath), "%s:%d/psmode", XML_NODE_CLIENT, index);
		    snprintf(value, sizeof(value), "%d", ci->psmode);
		    set_db(setPath, value);
		}
#endif		
 /* STAMODE dennis end */
		/* interface */
		snprintf(setPath, sizeof(setPath), "%s:%d/inf", XML_NODE_CLIENT,index);
		
		char viaIf[MAXLEN]= "N/A";
		if (ci->inf >=  0){ 
			snprintf(viaIf, strlen(ENVS_set.ifName), "%s", ENVS_set.ifName);
			snprintf(viaIf, sizeof(viaIf), "%s%d", viaIf, ci->inf);
		}
		snprintf(value, sizeof(value), "%s", viaIf);
		set_db(setPath, value);
		
		entry = entry->next;
		
		index++;
	}
	
	/* delete the old clients in xmldb */
	 del_clients_in_xmldb(index);
	
	if (skfd >= 0) close (skfd);

	return 0;
}


int main (int argc, char **argv)
{
	int i, j;
	int ret=0;
	char value[64];

	if (argc == 1)
		show_usage(-1);
	else
		parse_args(argc, argv);
	
	signal(SIGINT, signal_handler);
	signal(SIGTERM, signal_handler);
	signal(SIGKILL, signal_handler);

	memset(value, 0, sizeof(value));
	
	while(!signal_terminated)
	{
		if (ralink_poll())
		{
			ret = -1;
			break;
		}
		msleep(POLLING_PERIOD);
	}

	free_all_clients(&clients_head);
	
	/* clean the nodes. */
	char buf1[MAXLEN];
	memset(buf1, 0, sizeof(buf1));
	snprintf(buf1, sizeof(buf1), "%s#", XML_NODE_CLIENT);
	xmldbc_get_wb(NULL, 0, buf1, value, sizeof(value));
	j = atoi(value);
	for (i=0; i<j; i++)
	{
		sprintf(value, "%s:%d", XML_NODE_CLIENT, i+1);
		xmldbc_del(NULL, 0, value);
	}

	d_info("%s: terminated with ret=%d\n", __func__, ret);
	return ret;
}

